package imagescience.features;

import imagescience.images.Axes;
import imagescience.images.ByteImage;
import imagescience.images.Coordinates;
import imagescience.images.Dimensions;
import imagescience.images.FloatImage;
import imagescience.images.PixelSizes;
import imagescience.images.ShortImage;
import imagescience.utilities.Messenger;
import imagescience.utilities.Progressor;
import imagescience.utilities.Timer;

// **********************************************************************
// COPYRIGHT NOTICE AND DISCLAIMER
//
// Copyright (c) Erik Meijering. All rights reserved.
//
// Permission to copy and use this software and accompanying
// documentation for educational, research, and not-for-profit
// purposes, without fee and without a signed licensing agreement,
// is hereby granted, provided that the above copyright notice, this
// paragraph and the following two paragraphs appear in all copies.
// The copyright holder is free to make upgraded or improved versions
// of the software available for a fee or commercially only.
//
// IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE TO ANY PARTY FOR
// DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
// OF ANY KIND WHATSOEVER, ARISING OUT OF THE USE OF THIS SOFTWARE
// AND ITS DOCUMENTATION, EVEN IF HE HAS BEEN ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//
// THE COPYRIGHT HOLDER SPECIFICALLY DISCLAIMS ANY WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
// SOFTWARE AND ACCOMPANYING DOCUMENTATION IS PROVIDED "AS IS".
// THE COPYRIGHT HOLDER HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
// SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
//
// **********************************************************************


/** Provides methods for computing Gaussian-scaled derivatives of
    images. */
public final class Differentiator {

    /** The largest supported order of differentiation. */
    public static final int MAX_ORDER = 5;

    /** Method for differentiation of images of type
	<code>ByteImage</code>. Calls <code>fdiff()</code> with
	the same parameters if <code>Features.sepSmooth()</code> yields
	<code>true</code>, and <code>gauss()</code> if it yields
	<code>false</code>.
    */
    public FloatImage run(final ByteImage image,
			  final float scale,
			  final int xorder,
			  final int yorder,
			  final int zorder) {
	
	if (Features.sepSmooth()) return fdiff(new FloatImage(image),scale,xorder,yorder,zorder,false);
	else return gauss(new FloatImage(image),scale,xorder,yorder,zorder,false);
    }

    /** Method for differentiation of images of type
	<code>ShortImage</code>. Calls <code>fdiff()</code> with
	the same parameters if <code>Features.sepSmooth()</code> yields
	<code>true</code>, and <code>gauss()</code> if it yields
	<code>false</code>.
    */
    public FloatImage run(final ShortImage image,
			  final float scale,
			  final int xorder,
			  final int yorder,
			  final int zorder) {
	
	if (Features.sepSmooth()) return fdiff(new FloatImage(image),scale,xorder,yorder,zorder,false);
	else return gauss(new FloatImage(image),scale,xorder,yorder,zorder,false);
    }

    /** Method for differentiation of images of type
	<code>FloatImage</code>. Calls <code>fdiff()</code> with
	the same parameters if <code>Features.sepSmooth()</code> yields
	<code>true</code>, and <code>gauss()</code> if it yields
	<code>false</code>.
    */
    public FloatImage run(final FloatImage image,
			  final float scale,
			  final int xorder,
			  final int yorder,
			  final int zorder,
			  final boolean newimage) {
	
	if (Features.sepSmooth()) return fdiff(image,scale,xorder,yorder,zorder,newimage);
	else return gauss(image,scale,xorder,yorder,zorder,newimage);
    }

    /** Method for Gaussian smoothing of images of type
	<code>ByteImage</code>. See the <code>smooth</code>
	method for <code>FloatImage</code> objects for details on the
	parameters and exceptions.
    */
    public FloatImage smooth(final ByteImage image, final float scale) {
	
	return gauss(image,scale,0,0,0);
    }

    /** Method for Gaussian smoothing of images of type
	<code>ShortImage</code>. See the <code>smooth</code>
	method for <code>FloatImage</code> objects for details on the
	parameters and exceptions.
    */
    public FloatImage smooth(final ShortImage image, final float scale) {
	
	return gauss(image,scale,0,0,0);
    }

    /** Method for Gaussian smoothing of images of type
	<code>FloatImage</code>.

	@param image the input image to be smoothed.

	@param scale the smoothing scale.<br><ul> The scale is equal to
	the standard deviation of the Gaussian kernel and must be larger
	than or equal to zero. In order to enforce physical isotropy,
	for each dimension, the scale is divided by the pixel size of
	the image in that dimension.</ul>

	@param newimage indicates whether or not to return the result as
	a new image.<br><ul> If this parameter is <code>true</code>, the
	result of smoothing is returned as a newly created
	<code>FloatImage</code> object. If it is
	<code>false</code>, the input image is overwritten by the
	smoothed image.</ul>
	
	@return the smoothed image.

	@exception IllegalArgumentException if the scale is negative.

	@exception NullPointerException if the input image is
	<code>null</code>.
    */
    public FloatImage smooth(final FloatImage image,
			     final float scale,
			     final boolean newimage) {
	
	return gauss(image,scale,0,0,0,newimage);
    }

    /** Method for Gaussian-based differentiation of images of type
	<code>ByteImage</code>. See the <code>gauss</code> method
	for <code>FloatImage</code> objects for details on the
	parameters and exceptions.
    */
    public FloatImage gauss(final ByteImage image,
			    final float scale,
			    final int xorder,
			    final int yorder,
			    final int zorder) {
	
	return gauss(new FloatImage(image),scale,xorder,yorder,zorder,false);
    }

    /** Method for Gaussian-based differentiation of images of type
	<code>ShortImage</code>. See the <code>gauss</code>
	method for <code>FloatImage</code> objects for details on the
	parameters and exceptions.
    */
    public FloatImage gauss(final ShortImage image,
			    final float scale,
			    final int xorder,
			    final int yorder,
			    final int zorder) {
	
	return gauss(new FloatImage(image),scale,xorder,yorder,zorder,false);
    }

    /** Method for Gaussian-based differentiation of images of type
	<code>FloatImage</code>.

	@param image the input image to be differentiated.

	@param scale the smoothing scale at which derivatives are
	computed.<br><ul> The scale is equal to the standard deviation
	of the Gaussian kernel and must be larger than or equal to
	zero. In order to enforce physical isotropy, for each dimension,
	the scale is divided by the pixel size of the image in that
	dimension.</ul>

	@param xorder <code>yorder</code> - <code>zorder</code> - the
	order of differentiation in the x-, y-, and
	z-dimension.<br><ul>Currently, the largest supported order of
	differentiation is 5. If the order in any dimension is zero, the
	method applies smoothing in that dimension. Note that if the
	order of differentiation is larger than zero for a given
	dimension and the image size equals one in that dimension, the
	result is always zero.</ul>
	
	@param newimage indicates whether or not to return the result as
	a new image.<br><ul> If this parameter is <code>true</code>, the
	result of differentiation is returned as a newly created
	<code>FloatImage</code> object. If it is <code>false</code>, the
	input image is overwritten by the derivative image.</ul>
	
	@return the derivative image.

	@exception IllegalArgumentException if the scale is negative or
	if any of the differentiation-order parameters is negative or
	larger than <code>Differentiator.MAX_ORDER</code>.

	@exception NullPointerException if the input image is
	<code>null</code>.
    */
    public FloatImage gauss(final FloatImage image,
			    final float scale,
			    final int xorder,
			    final int yorder,
			    final int zorder,
			    final boolean newimage) {
	
	Features.printVersion("Differentiator");

	final Timer tmr = new Timer(); tmr.start();

	// Initialize:
	checkArguments(scale,xorder,yorder,zorder);

	final Dimensions dims = image.dimensions();
	msg.debug("X-size of input image: " + dims.x);
	msg.debug("Y-size of input image: " + dims.y);
	msg.debug("Z-size of input image: " + dims.z);
	msg.debug("T-size of input image: " + dims.t);

	final PixelSizes ps = image.pixelsizes();
	final float fhx = ps.x;
	final float fhy = ps.y;
	final float fhz = ps.z;
	msg.debug("Pixel size in x-dimension: " + fhx);
	msg.debug("Pixel size in y-dimension: " + fhy);
	msg.debug("Pixel size in z-dimension: " + fhz);

	final FloatImage deriv = (newimage ? image.duplicate() : image);

	final Progressor pgs = new Progressor();
	final boolean bScaleZero = (scale == 0.0f);
	if (bScaleZero) pgs.steps(0);
	else pgs.steps((dims.x > 1 ? dims.t*dims.z*dims.y : 0) +
		       (dims.y > 1 ? dims.t*dims.z*dims.x : 0) +
		       (dims.z > 1 ? dims.t*dims.z*dims.y : 0));
	
	msg.debug("Combining smoothing and differentiation");
	if (!bScaleZero) msg.debug("Gaussian differentiation at scale "+scale+" over pixel size");

	// Differentiate in x-dimension:
	if (bScaleZero || dims.x == 1)
	    if (xorder == 0) msg.debug("No operation in x-dimension");
	    else { msg.debug("Zeroing in x-dimension"); deriv.set(0.0f); }
	else {
	    final String derivInfo = derivInfo(xorder);
	    msg.debug(derivInfo+" in x-dimension");
	    final float fScaleX = scale/fhx;
	    final float[] gaussderiv = kernel(fScaleX,xorder,dims.x);
	    final int gdlenm1 = gaussderiv.length - 1;
	    final float[] ain = new float[dims.x + 2*gdlenm1];
	    final float[] aout = new float[dims.x];
	    final Coordinates coords = new Coordinates();
	    pgs.status(derivInfo+" in x-dimension");
	    deriv.axes(Axes.X);
	    for (coords.t=0; coords.t<dims.t; ++coords.t)
		for (coords.z=0; coords.z<dims.z; ++coords.z)
		    for (coords.y=0; coords.y<dims.y; ++coords.y) {
			coords.x = -gdlenm1; deriv.get(coords,ain);
			convolve(ain,aout,gaussderiv);
			coords.x = 0; deriv.set(coords,aout);
			pgs.step();
		    }
	}

	// Differentiate in y-dimension:
	if (bScaleZero || dims.y == 1)
	    if (yorder == 0) msg.debug("No operation in y-dimension");
	    else { msg.debug("Zeroing in y-dimension"); deriv.set(0.0f); }
	else {
	    final String derivInfo = derivInfo(yorder);
	    msg.debug(derivInfo+" in y-dimension");
	    final float fScaleY = scale/fhy;
	    final float[] gaussderiv = kernel(fScaleY,yorder,dims.y);
	    final int gdlenm1 = gaussderiv.length - 1;
	    final float[] ain = new float[dims.y + 2*gdlenm1];
	    final float[] aout = new float[dims.y];
	    final Coordinates coords = new Coordinates();
	    pgs.status(derivInfo+" in y-dimension");
	    deriv.axes(Axes.Y);
	    for (coords.t=0; coords.t<dims.t; ++coords.t)
		for (coords.z=0; coords.z<dims.z; ++coords.z)
		    for (coords.x=0; coords.x<dims.x; ++coords.x) {
			coords.y = -gdlenm1; deriv.get(coords,ain);
			convolve(ain,aout,gaussderiv);
			coords.y = 0; deriv.set(coords,aout);
			pgs.step();
		    }
	}

	// Differentiate in z-dimension:
	if (bScaleZero || dims.z == 1)
	    if (zorder == 0) msg.debug("No operation in z-dimension");
	    else { msg.debug("Zeroing in z-dimension"); deriv.set(0.0f); }
	else {
	    final String derivInfo = derivInfo(zorder);
	    msg.debug(derivInfo+" in z-dimension");
	    final float fScaleZ = scale/fhz;
	    final float[] gaussderiv = kernel(fScaleZ,zorder,dims.z);
	    final int gdlenm1 = gaussderiv.length - 1;
	    final float[] ain = new float[dims.z + 2*gdlenm1];
	    final float[] aout = new float[dims.z];
	    final Coordinates coords = new Coordinates();
	    pgs.status(derivInfo+" in z-dimension");
	    deriv.axes(Axes.Z);
	    for (coords.t=0; coords.t<dims.t; ++coords.t)
		for (coords.y=0; coords.y<dims.y; ++coords.y) {
		    for (coords.x=0; coords.x<dims.x; ++coords.x) {
			coords.z = -gdlenm1; deriv.get(coords,ain);
			convolve(ain,aout,gaussderiv);
			coords.z = 0; deriv.set(coords,aout);
		    }
		    pgs.step(dims.z);
		}
	}

	pgs.end();
	tmr.stop(msg);

	return deriv;
    }

    /** Method for finite-differences based differentiation of images of
	type <code>ByteImage</code>. See the <code>fdiff</code>
	method for <code>FloatImage</code> objects for details on the
	parameters and exceptions.
    */
    public FloatImage fdiff(final ByteImage image,
			    final float scale,
			    final int xorder,
			    final int yorder,
			    final int zorder) {
	
	return fdiff(new FloatImage(image),scale,xorder,yorder,zorder,false);
    }

    /** Method for finite-differences based differentiation of images of
	type <code>ShortImage</code>. See the <code>fdiff</code>
	method for <code>FloatImage</code> objects for details on the
	parameters and exceptions.
    */
    public FloatImage fdiff(final ShortImage image,
			    final float scale,
			    final int xorder,
			    final int yorder,
			    final int zorder) {
	
	return fdiff(new FloatImage(image),scale,xorder,yorder,zorder,false);
    }

    /** Method for finite-differences based differentiation of images of
	type <code>FloatImage</code>.

	@param image the input image to be differentiated.

	@param scale the smoothing scale at which derivatives are
	computed.<br><ul> The scale is equal to the standard deviation
	of the Gaussian kernel applied before differentiation and must
	be larger than or equal to zero. In order to enforce physical
	isotropy, for each dimension, the scale is divided by the pixel
	size of the image in that dimension.</ul>

	@param xorder <code>yorder</code> - <code>zorder</code> - the
	order of differentiation in the x-, y-, and
	z-dimension.<br><ul>Currently, the largest supported order of
	differentiation is 5. If the order in any dimension is zero, the
	method does nothing in that dimension. Note that if the order of
	differentiation is larger than zero for a given dimension and
	the image size equals one in that dimension, the result is
	always zero.</ul>
	
	@param newimage indicates whether or not to return the result as
	a new image.<br><ul> If this parameter is <code>true</code>, the
	result of differentiation is returned as a newly created
	<code>FloatImage</code> object. If it is
	<code>false</code>, the input image is overwritten by the
	derivative image.</ul>
	
	@return the derivative image.

	@exception IllegalArgumentException if the scale is negative or
	if any of the differentiation-order parameters is negative or
	larger than <code>Differentiator.MAX_ORDER</code>.

	@exception NullPointerException if the input image is
	<code>null</code>.
    */
    public FloatImage fdiff(final FloatImage image,
			    final float scale,
			    final int xorder,
			    final int yorder,
			    final int zorder,
			    final boolean newimage) {

	Features.printVersion("Differentiator");

	final Timer tmr = new Timer(); tmr.start();

	// Initialize:
	checkArguments(scale,xorder,yorder,zorder);

	final Dimensions dims = image.dimensions();
	msg.debug("X-size of input image: " + dims.x);
	msg.debug("Y-size of input image: " + dims.y);
	msg.debug("Z-size of input image: " + dims.z);
	msg.debug("T-size of input image: " + dims.t);

	final PixelSizes ps = image.pixelsizes();
	final float fhx = ps.x;
	final float fhy = ps.y;
	final float fhz = ps.z;
	msg.debug("Pixel size in x-dimension: " + fhx);
	msg.debug("Pixel size in y-dimension: " + fhy);
	msg.debug("Pixel size in z-dimension: " + fhz);

	final FloatImage deriv = (newimage ? image.duplicate() : image);

	final Progressor pgs = new Progressor();
	pgs.steps(((xorder > 0 && dims.x > 1) ? dims.t*dims.z*dims.y : 0) +
		  ((yorder > 0 && dims.y > 1) ? dims.t*dims.z*dims.x : 0) +
		  ((zorder > 0 && dims.z > 1) ? dims.t*dims.z*dims.y : 0));

	msg.debug("Seperating smoothing and differentiation");
	if (scale == 0.0f) msg.debug("No smoothing before differentiation");
	else {
	    msg.debug("Gaussian smoothing at scale "+scale+" over pixel size");
	    final boolean bDebug = msg.debug();
	    msg.debug(false);
	    gauss(deriv,scale,0,0,0,false);
	    msg.debug(bDebug);
	}
	
	// Differentiate in x-dimension:
	if (xorder == 0) msg.debug("No differentiation in x-dimension");
	else if (dims.x == 1) { msg.debug("Zeroing in x-dimension"); deriv.set(0.0f); }
	else {
	    final String derivInfo = derivInfo(xorder)+" in x-dimension";
	    msg.debug(derivInfo);
	    pgs.status(derivInfo);
	    deriv.axes(Axes.X);
	    final int iD2 = xorder/2;
	    final boolean bD1 = (xorder%2 != 0);
	    final float finvhx2 = 1/(fhx*fhx);
	    final float finv2hx = 1/(2*fhx);
	    final float[] array = new float[dims.x];
	    final Coordinates coords = new Coordinates();
	    for (coords.t=0; coords.t<dims.t; ++coords.t)
		for (coords.z=0; coords.z<dims.z; ++coords.z)
		    for (coords.y=0; coords.y<dims.y; ++coords.y) {
			deriv.get(coords,array);
			for (int d2=0; d2<iD2; ++d2) D2(array,finvhx2);
			if (bD1) D1(array,finv2hx);
			deriv.set(coords,array);
			pgs.step();
		    }
	}

	// Differentiate in y-dimension:
	if (yorder == 0) msg.debug("No differentiation in y-dimension");
	else if (dims.y == 1) { msg.debug("Zeroing in y-dimension"); deriv.set(0.0f); }
	else {
	    final String derivInfo = derivInfo(yorder)+" in y-dimension";
	    msg.debug(derivInfo);
	    pgs.status(derivInfo);
	    deriv.axes(Axes.Y);
	    final int iD2 = yorder/2;
	    final boolean bD1 = (yorder%2 != 0);
	    final float finvhy2 = 1/(fhy*fhy);
	    final float finv2hy = 1/(2*fhy);
	    final float[] array = new float[dims.y];
	    final Coordinates coords = new Coordinates();
	    for (coords.t=0; coords.t<dims.t; ++coords.t)
		for (coords.z=0; coords.z<dims.z; ++coords.z)
		    for (coords.x=0; coords.x<dims.x; ++coords.x) {
			deriv.get(coords,array);
			for (int d2=0; d2<iD2; ++d2) D2(array,finvhy2);
			if (bD1) D1(array,finv2hy);
			deriv.set(coords,array);
			pgs.step();
		    }
	}

	// Differentiate in z-dimension:
	if (zorder == 0) msg.debug("No differentiation in z-dimension");
	else if (dims.z == 1) { msg.debug("Zeroing in z-dimension"); deriv.set(0.0f); }
	else {
	    final String derivInfo = derivInfo(zorder)+" in z-dimension";
	    msg.debug(derivInfo);
	    pgs.status(derivInfo);
	    deriv.axes(Axes.Z);
	    final int iD2 = zorder/2;
	    final boolean bD1 = (zorder%2 != 0);
	    final float finvhz2 = 1/(fhz*fhz);
	    final float finv2hz = 1/(2*fhz);
	    final float[] array = new float[dims.z];
	    final Coordinates coords = new Coordinates();
	    for (coords.t=0; coords.t<dims.t; ++coords.t)
		for (coords.y=0; coords.y<dims.y; ++coords.y) {
		    for (coords.x=0; coords.x<dims.x; ++coords.x) {
			deriv.get(coords,array);
			for (int d2=0; d2<iD2; ++d2) D2(array,finvhz2);
			if (bD1) D1(array,finv2hz);
			deriv.set(coords,array);
		    }
		    pgs.step(dims.z);
		}
	}

	pgs.end();
	tmr.stop(msg);

	return deriv;
    }

    private float[] kernel(final float sigma, final int deriv, final int maxhlen) {

	// Initialize:
	float fAccuracy = 5.0f;
	if (deriv == 0) fAccuracy = 3.0f;
	else if (deriv <= 3) fAccuracy = 4.0f;

	int hlen = (int)(sigma*fAccuracy) + 1;
	if (hlen > maxhlen) hlen = maxhlen;
	final float[] kernel = new float[hlen];

	// Compute kernel:
	if (sigma == 0.0f) {
	    if (deriv == 0) kernel[0] = 1.0f;
	    else kernel[0] = 0.0f;
	} else switch (deriv) {
	case 0: {
	    float integral = 0;
	    final double mi2s2 = -1/(2*sigma*sigma);
	    for (int k=0; k<hlen; ++k) {
		kernel[k] = (float)Math.exp(k*k*mi2s2);
		integral += kernel[k];
	    }
	    integral *= 2.0f; integral -= kernel[0];
	    final float invint = 1/integral;
	    for (int k=0; k<hlen; ++k) kernel[k] *= invint;
	    break;
	}
	case 1: {
	    kernel[0] = 0.0f;
	    final double mi2s2 = -1/(2*sigma*sigma);
	    final double is3sqrt2pi = 1/(Math.sqrt(2*Math.PI)*sigma*sigma*sigma);
	    for (int k=1; k<hlen; ++k)
		kernel[k] = (float)(is3sqrt2pi*k*Math.exp(k*k*mi2s2));
	    break;
	}
	case 2: {
	    final double is2 = 1/(sigma*sigma);
	    final double mi2s2 = -0.5*is2;
	    final double is3sqrt2pi = 1/(Math.sqrt(2*Math.PI)*sigma*sigma*sigma);
	    for (int k=0; k<hlen; ++k) {
		final double k2 = k*k;
		kernel[k] = (float)(is3sqrt2pi*(k2*is2 - 1)*Math.exp(k2*mi2s2));
	    }
	    break;
	}
	case 3: {
	    kernel[0] = 0.0f;
	    final double is2 = 1/(sigma*sigma);
	    final double mi2s2 = -0.5*is2;
	    final double is5sqrt2pi = 1/(Math.sqrt(2*Math.PI)*sigma*sigma*sigma*sigma*sigma);
	    for (int k=1; k<hlen; ++k) {
		final double k2 = k*k;
		kernel[k] = (float)(is5sqrt2pi*k*(k2*is2 - 3)*Math.exp(k2*mi2s2));
	    }
	    break;
	}
	case 4: {
	    final double is2 = 1/(sigma*sigma);
	    final double is4 = is2*is2;
	    final double mi2s2 = -0.5*is2;
	    final double is5sqrt2pi = 1/(Math.sqrt(2*Math.PI)*sigma*sigma*sigma*sigma*sigma);
	    for (int k=0; k<hlen; ++k) {
		final double k2 = k*k;
		kernel[k] = (float)(is5sqrt2pi*(3 - 6*k2*is2 + k2*k2*is4)*Math.exp(k2*mi2s2));
	    }
	    break;
	}
	case 5: {
	    kernel[0] = 0.0f;
	    final double is2 = 1/(sigma*sigma);
	    final double is4 = is2*is2;
	    final double mi2s2 = -0.5*is2;
	    final double is7sqrt2pi = 1/(Math.sqrt(2*Math.PI)*sigma*sigma*sigma*sigma*sigma*sigma*sigma);
	    for (int k=1; k<hlen; ++k) {
		final double k2 = k*k;
		kernel[k] = (float)(is7sqrt2pi*k*(15 - 10*k2*is2 + k2*k2*is4)*Math.exp(k2*mi2s2));
	    }
	    break;
	}
	}

	return kernel;
    }

    private String derivInfo(final int deriv) {

	String derivInfo = null;
	switch (deriv) {
	case 0: derivInfo = "Smoothing"; break;
	case 1: derivInfo = "First-order differentiation"; break;
	case 2: derivInfo = "Second-order differentiation"; break;
	case 3: derivInfo = "Third-order differentiation "; break;
	case 4: derivInfo = "Fourth-order differentiation "; break;
	case 5: derivInfo = "Fifth-order differentiation "; break;
	}
	return derivInfo;
    }

    private void convolve(final float[] ain, final float[] aout, final float[] kernel) {

	// Mirror borders in input array:
	final int khlenm1 = kernel.length - 1;
	final int aolenm1 = aout.length - 1;
	for (int k=0, lm=khlenm1, lp=khlenm1, hm=khlenm1+aolenm1, hp=khlenm1+aolenm1; k<khlenm1; ++k) {
	    ain[--lm] = ain[++lp];
	    ain[++hp] = ain[--hm];
	}

	// Convolve with kernel:
	final float leftsign = (kernel[0] == 0.0f) ? -1.0f : 1.0f;
	for (int io=0, ii=khlenm1; io<=aolenm1; ++io, ++ii) {
	    float convres = ain[ii]*kernel[0];
	    for (int k=1, iimk=ii, iipk=ii; k<=khlenm1; ++k)
		convres += (leftsign*ain[--iimk] + ain[++iipk])*kernel[k];
	    aout[io] = convres;
	}
    }

    private void D1(final float[] array, final float scale) {
	final int max = array.length - 1;
	float valim1 = array[0];
	array[0] = 0.0f;
	for (int i=1, ip1=2; i<max; ++i, ++ip1) {
	    final float d1 = scale*(array[ip1] - valim1);
	    valim1 = array[i];
	    array[i] = d1;
	}
	array[max] = 0.0f;
    }

    private void D2(final float[] array, final float scale) {
	final int max = array.length - 1;
	float valim1 = array[0];
	array[0] = 2*(array[1] - valim1);
	for (int i=1, ip1=2; i<max; ++i, ++ip1) {
	    final float d2 = scale*(valim1 - 2*array[i] + array[ip1]);
	    valim1 = array[i];
	    array[i] = d2;
	}
	array[max] = 2*(valim1 - array[max]);
    }

    // This method is currently not used. It is an implementation of a
    // recursive exponential filter. When iterating it several times
    // with pole = (1 + (iterations - sqrt(iterations^2 +
    // 2*iterations*scale^2))/scale^2), this filter approximates the
    // Gaussian smoothing filter, with the advantage that, in principle,
    // it has a scale-independent, fixed computational cost per
    // pixel. Derivatives can subsequently be computed by taking finite
    // differences. Note, however, that this is numerically very
    // inaccurate, especially at higher scales.
    private void expfilter(final float[] array, final float pole) {

	// Initialize:
	float fS0,fZi,fZ2i;
	final int iLengthm1 = array.length - 1;
	final float fS1 = array[iLengthm1];
	final float fS2 = 1/(1 - pole*pole);
	final float factor = (1 - pole)*(1 - pole);
	
	// Initialize causal filter:
	fZi = pole;
	fZ2i = (float)Math.pow(pole,iLengthm1);
	fS0 = array[0] + fZ2i*array[iLengthm1];
	fZ2i = fZ2i*fZ2i/pole;
	for (int i=1; i<iLengthm1; ++i) {
	    fS0 += (fZi + fZ2i)*array[i];
	    fZi *= pole;
	    fZ2i /= pole;
	}
	array[0] = fS0/(1 - fZi*fZi);
	
	// Apply causal filter:
	for (int i=1, im1=0; i<=iLengthm1; ++i, ++im1)
	    array[i] += pole*array[im1];
	
	// Initialize anti-causal filter:
	array[iLengthm1] = fS2*(2*array[iLengthm1] - fS1);
	
	// Apply anti-causal filter:
	for (int i=iLengthm1-1, ip1=iLengthm1; i>=0; --i, --ip1)
	    array[i] += pole*array[ip1];

	// Apply scaling factor:
	for (int i=0; i<=iLengthm1; ++i) array[i] *= factor;
    }

    private void checkArguments(final float scale, final int xorder, final int yorder, final int zorder) {

	msg.debug("Checking arguments for differentiation");

	if (scale < 0.0f)
	    throw new IllegalArgumentException("Negative smoothing scale");

	if (xorder < 0 || xorder > MAX_ORDER)
	    throw new IllegalArgumentException("Differentiation order out of range in x-dimension");

	if (yorder < 0 || yorder > MAX_ORDER)
	    throw new IllegalArgumentException("Differentiation order out of range in y-dimension");

	if (zorder < 0 || zorder > MAX_ORDER)
	    throw new IllegalArgumentException("Differentiation order out of range in z-dimension");
    }

    private final Messenger msg = Features.msg;
}
