// **********************************************************************
// COPYRIGHT NOTICE AND DISCLAIMER
//
// Copyright (c) Erik Meijering. All rights reserved.
//
// Permission to copy and use this software and accompanying
// documentation for educational, research, and not-for-profit
// purposes, without fee and without a signed licensing agreement,
// is hereby granted, provided that the above copyright notice, this
// paragraph and the following two paragraphs appear in all copies.
// The copyright holder is free to make upgraded or improved versions
// of the software available for a fee or commercially only.
//
// IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE TO ANY PARTY FOR
// DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
// OF ANY KIND WHATSOEVER, ARISING OUT OF THE USE OF THIS SOFTWARE
// AND ITS DOCUMENTATION, EVEN IF HE HAS BEEN ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//
// THE COPYRIGHT HOLDER SPECIFICALLY DISCLAIMS ANY WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
// SOFTWARE AND ACCOMPANYING DOCUMENTATION IS PROVIDED "AS IS".
// THE COPYRIGHT HOLDER HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
// SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
//
// **********************************************************************

package imagescience.features;

import imagescience.images.Axes;
import imagescience.images.ByteImage;
import imagescience.images.Coordinates;
import imagescience.images.Dimensions;
import imagescience.images.FloatImage;
import imagescience.images.PixelSizes;
import imagescience.images.ShortImage;
import imagescience.utilities.FMath;
import imagescience.utilities.Messenger;
import imagescience.utilities.Progressor;
import imagescience.utilities.Timer;

/** Provides methods for detecting edges in images. */
public final class Edges {

    /** Method for detecting edges in images of type
	<code>ByteImage</code>. See the <code>run</code> method
	for <code>FloatImage</code> objects for details on the
	parameters and exceptions.
    */
    public FloatImage run(final ByteImage image,
			  final float scale,
			  final boolean nonmaxsup,
			  final float threslow,
			  final float threshigh) {

	return run(new FloatImage(image),scale,nonmaxsup,threslow,threshigh,true);
    }
	
    /** Method for detecting edges in images of type
	<code>ShortImage</code>. See the <code>run</code> method
	for <code>FloatImage</code> objects for details on the
	parameters and exceptions.
    */
    public FloatImage run(final ShortImage image,
			  final float scale,
			  final boolean nonmaxsup,
			  final float threslow,
			  final float threshigh) {

	return run(new FloatImage(image),scale,nonmaxsup,threslow,threshigh,true);
    }
	
    /** Method for detecting edges in images of type
	<code>FloatImage</code>.

	@param image the input image in which edges need to be
	detected.<br><ul>If the size of the image in the z-dimension
	equals one, this method will compute, for every pixel, the
	magnitude of the two-dimensional (2D) gradient vector. Otherwise
	it will compute for every pixel the magnitude of the full
	three-dimensional (3D) gradient vector. These computations are
	applied to every volume in the fourth dimension.</ul>

	@param scale the smoothing scale at which the required image
	derivatives are computed.<br><ul>The scale is equal to the
	standard deviation of the Gaussian kernel used for
	differentiation. In order to enforce physical isotropy, for each
	dimension, the scale is divided by the pixel size of the image
	in that dimension. If the scale parameter is negative, gradient
	vector computation and possible suppression of non-maximum
	gradient magnitudes is skipped. This allows the user to apply
	only thresholding to an already computed gradient magnitude
	image.</ul>

	@param nonmaxsup determines whether or not locally non-maximal
	gray-values are suppressed.<br><ul>To determine whether or not
	the gray-value of a pixel is a local maximum, this method
	applies linear interpolation in the direction of the gradient
	vector to compute gray-values at approximately one-pixel
	distance on each side of the given pixel, which are subsequently
	compared to the gray-value of the given pixel. Obviously, this
	operation requires computation of the gradient vector and
	therefore can only be carried out if the scale parameter is
	non-negative.</ul>

	@param threslow <code>threshigh</code> - the lower and higher
	threshold values.<br><ul>Threshold values of zero or less are
	ignored.  That is to say, if one of these parameters is zero or
	negative, the one larger than zero is used for applying hard
	thresholding, meaning that pixels with a gray-value larger than
	or equal to that threshold are put to 255, the other pixels to
	zero. If both of these parameters are zero or negative, no
	thresholding is applied. If both are larger than zero, they are
	used together in a hysteresis fashion, meaning that all pixels
	with a gray-value larger than or equal to the second threshold
	will be put to 255, all pixels with a gray-value less than the
	first threshold will be put to zero, and pixels with a
	gray-value larger than or equal to the first threshold will be
	put to 255 only if they are connected to pixels with a
	gray-value larger than or equal to the second threshold through
	pixels all with a gray-value larger than or equal to the first
	threshold. In 2D the method uses 8-connectivity, and in 3D it
	uses 26-connectivity.</ul>

	@param useimage indicates whether or not the contents of the
	input image may be altered by the method.<br><ul>If this
	parameter is <code>true</code>, the method may use the input image
	rather than a newly created image to store intermediate results,
	thereby saving memory. If it is <code>false</code>, the input image
	is left unaltered.</ul>
	
	@return an image showing the locations of edges according to the
	algorithm. If none of the operations of gradient vector
	computation, non-maxima suppression, or thresholding is carried
	out, the method returns a handle to (a copy of) the input image.

	@exception IllegalArgumentException if the value of the lower
	threshold parameter is larger than the value of the higher
	threshold parameter, in case they are both larger than or equal
	to zero.

	@exception NullPointerException if the input image is
	<code>null</code>.
    */
    public FloatImage run(final FloatImage image,
			  final float scale,
			  final boolean nonmaxsup,
			  final float threslow,
			  final float threshigh,
			  final boolean useimage) {
	
	Features.printVersion("Edges");

	final Timer tmr = new Timer(); tmr.start();

	// Initialize:
	msg.debug("Checking arguments for edge detection");
	boolean bGradient = true;
	boolean bNonMaxSup = nonmaxsup;
	if (scale < 0.0f) bGradient = bNonMaxSup = false;
	boolean bTLow = (threslow > 0.0f) ? true : false;
	boolean bTHigh = (threshigh > 0.0f) ? true : false;
	if (bTLow && bTHigh && (threslow > threshigh))
	    throw new IllegalArgumentException("Reversed hysteresis threshold range");

	final Dimensions dims = image.dimensions();
	msg.debug("X-size of input image: " + dims.x);
	msg.debug("Y-size of input image: " + dims.y);
	msg.debug("Z-size of input image: " + dims.z);
	msg.debug("T-size of input image: " + dims.t);

	final PixelSizes ps = image.pixelsizes();
	msg.debug("Pixel size in x-dimension: " + ps.x);
	msg.debug("Pixel size in y-dimension: " + ps.y);
	msg.debug("Pixel size in z-dimension: " + ps.z);

	FloatImage edgeImage = useimage ? image : image.duplicate();

	final boolean bDebug = msg.debug();

	// Detect edges:
	if (dims.z == 1) { // 2D case
	    if (bGradient) {
		// Compute the gradient vector:
		FloatImage Ix, Iy;
		if (Features.sepSmooth()) {
		    msg.debug("Seperating smoothing and differentiation");
		    msg.debug("Gaussian smoothing at scale "+scale);
		    msg.debug(false);
		    differ.smooth(edgeImage,scale,false);
		    msg.debug(bDebug);
		    msg.debug("Computing Ix");
		    msg.debug(false);
		    Ix = differ.fdiff(edgeImage,0.0f,1,0,0,true);
		    msg.debug(bDebug);
		    msg.debug("Computing Iy");
		    msg.debug(false);
		    Iy = differ.fdiff(edgeImage,0.0f,0,1,0,false);
		    msg.debug(bDebug);
		} else {
		    msg.debug("Combining smoothing and differentiation");
		    msg.debug("Gaussian differentiation at scale "+scale);
		    msg.debug("Computing Ix");
		    msg.debug(false);
		    Ix = differ.gauss(edgeImage,scale,1,0,0,true);
		    msg.debug(bDebug);
		    msg.debug("Computing Iy");
		    msg.debug(false);
		    Iy = differ.gauss(edgeImage,scale,0,1,0,false);
		    msg.debug(bDebug);
		}

		// Compute gradient magnitude: (in order to save memory, Ix
		// is re-used in case non-maxima suppression is not applied)
		msg.debug("Computing 2D gradient magnitude image");
		edgeImage = bNonMaxSup ? new FloatImage(dims) : Ix;
		final Progressor pgs = new Progressor("Computing gradient magnitude");
		pgs.steps(dims.t*dims.z*dims.y);
		Ix.axes(Axes.X); Iy.axes(Axes.X); edgeImage.axes(Axes.X);
		final float[] aIx = new float[dims.x];
		final float[] aIy = new float[dims.x];
		final Coordinates coords = new Coordinates();
		
		for (coords.t=0; coords.t<dims.t; ++coords.t)
		    for (coords.y=0; coords.y<dims.y; ++coords.y) {
			Ix.get(coords,aIx); Iy.get(coords,aIy);
			for (int x=0; x<dims.x; ++x)
			    aIx[x] = (float)Math.sqrt(aIx[x]*aIx[x] + aIy[x]*aIy[x]);
			edgeImage.set(coords,aIx);
			pgs.step();
		    }
		pgs.end();
		
		// Apply non-maxima suppression if requested: (using
		// mirror-boundary conditions and linear interpolation)
		if (bNonMaxSup) {
		    msg.debug("Suppressing non-maximum gradient magnitudes 2D");
		    final Progressor pgss = new Progressor("Suppressing non-maxima");
		    pgss.steps(dims.t*dims.z*dims.y);
		    final FloatImage supImage = Ix;
		    final float[][] gm = new float[3][dims.x+2];
		    final Coordinates cgm = new Coordinates(); cgm.x = -1;
		    final int dimsym1 = dims.y - 1;
		    float[] atmp = null;
		    coords.reset();

		    if (dims.y == 1) {
			for (coords.t=0, cgm.t=0; coords.t<dims.t; ++coords.t, ++cgm.t) {

			    // One row only:
			    Ix.get(coords,aIx); Iy.get(coords,aIy);
			    edgeImage.get(cgm,gm[0]); gm[2] = gm[1] = gm[0];
			    suppress2D(gm,aIx,aIy);
			    supImage.set(coords,aIx);
			    pgss.step();
			}
		    } else {
			for (coords.t=0, cgm.t=0; coords.t<dims.t; ++coords.t, ++cgm.t) {
			    
			    // First row:
			    coords.y = 0; Ix.get(coords,aIx); Iy.get(coords,aIy);
			    cgm.y = 0; edgeImage.get(cgm,gm[1]);
			    cgm.y = 1; edgeImage.get(cgm,gm[0]); edgeImage.get(cgm,gm[2]);
			    suppress2D(gm,aIx,aIy);
			    supImage.set(coords,aIx);
			    pgss.step();
			    
			    // Intermediate rows:
			    for (coords.y=1, cgm.y=2; coords.y<dimsym1; ++coords.y, ++cgm.y) {
				Ix.get(coords,aIx); Iy.get(coords,aIy);
				atmp=gm[0]; gm[0]=gm[1]; gm[1]=gm[2]; gm[2]=atmp;
				edgeImage.get(cgm,gm[2]);
				suppress2D(gm,aIx,aIy);
				supImage.set(coords,aIx);
				pgss.step();
			    }
			    
			    // Last row:
			    Ix.get(coords,aIx); Iy.get(coords,aIy);
			    atmp=gm[0]; gm[0]=gm[1]; gm[1]=gm[2]; gm[2]=atmp;
			    cgm.y = dims.y-2; edgeImage.get(cgm,gm[2]);
			    suppress2D(gm,aIx,aIy);
			    supImage.set(coords,aIx);
			    pgss.step();
			}
		    }
		    
		    pgss.end();
		    edgeImage = supImage;
		}
	    }
	} else { // 3D case
	    if (bGradient) {
		// Compute the gradient vector:
		FloatImage Ix, Iy, Iz;
		if (Features.sepSmooth()) {
		    msg.debug("Seperating smoothing and differentiation");
		    msg.debug("Gaussian smoothing at scale "+scale);
		    msg.debug(false);
		    differ.smooth(edgeImage,scale,false);
		    msg.debug(bDebug);
		    msg.debug("Computing Ix");
		    msg.debug(false);
		    Ix = differ.fdiff(edgeImage,0.0f,1,0,0,true);
		    msg.debug(bDebug);
		    msg.debug("Computing Iy");
		    msg.debug(false);
		    Iy = differ.fdiff(edgeImage,0.0f,0,1,0,true);
		    msg.debug(bDebug);
		    msg.debug("Computing Iz");
		    msg.debug(false);
		    Iz = differ.fdiff(edgeImage,0.0f,0,0,1,false);
		    msg.debug(bDebug);
		} else {
		    msg.debug("Combining smoothing and differentiation");
		    msg.debug("Gaussian differentiation at scale "+scale);
		    msg.debug("Computing Ix");
		    msg.debug(false);
		    Ix = differ.gauss(edgeImage,scale,1,0,0,true);
		    msg.debug(bDebug);
		    msg.debug("Computing Iy");
		    msg.debug(false);
		    Iy = differ.gauss(edgeImage,scale,0,1,0,true);
		    msg.debug(bDebug);
		    msg.debug("Computing Iz");
		    msg.debug(false);
		    Iz = differ.gauss(edgeImage,scale,0,0,1,false);
		    msg.debug(bDebug);
		}

		// Compute gradient magnitude: (in order to save memory, Ix
		// is re-used in case non-maxima suppression is not applied)
		msg.debug("Computing 3D gradient magnitude image");
		edgeImage = bNonMaxSup ? new FloatImage(dims) : Ix;
		final Progressor pgs = new Progressor("Computing gradient magnitude");
		pgs.steps(dims.t*dims.z*dims.y);
		Ix.axes(Axes.X); Iy.axes(Axes.X); Iz.axes(Axes.X); edgeImage.axes(Axes.X);
		final float[] aIx = new float[dims.x];
		final float[] aIy = new float[dims.x];
		final float[] aIz = new float[dims.x];
		final Coordinates coords = new Coordinates();
		
		for (coords.t=0; coords.t<dims.t; ++coords.t)
		    for (coords.z=0; coords.z<dims.z; ++coords.z)
			for (coords.y=0; coords.y<dims.y; ++coords.y) {
			    Ix.get(coords,aIx); Iy.get(coords,aIy); Iz.get(coords,aIz);
			    for (int x=0; x<dims.x; ++x)
				aIx[x] = (float)Math.sqrt(aIx[x]*aIx[x] + aIy[x]*aIy[x] + aIz[x]*aIz[x]);
			    edgeImage.set(coords,aIx);
			    pgs.step();
			}
		pgs.end();
		
		// Apply non-maxima suppression if requested: (using
		// mirror-boundary conditions and linear interpolation)
		if (bNonMaxSup) {
		    msg.debug("Suppressing non-maximum gradient magnitudes 3D");
		    final Progressor pgss = new Progressor("Suppressing non-maxima");
		    pgss.steps(dims.t*dims.z);
		    Ix.axes(Axes.X+Axes.Y);
		    Iy.axes(Axes.X+Axes.Y);
		    Iz.axes(Axes.X+Axes.Y);
		    final FloatImage supImage = Ix;
		    edgeImage.axes(Axes.X+Axes.Y);
		    final float[][][] gm = new float[3][dims.y+2][dims.x+2];
		    final float[][] aaIx = new float[dims.y][dims.x];
		    final float[][] aaIy = new float[dims.y][dims.x];
		    final float[][] aaIz = new float[dims.y][dims.x];
		    final Coordinates cgm = new Coordinates(); cgm.y = cgm.x = -1;
		    final int dimszm1 = dims.z - 1;
		    float[][] atmp = null;
		    coords.reset();

		    for (coords.t=0, cgm.t=0; coords.t<dims.t; ++coords.t, ++cgm.t) {

			// First slice:
			coords.z = 0; Ix.get(coords,aaIx); Iy.get(coords,aaIy); Iz.get(coords,aaIz);
			cgm.z = 0; edgeImage.get(cgm,gm[1]);
			cgm.z = 1; edgeImage.get(cgm,gm[0]); edgeImage.get(cgm,gm[2]);
			suppress3D(gm,aaIx,aaIy,aaIz);
			supImage.set(coords,aaIx);
			pgss.step();

			// Intermediate slices:
			for (coords.z=1, cgm.z=2; coords.z<dimszm1; ++coords.z, ++cgm.z) {
			    Ix.get(coords,aaIx); Iy.get(coords,aaIy); Iz.get(coords,aaIz);
			    atmp=gm[0]; gm[0]=gm[1]; gm[1]=gm[2]; gm[2]=atmp;
			    edgeImage.get(cgm,gm[2]);
			    suppress3D(gm,aaIx,aaIy,aaIz);
			    supImage.set(coords,aaIx);
			    pgss.step();
			}

			// Last slice:
			Ix.get(coords,aaIx); Iy.get(coords,aaIy); Iz.get(coords,aaIz);
			atmp=gm[0]; gm[0]=gm[1]; gm[1]=gm[2]; gm[2]=atmp;
			cgm.z = dims.z-2; edgeImage.get(cgm,gm[2]);
			suppress3D(gm,aaIx,aaIy,aaIz);
			supImage.set(coords,aaIx);
			pgss.step();
		    }
		    
		    pgss.end();
		    edgeImage = supImage;
		}
	    }
	}

	// Apply thresholding if requested:
	if (bTLow && !bTHigh) thresholder.hard(edgeImage,threslow);
	else if (!bTLow && bTHigh) thresholder.hard(edgeImage,threshigh);
	else if (bTLow && bTHigh) thresholder.hysteresis(edgeImage,threslow,threshigh);

	tmr.stop(msg);
	return edgeImage;
    }

    private void suppress2D(final float[][] gm,
			    final float[] aIx,
			    final float[] aIy) {

	// Initialize:
	final int dimsx = aIx.length;
	final int dimsxp1 = dimsx + 1;
	final int dimsxm1 = dimsx - 1;
	float rx, ry, fx, fy, gmval, gmval1, gmval2;
	float fdx, fdy, f1mdx, f1mdy;
	int ix, iy, ixp1, iyp1;

	// Mirror x-borders:
	if (dimsx == 1) {
	    gm[0][0] = gm[0][1]; gm[0][dimsxp1] = gm[0][dimsx];
	    gm[1][0] = gm[1][1]; gm[1][dimsxp1] = gm[1][dimsx];
	    gm[2][0] = gm[2][1]; gm[2][dimsxp1] = gm[2][dimsx];
	} else {
	    gm[0][0] = gm[0][2]; gm[0][dimsxp1] = gm[0][dimsxm1];
	    gm[1][0] = gm[1][2]; gm[1][dimsxp1] = gm[1][dimsxm1];
	    gm[2][0] = gm[2][2]; gm[2][dimsxp1] = gm[2][dimsxm1];
	}

	// Suppress non-maxima:
	for (int x=0, xp1=1; x<dimsx; ++x, ++xp1) {

	    gmval = gm[1][xp1];
	    if (gmval == 0.0f) aIx[x] = 0.0f;
	    else {
		// Compute direction vector:
		rx = 0.7f*aIx[x]/gmval;
		ry = 0.7f*aIy[x]/gmval;
		
		// Compute gradient magnitude in one direction:
		fx = xp1 + rx; fy = 1 + ry;
		ix = FMath.floor(fx); iy = FMath.floor(fy);
		ixp1 = ix + 1; iyp1 = iy + 1;
		fdx = fx - ix; fdy = fy - iy;
		f1mdx = 1 - fdx; f1mdy = 1 - fdy;

		gmval1 = (f1mdy*f1mdx*gm[iy][ix] +
			  f1mdy*fdx*gm[iy][ixp1] +
			  fdy*f1mdx*gm[iyp1][ix] +
			  fdy*fdx*gm[iyp1][ixp1]);
		
		// Compute gradient magnitude in opposite direction:
		fx = xp1 - rx; fy = 1 - ry;
		ix = FMath.floor(fx); iy = FMath.floor(fy);
		ixp1 = ix + 1; iyp1 = iy + 1;
		fdx = fx - ix; fdy = fy - iy;
		f1mdx = 1 - fdx; f1mdy = 1 - fdy;

		gmval2 = (f1mdy*f1mdx*gm[iy][ix] +
			  f1mdy*fdx*gm[iy][ixp1] +
			  fdy*f1mdx*gm[iyp1][ix] +
			  fdy*fdx*gm[iyp1][ixp1]);
		
		// Suppress current gradient magnitude if non-maximum:
		if (gmval1 >= gmval || gmval2 >= gmval) aIx[x] = 0.0f;
		else aIx[x] = gmval;
	    }
	}
    }

    private void suppress3D(final float[][][] gm,
			    final float[][] aaIx,
			    final float[][] aaIy,
			    final float[][] aaIz) {

	// Initialize:
	final int dimsy = aaIx.length;
	final int dimsyp1 = dimsy + 1;
	final int dimsym1 = dimsy - 1;
	final int dimsx = aaIx[0].length;
	final int dimsxp1 = dimsx + 1;
	final int dimsxm1 = dimsx - 1;
	float rx, ry, rz, fx, fy, fz, gmval, gmval1, gmval2;
	float fdx, fdy, fdz, f1mdx, f1mdy, f1mdz;
	int ix, iy, iz, ixp1, iyp1, izp1;

	// Mirror x-borders:
	if (dimsx == 1) for (int z=0; z<3; ++z) {
	    final float[][] slice = gm[z];
	    for (int y=1; y<dimsyp1; ++y)
		{ slice[y][0] = slice[y][1]; slice[y][dimsxp1] = slice[y][dimsx]; }
	} else for (int z=0; z<3; ++z) {
	    final float[][] slice = gm[z];
	    for (int y=1; y<dimsyp1; ++y)
		{ slice[y][0] = slice[y][2]; slice[y][dimsxp1] = slice[y][dimsxm1]; }
	}
	
	// Mirror y-borders:
	if (dimsy == 1) for (int z=0; z<3; ++z) {
	    final float[] y0 = gm[z][0];
	    final float[] y1 = gm[z][1];
	    final float[] y2 = gm[z][2];
	    for (int x=0; x<=dimsxp1; ++x) y0[x] = y2[x] = y1[x];
	} else for (int z=0; z<3; ++z) {
	    final float[] y0 = gm[z][0];
	    final float[] y2 = gm[z][2];
	    final float[] ydimsym1 = gm[z][dimsym1];
	    final float[] ydimsyp1 = gm[z][dimsyp1];
	    for (int x=0; x<=dimsxp1; ++x) { y0[x] = y2[x]; ydimsyp1[x] = ydimsym1[x]; }
	}

	// Suppress non-maxima:
	final float[][] gm1 = gm[1];

	for (int y=0, yp1=1; y<dimsy; ++y, ++yp1) {

	    final float[] gm1yp1 = gm1[yp1];
	    final float[] aIx = aaIx[y];
	    final float[] aIy = aaIy[y];
	    final float[] aIz = aaIz[y];

	    for (int x=0, xp1=1; x<dimsx; ++x, ++xp1) {
		
		gmval = gm1yp1[xp1];
		if (gmval == 0.0f) aIx[x] = 0.0f;
		else {
		    // Compute direction vector:
		    rx = 0.7f*aIx[x]/gmval;
		    ry = 0.7f*aIy[x]/gmval;
		    rz = 0.7f*aIz[x]/gmval;
		    
		    // Compute gradient magnitude in one direction:
		    fx = xp1 + rx; fy = yp1 + ry; fz = 1 + rz;
		    ix = FMath.floor(fx); iy = FMath.floor(fy); iz = FMath.floor(fz);
		    ixp1 = ix + 1; iyp1 = iy + 1; izp1 = iz + 1;
		    fdx = fx - ix; fdy = fy - iy; fdz = fz - iz;
		    f1mdx = 1 - fdx; f1mdy = 1 - fdy; f1mdz = 1 - fdz;
		    
		    gmval1 = (f1mdz*f1mdy*f1mdx*gm[iz][iy][ix] +
			      f1mdz*f1mdy*fdx*gm[iz][iy][ixp1] +
			      f1mdz*fdy*f1mdx*gm[iz][iyp1][ix] +
			      f1mdz*fdy*fdx*gm[iz][iyp1][ixp1] +
			      fdz*f1mdy*f1mdx*gm[izp1][iy][ix] +
			      fdz*f1mdy*fdx*gm[izp1][iy][ixp1] +
			      fdz*fdy*f1mdx*gm[izp1][iyp1][ix] +
			      fdz*fdy*fdx*gm[izp1][iyp1][ixp1]);
		    
		    // Compute gradient magnitude in opposite direction:
		    fx = xp1 - rx; fy = yp1 - ry; fz = 1 - rz;
		    ix = FMath.floor(fx); iy = FMath.floor(fy); iz = FMath.floor(fz);
		    ixp1 = ix + 1; iyp1 = iy + 1; izp1 = iz + 1;
		    fdx = fx - ix; fdy = fy - iy; fdz = fz - iz;
		    f1mdx = 1 - fdx; f1mdy = 1 - fdy; f1mdz = 1 - fdz;
		    
		    gmval2 = (f1mdz*f1mdy*f1mdx*gm[iz][iy][ix] +
			      f1mdz*f1mdy*fdx*gm[iz][iy][ixp1] +
			      f1mdz*fdy*f1mdx*gm[iz][iyp1][ix] +
			      f1mdz*fdy*fdx*gm[iz][iyp1][ixp1] +
			      fdz*f1mdy*f1mdx*gm[izp1][iy][ix] +
			      fdz*f1mdy*fdx*gm[izp1][iy][ixp1] +
			      fdz*fdy*f1mdx*gm[izp1][iyp1][ix] +
			      fdz*fdy*fdx*gm[izp1][iyp1][ixp1]);
		    
		    // Suppress current gradient magnitude if non-maximum:
		    if (gmval1 >= gmval || gmval2 >= gmval) aIx[x] = 0.0f;
		    else aIx[x] = gmval;
		}
	    }
	}
    }

    private final Messenger msg = Features.msg;
    private final Differentiator differ = new Differentiator();
    private final Thresholder thresholder = new Thresholder();
}
