// **********************************************************************
// COPYRIGHT NOTICE AND DISCLAIMER
//
// Copyright (c) Erik Meijering. All rights reserved.
//
// Permission to copy and use this software and accompanying
// documentation for educational, research, and not-for-profit
// purposes, without fee and without a signed licensing agreement,
// is hereby granted, provided that the above copyright notice, this
// paragraph and the following two paragraphs appear in all copies.
// The copyright holder is free to make upgraded or improved versions
// of the software available for a fee or commercially only.
//
// IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE TO ANY PARTY FOR
// DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
// OF ANY KIND WHATSOEVER, ARISING OUT OF THE USE OF THIS SOFTWARE
// AND ITS DOCUMENTATION, EVEN IF HE HAS BEEN ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//
// THE COPYRIGHT HOLDER SPECIFICALLY DISCLAIMS ANY WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
// SOFTWARE AND ACCOMPANYING DOCUMENTATION IS PROVIDED "AS IS".
// THE COPYRIGHT HOLDER HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
// SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
//
// **********************************************************************

package imagescience.features;

import imagescience.images.Axes;
import imagescience.images.ByteImage;
import imagescience.images.Coordinates;
import imagescience.images.Dimensions;
import imagescience.images.FloatImage;
import imagescience.images.PixelSizes;
import imagescience.images.ShortImage;
import imagescience.utilities.Messenger;
import imagescience.utilities.Progressor;
import imagescience.utilities.Timer;

/** Provides methods for computing eigenimages of the
    structure tensor. */
public final class Structure {

    /** Method for computing structure tensor eigenimages of images of
	type <code>ByteImage</code>. See the <code>run</code>
	method for <code>FloatImage</code> objects for details on the
	parameters and exceptions.
    */
    public FloatImage[] run(final ByteImage image,
			    final float sscale,
			    final float iscale) {

	return run(new FloatImage(image),sscale,iscale,true);
    }

    /** Method for computing structure tensor eigenimages of images of
	type <code>ShortImage</code>. See the <code>run</code>
	method for <code>FloatImage</code> objects for details on the
	parameters and exceptions.
    */
    public FloatImage[] run(final ShortImage image,
			    final float sscale,
			    final float iscale) {

	return run(new FloatImage(image),sscale,iscale,true);
    }

    /** Method for computing structure tensor eigenimages of images of
	type <code>FloatImage</code>.

	@param image the input image for which eigenimages need to be
	computed.<br><ul>If the size of the image in the z-dimension
	equals one, this method will compute, for every pixel, the
	two-dimensional (2D) structure tensor and its two
	eigenvalues. Otherwise it will compute for every pixel the full
	three-dimensional (3D) structure tensor and its three
	eigenvalues. These computations are applied to every volume in
	the fourth dimension.</ul>

	@param sscale the smoothing scale at which the required image
	derivatives are computed.<br><ul> This scale is equal to the
	standard deviation of the Gaussian kernel used for
	differentiation and must be larger than zero. In order to
	enforce physical isotropy, for each dimension, the scale is
	divided by the pixel size of the image in that dimension.</ul>

	@param iscale the integration scale.<br><ul> This scale is equal
	to the standard deviation of the Gaussian kernel used for
	integrating the components of the structure tensor and must be
	larger than zero. In order to enforce physical isotropy, for
	each dimension, the scale is divided by the pixel size of the
	image in that dimension.</ul>

	@param useimage indicates whether or not the contents of the
	input image may be altered by the method.<br><ul>If this
	parameter is <code>true</code>, the method may use the input image
	rather than a newly created image to store intermediate results,
	thereby saving memory. If it is <code>false</code>, the input image
	is left unaltered.</ul>
	
	@return an array containing the eigenimages.<br><ul>If only the
	two-dimensional (2D) structure tensor and its two eigenvalues
	were computed for every pixel, the returned array contains two
	eigenimages: at index 0 the image with, for every pixel, the
	largest eigenvalue; at index 1 the image with, for every pixel,
	the smallest eigenvalue. If the full three-dimensional (3D)
	structure tensor and its three eigenvalues were computed for
	every pixel, the returned array contains three eigenimages: at
	index 0 the image with, for every pixel, the largest eigenvalue;
	at index 1 the image with, for every pixel, the middle
	eigenvalue; at index 2 the image with, for every pixel, the
	smallest eigenvalue.</ul>

	@exception IllegalArgumentException if the scale is negative.

	@exception NullPointerException if the input image is
	<code>null</code>.
    */
    public FloatImage[] run(final FloatImage image,
			    final float sscale,
			    final float iscale,
			    final boolean useimage) {

	Features.printVersion("Structure");

	final Timer tmr = new Timer(); tmr.start();

	// Initialize:
	msg.debug("Checking arguments for structure computation");
	if (sscale < 0.0f) throw new IllegalArgumentException("Negative smoothing scale");
	if (iscale < 0.0f) throw new IllegalArgumentException("Negative integration scale");

	final Dimensions dims = image.dimensions();
	msg.debug("X-size of input image: " + dims.x);
	msg.debug("Y-size of input image: " + dims.y);
	msg.debug("Z-size of input image: " + dims.z);
	msg.debug("T-size of input image: " + dims.t);

	final PixelSizes ps = image.pixelsizes();
	msg.debug("Pixel size in x-dimension: " + ps.x);
	msg.debug("Pixel size in y-dimension: " + ps.y);
	msg.debug("Pixel size in z-dimension: " + ps.z);

	FloatImage[] eigenimages = null;

	// Compute structure tensor (2D or 3D) and eigenimages:
	if (dims.z == 1) { // 2D case

	    // Compute the components of the structure tensor:
	    final boolean bDebug = msg.debug();
	    FloatImage Ix2, Iy2;
	    if (Features.sepSmooth()) {
		msg.debug("Seperating smoothing and differentiation");
		msg.debug("Gaussian smoothing at scale "+sscale);
		msg.debug(false);
		final FloatImage smoothImage = useimage ? image : image.duplicate();
		differ.smooth(smoothImage,sscale,false);
		msg.debug(bDebug);
		msg.debug("Computing Ix");
		msg.debug(false);
		Ix2 = differ.fdiff(smoothImage,0.0f,1,0,0,true);
		msg.debug(bDebug);
		msg.debug("Computing Iy");
		msg.debug(false);
		Iy2 = differ.fdiff(smoothImage,0.0f,0,1,0,false);
		msg.debug(bDebug);
	    } else {
		msg.debug("Combining smoothing and differentiation");
		msg.debug("Gaussian differentiation at scale "+sscale);
		msg.debug("Computing Ix");
		msg.debug(false);
		Ix2 = differ.gauss(image,sscale,1,0,0,true);
		msg.debug(bDebug);
		msg.debug("Computing Iy");
		msg.debug(false);
		Iy2 = differ.gauss(image,sscale,0,1,0,useimage?false:true);
		msg.debug(bDebug);
	    }
	    msg.debug("Computing IxIy");
	    final FloatImage IxIy = Ix2.duplicate();
	    IxIy.multiply(Iy2);
	    msg.debug("Squaring Ix");
	    Ix2.square();
	    msg.debug("Squaring Iy");
	    Iy2.square();

	    // Integrate the components of the tensor:
	    if (iscale > 0.0f) {
		msg.debug("Gaussian integration at scale "+iscale);
		msg.debug("Integrating IxIx");
		msg.debug(false);
		differ.smooth(Ix2,iscale,false);
		msg.debug(bDebug);
		msg.debug("Integrating IxIy");
		msg.debug(false);
		differ.smooth(IxIy,iscale,false);
		msg.debug(bDebug);
		msg.debug("Integrating IyIy");
		msg.debug(false);
		differ.smooth(Iy2,iscale,false);
		msg.debug(bDebug);
	    } else msg.debug("No integration");

	    // Compute the eigenimages: (note that in order to minimize
	    // memory usage, Ix2 and Iy2 are re-used)
	    msg.debug("Computing 2D eigenimages");
	    final Progressor pgs = new Progressor("Computing eigenimages");
	    pgs.steps(dims.t*dims.z*dims.y);
	    Ix2.axes(Axes.X); IxIy.axes(Axes.X); Iy2.axes(Axes.X);
	    final float[] axx = new float[dims.x];
	    final float[] axy = new float[dims.x];
	    final float[] ayy = new float[dims.x];
	    final Coordinates coords = new Coordinates();

	    msg.debug("Comparing and storing eigenvalues");
	    for (coords.t=0; coords.t<dims.t; ++coords.t)
		for (coords.y=0; coords.y<dims.y; ++coords.y) {
		    Ix2.get(coords,axx);
		    IxIy.get(coords,axy);
		    Iy2.get(coords,ayy);
		    for (int x=0; x<dims.x; ++x) {
			final float b = -(axx[x] + ayy[x]);
			final float c = axx[x]*ayy[x] - axy[x]*axy[x];
			final float q = (float)(-0.5*(b + (b < 0.0f ? -1.0f : 1.0f)*Math.sqrt(b*b - 4*c)));
			final float absh1 = Math.abs(q);
			final float absh2 = Math.abs(c/q);
			if (absh1 > absh2) { axx[x] = absh1; ayy[x] = absh2; }
			else { axx[x] = absh2; ayy[x] = absh1; }
		    }
		    Ix2.set(coords,axx);
		    Iy2.set(coords,ayy);
		    pgs.step();
		}
	    pgs.end();

	    eigenimages = new FloatImage[2];
	    eigenimages[0] = Ix2;
	    eigenimages[1] = Iy2;

	} else { // 3D case

	    // Compute the components of the structure tensor:
	    final boolean bDebug = msg.debug();
	    FloatImage Ix2, Iy2, Iz2;
	    if (Features.sepSmooth()) {
		msg.debug("Seperating smoothing and differentiation");
		msg.debug("Gaussian smoothing at scale "+sscale);
		msg.debug(false);
		final FloatImage smoothImage = useimage ? image : image.duplicate();
		differ.smooth(smoothImage,sscale,false);
		msg.debug(bDebug);
		msg.debug("Computing Ix");
		msg.debug(false);
		Ix2 = differ.fdiff(smoothImage,0.0f,1,0,0,true);
		msg.debug(bDebug);
		msg.debug("Computing Iy");
		msg.debug(false);
		Iy2 = differ.fdiff(smoothImage,0.0f,0,1,0,true);
		msg.debug(bDebug);
		msg.debug("Computing Iz");
		msg.debug(false);
		Iz2 = differ.fdiff(smoothImage,0.0f,0,0,1,false);
		msg.debug(bDebug);
	    } else {
		msg.debug("Combining smoothing and differentiation");
		msg.debug("Gaussian differentiation at scale "+sscale);
		msg.debug("Computing Ix");
		msg.debug(false);
		Ix2 = differ.gauss(image,sscale,1,0,0,true);
		msg.debug(bDebug);
		msg.debug("Computing Iy");
		msg.debug(false);
		Iy2 = differ.gauss(image,sscale,0,1,0,true);
		msg.debug(bDebug);
		msg.debug("Computing Iz");
		msg.debug(false);
		Iz2 = differ.gauss(image,sscale,0,0,1,useimage?false:true);
		msg.debug(bDebug);
	    }
	    msg.debug("Computing IxIy");
	    final FloatImage IxIy = Ix2.duplicate();
	    IxIy.multiply(Iy2);
	    msg.debug("Computing IxIz");
	    final FloatImage IxIz = Ix2.duplicate();
	    IxIz.multiply(Iz2);
	    msg.debug("Computing IyIz");
	    final FloatImage IyIz = Iy2.duplicate();
	    IyIz.multiply(Iz2);
	    msg.debug("Squaring Ix");
	    Ix2.square();
	    msg.debug("Squaring Iy");
	    Iy2.square();
	    msg.debug("Squaring Iz");
	    Iz2.square();

	    // Integrate the components of the tensor:
	    if (iscale > 0.0f) {
		msg.debug("Gaussian integration at scale "+iscale);
		msg.debug("Integrating IxIx");
		msg.debug(false);
		differ.smooth(Ix2,iscale,false);
		msg.debug(bDebug);
		msg.debug("Integrating IxIy");
		msg.debug(false);
		differ.smooth(IxIy,iscale,false);
		msg.debug(bDebug);
		msg.debug("Integrating IxIz");
		msg.debug(false);
		differ.smooth(IxIz,iscale,false);
		msg.debug(bDebug);
		msg.debug("Integrating IyIy");
		msg.debug(false);
		differ.smooth(Iy2,iscale,false);
		msg.debug(bDebug);
		msg.debug("Integrating IyIz");
		msg.debug(false);
		differ.smooth(IyIz,iscale,false);
		msg.debug(bDebug);
		msg.debug("Integrating IzIz");
		msg.debug(false);
		differ.smooth(Iz2,iscale,false);
		msg.debug(bDebug);
	    } else msg.debug("No integration");

	    // Compute the eigenimages: (note that in order to minimize
	    // memory usage, Ix2, Iy2, and Iz2 are re-used)
	    msg.debug("Computing 3D eigenimages");
	    final Progressor pgs = new Progressor("Computing eigenimages");
	    pgs.steps(dims.t*dims.z*dims.y);
	    Ix2.axes(Axes.X); IxIy.axes(Axes.X); IxIz.axes(Axes.X);
	    Iy2.axes(Axes.X); IyIz.axes(Axes.X); Iz2.axes(Axes.X);
	    final float[] axx = new float[dims.x];
	    final float[] axy = new float[dims.x];
	    final float[] axz = new float[dims.x];
	    final float[] ayy = new float[dims.x];
	    final float[] ayz = new float[dims.x];
	    final float[] azz = new float[dims.x];
	    final Coordinates coords = new Coordinates();

	    msg.debug("Comparing and storing eigenvalues");
	    for (coords.t=0; coords.t<dims.t; ++coords.t)
		for (coords.z=0; coords.z<dims.z; ++coords.z)
		    for (coords.y=0; coords.y<dims.y; ++coords.y) {
			Ix2.get(coords,axx);
			IxIy.get(coords,axy);
			IxIz.get(coords,axz);
			Iy2.get(coords,ayy);
			IyIz.get(coords,ayz);
			Iz2.get(coords,azz);
			for (int x=0; x<dims.x; ++x) {
			    final float fxx = axx[x];
			    final float fxy = axy[x];
			    final float fxz = axz[x];
			    final float fyy = ayy[x];
			    final float fyz = ayz[x];
			    final float fzz = azz[x];
			    final float a = -(fxx + fyy + fzz);
			    final float b = fxx*fyy + fxx*fzz + fyy*fzz - fxy*fxy - fxz*fxz - fyz*fyz;
			    final float c = fxx*(fyz*fyz - fyy*fzz) + fyy*fxz*fxz + fzz*fxy*fxy - 2*fxy*fxz*fyz;
			    final double q = (a*a - 3*b)/9;
			    final double r = (a*a*a - 4.5f*a*b + 13.5f*c)/27f;
			    final double sqrtq = (q > 0) ? Math.sqrt(q) : 0.0;
			    final double rsqq3 = r/(sqrtq*sqrtq*sqrtq);
			    final double angle = (rsqq3*rsqq3 <= 1) ? Math.acos(rsqq3) : Math.acos(rsqq3 < 0.0f ? -1.0f : 1.0f);
			    float absh1 = (float)Math.abs(-2.0*sqrtq*Math.cos(angle/3) - a/3);
			    float absh2 = (float)Math.abs(-2.0*sqrtq*Math.cos((angle + TWOPI)/3) - a/3);
			    float absh3 = (float)Math.abs(-2.0*sqrtq*Math.cos((angle - TWOPI)/3) - a/3);
			    if (absh2 < absh3) { final float tmp = absh2; absh2 = absh3; absh3 = tmp; }
			    if (absh1 < absh2) { final float tmp1 = absh1; absh1 = absh2; absh2 = tmp1;
			    if (absh2 < absh3) { final float tmp2 = absh2; absh2 = absh3; absh3 = tmp2; }}
			    axx[x] = absh1;
			    ayy[x] = absh2;
			    azz[x] = absh3;
			}
			Ix2.set(coords,axx);
			Iy2.set(coords,ayy);
			Iz2.set(coords,azz);
			pgs.step();
		    }
	    pgs.end();

	    eigenimages = new FloatImage[3];
	    eigenimages[0] = Ix2;
	    eigenimages[1] = Iy2;
	    eigenimages[2] = Iz2;
	}

	tmr.stop(msg);
	return eigenimages;
    }

    private final Messenger msg = Features.msg;
    private final Differentiator differ = new Differentiator();

    private final double TWOPI = 2*Math.PI;
}
