// **********************************************************************
// COPYRIGHT NOTICE AND DISCLAIMER
//
// Copyright (c) Erik Meijering. All rights reserved.
//
// Permission to copy and use this software and accompanying
// documentation for educational, research, and not-for-profit
// purposes, without fee and without a signed licensing agreement,
// is hereby granted, provided that the above copyright notice, this
// paragraph and the following two paragraphs appear in all copies.
// The copyright holder is free to make upgraded or improved versions
// of the software available for a fee or commercially only.
//
// IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE TO ANY PARTY FOR
// DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
// OF ANY KIND WHATSOEVER, ARISING OUT OF THE USE OF THIS SOFTWARE
// AND ITS DOCUMENTATION, EVEN IF HE HAS BEEN ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//
// THE COPYRIGHT HOLDER SPECIFICALLY DISCLAIMS ANY WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
// SOFTWARE AND ACCOMPANYING DOCUMENTATION IS PROVIDED "AS IS".
// THE COPYRIGHT HOLDER HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
// SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
//
// **********************************************************************

package imagescience.features;

import imagescience.images.Axes;
import imagescience.images.ByteImage;
import imagescience.images.Coordinates;
import imagescience.images.Dimensions;
import imagescience.images.FloatImage;
import imagescience.utilities.Messenger;
import imagescience.utilities.Progressor;

/** Provides methods for thresholding images. */
public final class Thresholder {

    /** Method for applying hard thresholding to images of type
	<code>FloatImage</code>.

	@param image the input image to be thresholded.

	@param threshold the threshold value.<br><ul>Pixels with a
	gray-value larger than or equal to the value of this parameter
	are put to 255 and pixels with a gray-value less than this
	parameter are put to zero.</ul>

	@exception NullPointerException if the input image is
	<code>null</code>.
    */
    public void hard(final FloatImage image,
		     final float threshold) {
	
	// Initialize:
	msg.debug("Applying hard thresholding at value "+threshold);
	final Dimensions dims = image.dimensions();
	final Progressor pgs = new Progressor("Applying hard thresholding");
	pgs.steps(dims.t*dims.z);
	image.axes(Axes.X);
	final float[] array = new float[dims.x];
	final Coordinates coords = new Coordinates();
		
	// Threshold:
	for (coords.t=0; coords.t<dims.t; ++coords.t)
	    for (coords.z=0; coords.z<dims.z; ++coords.z) {
		for (coords.y=0; coords.y<dims.y; ++coords.y) {
		    image.get(coords,array);
		    for (int x=0; x<dims.x; ++x)
			if (array[x] >= threshold) array[x] = fON;
			else array[x] = fOFF;
		    image.set(coords,array);
		}
		pgs.step();
	    }
	pgs.end();
    }

    /** Method for applying hysteresis thresholding to images of type
	<code>FloatImage</code>.

	@param image the input image to be thresholded.

	@param threslow <code>threshigh</code> - the lower and higher
	threshold values.<br><ul>The parameters determine the range of
	the hysteresis. All pixels with a gray-value larger than or
	equal to the second threshold will be put to 255, all pixels
	with a gray-value less than the first threshold will be put to
	zero, and pixels with a gray-value larger than or equal to the
	first threshold will be put to 255 only if they are connected to
	pixels with a gray-value larger than or equal to the second
	threshold through pixels all with a gray-value larger than or
	equal to the first threshold. In 2D the method uses
	8-connectivity, and in 3D it uses 26-connectivity.</ul>

	@exception NullPointerException if the input image is
	<code>null</code>.
    */
    public void hysteresis(final FloatImage image,
			   final float threslow,
			   final float threshigh) {
	
	// Initialize:
	msg.debug("Applying hysteresis thresholding in range ["+threslow+","+threshigh+"]");
	final Dimensions dims = image.dimensions();
	final Progressor pgs = new Progressor("Applying hysteresis thresholding");
	pgs.steps(dims.t*dims.z*dims.y);
	image.axes(Axes.X);
	final float[] array = new float[dims.x];
	final byte[] marray = new byte[dims.x+2];
	final Coordinates coords = new Coordinates();
	final Coordinates mcoords = new Coordinates();
	final Coordinates cand = new Coordinates();
	final Candidates cands = new Candidates();

	// Classify and expand:
	if (dims.z == 1) { // 2D case

	    // Create map with border:
	    final Dimensions mdims = new Dimensions(dims.x+2,dims.y+2,dims.z,dims.t);
	    final ByteImage map = new ByteImage(mdims); map.axes(Axes.X);
	    for (coords.t=mcoords.t=0; coords.t<dims.t; ++coords.t, ++mcoords.t)
		for (coords.y=0, mcoords.y=1; coords.y<dims.y; ++coords.y, ++mcoords.y) {
		    coords.x=0; image.get(coords,array);
		    for (mcoords.x=1; coords.x<dims.x; ++coords.x, ++mcoords.x)
			if (array[coords.x] >= threshigh) marray[mcoords.x] = ON;
			else if (array[coords.x] >= threslow) marray[mcoords.x] = CAND;
			else marray[mcoords.x] = OFF;
		    mcoords.x=0; map.set(mcoords,marray);
		}
	    
	    // Expand ON pixels:
	    for (mcoords.t=0; mcoords.t<dims.t; ++mcoords.t)
		for (mcoords.y=1; mcoords.y<=dims.y; ++mcoords.y) {
		    mcoords.x=0; map.get(mcoords,marray);
		    for (mcoords.x=1; mcoords.x<=dims.x; ++mcoords.x)
			if (marray[mcoords.x] == ON) {
			    cands.push(mcoords);
			    // Search for candidates iteratively in 8-connected neighborhoods:
			    while (cands.size > 0) {
				cands.pop(cand); map.set(cand,ON);
				++cand.x; if (map.get(cand)==CAND) { cands.push(cand); map.set(cand,QUEUED); }
				++cand.y; if (map.get(cand)==CAND) { cands.push(cand); map.set(cand,QUEUED); }
				--cand.x; if (map.get(cand)==CAND) { cands.push(cand); map.set(cand,QUEUED); }
				--cand.x; if (map.get(cand)==CAND) { cands.push(cand); map.set(cand,QUEUED); }
				--cand.y; if (map.get(cand)==CAND) { cands.push(cand); map.set(cand,QUEUED); }
				--cand.y; if (map.get(cand)==CAND) { cands.push(cand); map.set(cand,QUEUED); }
				++cand.x; if (map.get(cand)==CAND) { cands.push(cand); map.set(cand,QUEUED); }
				++cand.x; if (map.get(cand)==CAND) { cands.push(cand); map.set(cand,QUEUED); }
			    }
			}
		    pgs.step();
		}

	    // Copy ON pixels:
	    for (coords.t=mcoords.t=0; coords.t<dims.t; ++coords.t, ++mcoords.t)
		for (coords.y=0, mcoords.y=1; coords.y<dims.y; ++coords.y, ++mcoords.y) {
		    mcoords.x=0; map.get(mcoords,marray);
		    for (coords.x=0, mcoords.x=1; coords.x<dims.x; ++coords.x, ++mcoords.x)
			if (marray[mcoords.x] == ON) array[coords.x] = fON;
			else array[coords.x] = fOFF;
		    coords.x=0; image.set(coords,array);
		}
	    
	} else { // 3D case

	    // Create map with border:
	    final Dimensions mdims = new Dimensions(dims.x+2,dims.y+2,dims.z+2,dims.t);
	    final ByteImage map = new ByteImage(mdims); map.axes(Axes.X);
	    for (coords.t=mcoords.t=0; coords.t<dims.t; ++coords.t, ++mcoords.t)
		for (coords.z=0, mcoords.z=1; coords.z<dims.z; ++coords.z, ++mcoords.z)
		    for (coords.y=0, mcoords.y=1; coords.y<dims.y; ++coords.y, ++mcoords.y) {
			coords.x=0; image.get(coords,array);
			for (mcoords.x=1; coords.x<dims.x; ++coords.x, ++mcoords.x)
			    if (array[coords.x] >= threshigh) marray[mcoords.x] = ON;
			    else if (array[coords.x] >= threslow) marray[mcoords.x] = CAND;
			    else marray[mcoords.x] = OFF;
			mcoords.x=0; map.set(mcoords,marray);
		    }
	    
	    // Expand ON voxels:
	    for (mcoords.t=0; mcoords.t<dims.t; ++mcoords.t)
		for (mcoords.z=1; mcoords.z<=dims.z; ++mcoords.z)
		    for (mcoords.y=1; mcoords.y<=dims.y; ++mcoords.y) {
			mcoords.x=0; map.get(mcoords,marray);
			for (mcoords.x=1; mcoords.x<=dims.x; ++mcoords.x)
			    if (marray[mcoords.x] == ON) {
				cands.push(mcoords);
				// Search for candidates iteratively in 26-connected neighborhoods:
				while (cands.size > 0) {
				    cands.pop(cand); map.set(cand,ON);
				    // Back 9 voxels:
				    ++cand.z; if (map.get(cand)==CAND) { cands.push(cand); map.set(cand,QUEUED); }
				    ++cand.x; if (map.get(cand)==CAND) { cands.push(cand); map.set(cand,QUEUED); }
				    ++cand.y; if (map.get(cand)==CAND) { cands.push(cand); map.set(cand,QUEUED); }
				    --cand.x; if (map.get(cand)==CAND) { cands.push(cand); map.set(cand,QUEUED); }
				    --cand.x; if (map.get(cand)==CAND) { cands.push(cand); map.set(cand,QUEUED); }
				    --cand.y; if (map.get(cand)==CAND) { cands.push(cand); map.set(cand,QUEUED); }
				    --cand.y; if (map.get(cand)==CAND) { cands.push(cand); map.set(cand,QUEUED); }
				    ++cand.x; if (map.get(cand)==CAND) { cands.push(cand); map.set(cand,QUEUED); }
				    ++cand.x; if (map.get(cand)==CAND) { cands.push(cand); map.set(cand,QUEUED); }
				    // Middle 8 voxels:
				    --cand.z; if (map.get(cand)==CAND) { cands.push(cand); map.set(cand,QUEUED); }
				    ++cand.y; if (map.get(cand)==CAND) { cands.push(cand); map.set(cand,QUEUED); }
				    ++cand.y; if (map.get(cand)==CAND) { cands.push(cand); map.set(cand,QUEUED); }
				    --cand.x; if (map.get(cand)==CAND) { cands.push(cand); map.set(cand,QUEUED); }
				    --cand.x; if (map.get(cand)==CAND) { cands.push(cand); map.set(cand,QUEUED); }
				    --cand.y; if (map.get(cand)==CAND) { cands.push(cand); map.set(cand,QUEUED); }
				    --cand.y; if (map.get(cand)==CAND) { cands.push(cand); map.set(cand,QUEUED); }
				    ++cand.x; if (map.get(cand)==CAND) { cands.push(cand); map.set(cand,QUEUED); }
				    // Front 9 voxels:
				    --cand.z; if (map.get(cand)==CAND) { cands.push(cand); map.set(cand,QUEUED); }
				    ++cand.x; if (map.get(cand)==CAND) { cands.push(cand); map.set(cand,QUEUED); }
				    ++cand.y; if (map.get(cand)==CAND) { cands.push(cand); map.set(cand,QUEUED); }
				    ++cand.y; if (map.get(cand)==CAND) { cands.push(cand); map.set(cand,QUEUED); }
				    --cand.x; if (map.get(cand)==CAND) { cands.push(cand); map.set(cand,QUEUED); }
				    --cand.x; if (map.get(cand)==CAND) { cands.push(cand); map.set(cand,QUEUED); }
				    --cand.y; if (map.get(cand)==CAND) { cands.push(cand); map.set(cand,QUEUED); }
				    --cand.y; if (map.get(cand)==CAND) { cands.push(cand); map.set(cand,QUEUED); }
				    ++cand.x; ++cand.y; if (map.get(cand)==CAND) { cands.push(cand); map.set(cand,QUEUED); }
				}
			    }
			pgs.step();
		    }

	    // Copy ON voxels:
	    for (coords.t=mcoords.t=0; coords.t<dims.t; ++coords.t, ++mcoords.t)
		for (coords.z=0, mcoords.z=1; coords.z<dims.z; ++coords.z, ++mcoords.z)
		    for (coords.y=0, mcoords.y=1; coords.y<dims.y; ++coords.y, ++mcoords.y) {
			mcoords.x=0; map.get(mcoords,marray);
			for (coords.x=0, mcoords.x=1; coords.x<dims.x; ++coords.x, ++mcoords.x)
			    if (marray[mcoords.x] == ON) array[coords.x] = fON;
			    else array[coords.x] = fOFF;
			coords.x=0; image.set(coords,array);
		    }
	}

	pgs.end();
    }
    
    private final Messenger msg = Features.msg;

    private final static byte ON = 10;
    private final static byte OFF = 0;
    private final static byte CAND = 2;
    private final static byte QUEUED = 9;

    private final static float fON = 255.0f;
    private final static float fOFF = 0.0f;
}

final class Candidates {

    private int iCap = 5000;
    private final static int CAPINC = 5000;
    int size = 0;

    private short[] cx = new short[iCap];
    private short[] cy = new short[iCap];
    private short[] cz = new short[iCap];
    private short[] ct = new short[iCap];

    void push(final Coordinates coords) {
	if (size == iCap) inccap();
	cx[size] = (short)coords.x;
	cy[size] = (short)coords.y;
	cz[size] = (short)coords.z;
	ct[size] = (short)coords.t;
	++size;
    }
	
    void pop(final Coordinates coords) {
	--size;
	coords.x = cx[size];
	coords.y = cy[size];
	coords.z = cz[size];
	coords.t = ct[size];
    }

    private void inccap() {
	final int iNewCap = iCap + CAPINC;
	final short[] newcx = new short[iNewCap];
	final short[] newcy = new short[iNewCap];
	final short[] newcz = new short[iNewCap];
	final short[] newct = new short[iNewCap];
	for (int i=0; i<iCap; ++i) {
	    newcx[i] = cx[i];
	    newcy[i] = cy[i];
	    newcz[i] = cz[i];
	    newct[i] = ct[i];
	}
	cx = newcx;
	cy = newcy;
	cz = newcz;
	ct = newct;
	iCap = iNewCap;
    }
}
