// **********************************************************************
// COPYRIGHT NOTICE AND DISCLAIMER
//
// Copyright (c) Erik Meijering. All rights reserved.
//
// Permission to copy and use this software and accompanying
// documentation for educational, research, and not-for-profit
// purposes, without fee and without a signed licensing agreement,
// is hereby granted, provided that the above copyright notice, this
// paragraph and the following two paragraphs appear in all copies.
// The copyright holder is free to make upgraded or improved versions
// of the software available for a fee or commercially only.
//
// IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE TO ANY PARTY FOR
// DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
// OF ANY KIND WHATSOEVER, ARISING OUT OF THE USE OF THIS SOFTWARE
// AND ITS DOCUMENTATION, EVEN IF HE HAS BEEN ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//
// THE COPYRIGHT HOLDER SPECIFICALLY DISCLAIMS ANY WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
// SOFTWARE AND ACCOMPANYING DOCUMENTATION IS PROVIDED "AS IS".
// THE COPYRIGHT HOLDER HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
// SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
//
// **********************************************************************

package imagescience.images;

import ij.ImageStack;
import imagescience.utilities.FMath;

/** Represents images of up to four dimensions containing data of type
    <code>byte</code>. */
public final class ByteImage extends Image {
    
    // Pixel array:
    final byte[][][] pixels;

    /** Constructor for creating a one-dimensional image of <code>byte</code>s.

	@exception IllegalArgumentException if the parameter is zero or negative.
    */
    public ByteImage(final int xdim) {
	
	super(xdim,1,1,1);

	pixels = new byte[1][1][xdim];
    }

    /** Constructor for creating a two-dimensional image of <code>byte</code>s.

	@exception IllegalArgumentException if any of the parameters is zero or negative.
    */
    public ByteImage(final int xdim, final int ydim) {
	
	super(xdim,ydim,1,1);

	pixels = new byte[1][1][xdim*ydim];
    }

    /** Constructor for creating a three-dimensional image of <code>byte</code>s.

	@exception IllegalArgumentException if any of the parameters is zero or negative.
    */
    public ByteImage(final int xdim, final int ydim, final int zdim) {
	
	super(xdim,ydim,zdim,1);

	pixels = new byte[1][zdim][xdim*ydim];
    }

    /** Constructor for creating a four-dimensional image of <code>byte</code>s.

	@exception IllegalArgumentException if any of the parameters is zero or negative.
    */
    public ByteImage(final int xdim, final int ydim, final int zdim, final int tdim) {
	
	super(xdim,ydim,zdim,tdim);

	pixels = new byte[tdim][zdim][xdim*ydim];
    }

    /** Constructor for creating a four-dimensional image of
	<code>byte</code>s. The size of the image in the four dimensions
	is determined by the field values of the parameter.

	@exception NullPointerException if the parameter is <code>null</code>.
    */
    public ByteImage(final Dimensions dims) {
	
	super(dims.x,dims.y,dims.z,dims.t);

	pixels = new byte[tdim][zdim][xdim*ydim];
    }

    /** Copy constructor. All pixel data and corresponding information
        contained in the given image is copied, and nothing in the
        resulting image is shared with the given image.

	@exception NullPointerException if the parameter is <code>null</code>.
    */
    public ByteImage(final ByteImage image) {
	
	super(image.xdim,image.ydim,image.zdim,image.tdim);

	pixels = new byte[tdim][zdim][xdim*ydim];

	final int xydim = xdim*ydim;
	for (int t=0; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy)
		    pixels[t][z][xy] = image.pixels[t][z][xy];

	axes = image.axes;

	xpixdim = image.xpixdim;
	ypixdim = image.ypixdim;
	zpixdim = image.zpixdim;
	tpixdim = image.tpixdim;
    }

    /** Copy constructor. All pixel data and corresponding information
        contained in the given image is copied, and nothing in the
        resulting image is shared with the given image. The pixel values
        in the resulting image are equal to those in the given image but
        clipped to the range [0,255].

	@exception NullPointerException if the parameter is <code>null</code>.
    */
    public ByteImage(final ShortImage image) {
	
	super(image.xdim,image.ydim,image.zdim,image.tdim);

	pixels = new byte[tdim][zdim][xdim*ydim];

	final int xydim = xdim*ydim;
	for (int t=0; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy) {
		    final int value = 0xFFFF & image.pixels[t][z][xy];
		    pixels[t][z][xy] = (byte)((value > 255) ? 255 : value);
		}

	axes = image.axes;

	xpixdim = image.xpixdim;
	ypixdim = image.ypixdim;
	zpixdim = image.zpixdim;
	tpixdim = image.tpixdim;
    }

    /** Copy constructor. All pixel data and corresponding information
        contained in the given image is copied, and nothing in the
        resulting image is shared with the given image. The pixel values
        in the resulting image are equal to those in the given image but
        clipped to the range [0,255] and rounded to the nearest
        <code>byte</code>.

	@exception NullPointerException if the parameter is <code>null</code>.
    */
    public ByteImage(final FloatImage image) {
	
	super(image.xdim,image.ydim,image.zdim,image.tdim);

	pixels = new byte[tdim][zdim][xdim*ydim];

	final int xydim = xdim*ydim;
	for (int t=0; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy) {
		    final float value = image.pixels[t][z][xy];
		    pixels[t][z][xy] = (byte)FMath.round((value > 255.0f) ? 255.0f : ((value < 0.0f) ? 0.0f : value));
		}

	axes = image.axes;

	xpixdim = image.xpixdim;
	ypixdim = image.ypixdim;
	zpixdim = image.zpixdim;
	tpixdim = image.tpixdim;
    }

    /** Copy constructor that allows adding borders. Creates a new image
        whose size in every dimension is equal to that of the given
        image plus twice the value of the given borders size in that
        dimension. All pixel data and corresponding information
        contained in the given image is copied, and nothing in the
        resulting image is shared with the given image. The
        <code>borders</code> parameter determines the starting position
        for copying into the new image: the data is copied in such a way
        that in each dimension, the resulting image has an empty border
        of given size on both sides.

	@exception NullPointerException if any of the parameters is
	<code>null</code>.
    */
    public ByteImage(final ByteImage image, final Borders borders) {
	
	super(image.xdim+2*borders.x,image.ydim+2*borders.y,image.zdim+2*borders.z,image.tdim+2*borders.t);

	pixels = new byte[tdim][zdim][xdim*ydim];

	for (int tin=0, tnew=borders.t; tin<image.tdim; ++tin, ++tnew)
	    for (int zin=0, znew=borders.z; zin<image.zdim; ++zin, ++znew)
		for (int y=0, xyin=0, ynew=borders.y*xdim+borders.x; y<image.ydim; ++y, ynew+=xdim)
		    for (int x=0, xynew=ynew; x<image.xdim; ++x, ++xyin, ++xynew)
			pixels[tnew][znew][xynew] = image.pixels[tin][zin][xyin];

	axes = image.axes;

	xpixdim = image.xpixdim;
	ypixdim = image.ypixdim;
	zpixdim = image.zpixdim;
	tpixdim = image.tpixdim;
    }

    /** Copy constructor that allows adding borders. Creates a new image
        whose size in every dimension is equal to that of the given
        image plus twice the value of the given borders size in that
        dimension. All pixel data and corresponding information
        contained in the given image is copied, and nothing in the
        resulting image is shared with the given image. The
        <code>borders</code> parameter determines the starting position
        for copying into the new image: the data is copied in such a way
        that in each dimension, the resulting image has an empty border
        of given size on both sides. The pixel values in the resulting
        image are equal to the corresponding pixel values in the given
        image but clipped to the range [0,255].

	@exception NullPointerException if any of the parameters is
	<code>null</code>.
    */
    public ByteImage(final ShortImage image, final Borders borders) {
	
	super(image.xdim+2*borders.x,image.ydim+2*borders.y,image.zdim+2*borders.z,image.tdim+2*borders.t);

	pixels = new byte[tdim][zdim][xdim*ydim];

	for (int tin=0, tnew=borders.t; tin<image.tdim; ++tin, ++tnew)
	    for (int zin=0, znew=borders.z; zin<image.zdim; ++zin, ++znew)
		for (int y=0, xyin=0, ynew=borders.y*xdim+borders.x; y<image.ydim; ++y, ynew+=xdim)
		    for (int x=0, xynew=ynew; x<image.xdim; ++x, ++xyin, ++xynew) {
			final int value = 0xFFFF & image.pixels[tin][zin][xyin];
			pixels[tnew][znew][xynew] = (byte)((value > 255) ? 255 : value);
		    }

	axes = image.axes;

	xpixdim = image.xpixdim;
	ypixdim = image.ypixdim;
	zpixdim = image.zpixdim;
	tpixdim = image.tpixdim;
    }

    /** Copy constructor that allows adding borders. Creates a new image
        whose size in every dimension is equal to that of the given
        image plus twice the value of the given borders size in that
        dimension. All pixel data and corresponding information
        contained in the given image is copied, and nothing in the
        resulting image is shared with the given image. The
        <code>borders</code> parameter determines the starting position
        for copying into the new image: the data is copied in such a way
        that in each dimension, the resulting image has an empty border
        of given size on both sides. The pixel values in the resulting
        image are equal to the corresponding pixel values in the given
        image but clipped to the range [0,255] and rounded to the
        nearest <code>byte</code>.

	@exception NullPointerException if any of the parameters is
	<code>null</code>.
    */
    public ByteImage(final FloatImage image, final Borders borders) {
	
	super(image.xdim+2*borders.x,image.ydim+2*borders.y,image.zdim+2*borders.z,image.tdim+2*borders.t);

	pixels = new byte[tdim][zdim][xdim*ydim];

	for (int tin=0, tnew=borders.t; tin<image.tdim; ++tin, ++tnew)
	    for (int zin=0, znew=borders.z; zin<image.zdim; ++zin, ++znew)
		for (int y=0, xyin=0, ynew=borders.y*xdim+borders.x; y<image.ydim; ++y, ynew+=xdim)
		    for (int x=0, xynew=ynew; x<image.xdim; ++x, ++xyin, ++xynew) {
			final float value = image.pixels[tin][zin][xyin];
			pixels[tnew][znew][xynew] = (byte)FMath.round((value > 255.0f) ? 255.0f : ((value < 0.0f) ? 0.0f : value));
		    }

	axes = image.axes;

	xpixdim = image.xpixdim;
	ypixdim = image.ypixdim;
	zpixdim = image.zpixdim;
	tpixdim = image.tpixdim;
    }

    /** Constructor for creating a three-dimensional image wrapper
        object. The size information is copied from the given
        <code>ImageStack</code> object. The actual pixel data, however,
        is not copied to, but shared with the newly created image.

	@exception IllegalArgumentException if the data stored in
	<code>stack</code> is not of type <code>byte</code>.

	@exception NullPointerException if the parameter is <code>null</code>.
    */
    public ByteImage(final ImageStack stack) {
	
	super(stack.getWidth(),stack.getHeight(),stack.getSize(),1);

	final Object[] spixels = stack.getImageArray();
	if (!(spixels[0] instanceof byte[]))
	    throw new IllegalArgumentException("Wrong pixel type");

	pixels = new byte[tdim][zdim][];
	for (int z=0; z<zdim; ++z)
	    pixels[0][z] = (byte[])spixels[z];
    }

    /** Constructor for creating a four-dimensional image wrapper
        object. The size information is copied from the given
        <code>ImageStack</code> object. The actual pixel data, however,
        is not copied to, but shared with the newly created image.

	@exception IllegalArgumentException if the data stored in
	<code>stack</code> is not of type <code>byte</code>, or if the number of
	slices in <code>stack</code> is not an integral multiple of the
	value of the second parameter.

	@exception NullPointerException if the first parameter is <code>null</code>.
    */
    public ByteImage(final ImageStack stack, final int slicespervolume) {

	super(stack.getWidth(),stack.getHeight(),slicespervolume,stack.getSize()/slicespervolume);

	if (zdim*tdim != stack.getSize())
	    throw new IllegalArgumentException("Incorrect number of slices per volume");

	final Object[] spixels = stack.getImageArray();
	if (!(spixels[0] instanceof byte[]))
	    throw new IllegalArgumentException("Wrong pixel type");

	pixels = new byte[tdim][zdim][];
	for (int t=0, slice=0; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z, ++slice)
		pixels[t][z] = (byte[])spixels[slice];
    }

    /** Returns a handle to the internal pixel array of the current
        image. In the current implementation, this is a
        three-dimensional array, where the first two indices from the
        left correspond to the t- and z-dimension, respectively, while
        the right-most index addresses the x-y-dimension, containing the
        sequence of pixel rows. One should be cautious in using this
        method, as the organization of the internal array containing the
        actual pixel data may change in future versions of the class. */
    public byte[][][] array() { return pixels; }

    public ImageStack stack() {

	ImageStack stack = new ImageStack(xdim,ydim);

	for (int t=0; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z)
		stack.addSlice("",pixels[t][z]);

	return stack;
    }

    /** Returns an exact copy of the image. That is to say, all pixel
        data and corresponding information contained in the object is
        duplicated, and nothing in the returned object is shared with
        the object. */
    public ByteImage duplicate() {

	ByteImage dupi = new ByteImage(xdim,ydim,zdim,tdim);

	final int xydim = xdim*ydim;
	for (int t=0; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy)
		    dupi.pixels[t][z][xy] = pixels[t][z][xy];

	dupi.axes = axes;

	dupi.xpixdim = xpixdim;
	dupi.ypixdim = ypixdim;
	dupi.zpixdim = zpixdim;
	dupi.tpixdim = tpixdim;

	return dupi;
    }

    public int memory() { return (xdim*ydim*zdim*tdim)/1024; }

    /** Returns a floating-point representation of the minimum value in
        the image. Since pixel values are taken in the unsigned sense,
        the returned value is always in the range [0,255]. */
    public float minimum() {

	final int xydim = xdim*ydim;
	float min = Float.MAX_VALUE;
	for (int t=0; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy) {
		    final float value = 0xFF & pixels[t][z][xy];
		    if (value < min) min = value;
		}

	return min;
    }

    /** Returns a floating-point representation of the maximum value in
        the image. Since pixel values are taken in the unsigned sense,
        the returned value is always in the range [0,255]. */
    public float maximum() {

	final int xydim = xdim*ydim;
	float max = -Float.MAX_VALUE;
	for (int t=0; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy) {
		    final float value = 0xFF & pixels[t][z][xy];
		    if (value > max) max = value;
		}

	return max;
    }

    /** Returns the pixel value at the given coordinates. Note that the
        pixel value is taken in the unsigned sense and is therefore
        always in the range [0,255].

	@exception ArrayIndexOutOfBoundsException if any of the
	coordinates is out of range.

	@exception NullPointerException if the parameter is <code>null</code>.
    */
    public float get(final Coordinates coords) {

	return (0xFF & pixels[coords.t][coords.z][xdim*coords.y + coords.x]);
    }

    /** Reads pixel values in one dimension into the given array. The
        dimension from which pixel values are read is determined by the
        currently active Cartesian coordinate axis. Exactly one axis
        must be active in order for this method to work. The given
        coordinates indicate the starting position of the given pixel
        array, relative to the image. That is to say, they determine the
        starting (x,y,z,t) position in the image for reading pixel
        values into the array. Note that the coordinate corresponding to
        the active axis is allowed to be negative. However, pixel
        positions outside the image in the active dimension are not
        actually read but simply skipped. This is useful for creating
        borders around the pixel data read into the given array.

	@exception IllegalStateException if more than one coordinate
	axis is active.

	@exception ArrayIndexOutOfBoundsException if the coordinates
        corresponding to the non-active axes are outside the range
        determined by the size of the image.

	@exception NullPointerException if any of the parameters is
	<code>null</code>.
    */
    public void get(final Coordinates coords, final byte[] values) {

	switch (axes) {

	case Axes.X: {
	    int vxstart = 0;
	    int pxstart = coords.x;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    int pxstop = coords.x + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    for (int x=pxstart, px=coords.y*xdim + pxstart, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
		values[vx] = pixels[coords.t][coords.z][px];
	    break;
	}
	case Axes.Y: {
	    int vystart = 0;
	    int pystart = coords.y;
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    int pystop = coords.y + values.length;
	    if (pystop > ydim) pystop = ydim;
	    for (int y=pystart, py=pystart*xdim + coords.x, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		values[vy] = pixels[coords.t][coords.z][py];
	    break;
	}
	case Axes.Z: {
	    int vzstart = 0;
	    int pzstart = coords.z;
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    int pzstop = coords.z + values.length;
	    if (pzstop > zdim) pzstop = zdim;
	    final int pxy = coords.y*xdim + coords.x;
	    for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		values[vz] = pixels[coords.t][pz][pxy];
	    break;
	}
	case Axes.T: {
	    int vtstart = 0;
	    int ptstart = coords.t;
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int ptstop = coords.t + values.length;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pz = coords.z;
	    final int pxy = coords.y*xdim + coords.x;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		values[vt] = pixels[pt][pz][pxy];
	    break;
	}
	default: // wrong number of active axes
	    throw new IllegalStateException("Wrong number of active axes");
	}
    }

    /** Reads pixel values in one dimension into the given array. Note
        that the pixel values are taken in the unsigned sense and are
        therefore always in the range [0,255]. The dimension from which
        pixel values are read is determined by the currently active
        Cartesian coordinate axis. Exactly one axis must be active in
        order for this method to work. The given coordinates indicate
        the starting position of the given pixel array, relative to the
        image. That is to say, they determine the starting (x,y,z,t)
        position in the image for reading pixel values into the
        array. Note that the coordinate corresponding to the active axis
        is allowed to be negative. However, pixel positions outside the
        image in the active dimension are not actually read but simply
        skipped. This is useful for creating borders around the pixel
        data read into the given array.

	@exception IllegalStateException if more than one coordinate
	axis is active.

	@exception ArrayIndexOutOfBoundsException if the coordinates
        corresponding to the non-active axes are outside the range
        determined by the size of the image.

	@exception NullPointerException if any of the parameters is
	<code>null</code>.
    */
    public void get(final Coordinates coords, final float[] values) {

	switch (axes) {

	case Axes.X: {
	    int vxstart = 0;
	    int pxstart = coords.x;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    int pxstop = coords.x + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    for (int x=pxstart, px=coords.y*xdim + pxstart, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
		values[vx] = 0xFF & pixels[coords.t][coords.z][px];
	    break;
	}
	case Axes.Y: {
	    int vystart = 0;
	    int pystart = coords.y;
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    int pystop = coords.y + values.length;
	    if (pystop > ydim) pystop = ydim;
	    for (int y=pystart, py=pystart*xdim + coords.x, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		values[vy] = 0xFF & pixels[coords.t][coords.z][py];
	    break;
	}
	case Axes.Z: {
	    int vzstart = 0;
	    int pzstart = coords.z;
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    int pzstop = coords.z + values.length;
	    if (pzstop > zdim) pzstop = zdim;
	    final int pxy = coords.y*xdim + coords.x;
	    for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		values[vz] = 0xFF & pixels[coords.t][pz][pxy];
	    break;
	}
	case Axes.T: {
	    int vtstart = 0;
	    int ptstart = coords.t;
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int ptstop = coords.t + values.length;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pz = coords.z;
	    final int pxy = coords.y*xdim + coords.x;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		values[vt] = 0xFF & pixels[pt][pz][pxy];
	    break;
	}
	default: // wrong number of active axes
	    throw new IllegalStateException("Wrong number of active axes");
	}
    }

    /** Reads pixel values in two dimensions into the given array. The
        dimensions from which pixel values are read are determined by
        the currently active Cartesian coordinate axes. Exactly two axes
        must be active in order for this method to work. The two indices
        of the given pixel array correspond to these axes, where the
        order of the indices from left to right is from the highest to
        the lowest dimension. The given coordinates indicate the
        starting position of the given pixel array, relative to the
        image. That is to say, they determine the starting (x,y,z,t)
        position in the image for reading pixel values into the
        array. Note that the coordinates corresponding to the active
        axes are allowed to be negative. However, pixel positions
        outside the image in the active dimensions are not actually read
        but simply skipped. This is useful for creating borders around
        the pixel data read into the given array.

	@exception IllegalStateException if not exactly two axes are active.

	@exception ArrayIndexOutOfBoundsException if the coordinates
        corresponding to the non-active axes are outside the range
        determined by the size of the image.

	@exception NullPointerException if any of the parameters is
	<code>null</code>.
    */
    public void get(final Coordinates coords, final byte[][] values) {

	switch (axes) {

	case Axes.YX: {
	    int vxstart = 0;
	    int vystart = 0;
	    int pxstart = coords.x;
	    int pystart = coords.y;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    int pxstop = coords.x + values[0].length;
	    int pystop = coords.y + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pystop > ydim) pystop = ydim;
	    for (int y=pystart, py=pystart*xdim + pxstart, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		for (int x=pxstart, px=py, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
		    values[vy][vx] = pixels[coords.t][coords.z][px];
	    break;
	}
	case Axes.ZX: {
	    int vxstart = 0;
	    int vzstart = 0;
	    int pxstart = coords.x;
	    int pzstart = coords.z;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    int pxstop = coords.x + values[0].length;
	    int pzstop = coords.z + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pzstop > zdim) pzstop = zdim;
	    final int pxy = coords.y*xdim + pxstart;
	    for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		for (int x=pxstart, px=pxy, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
		    values[vz][vx] = pixels[coords.t][pz][px];
	    break;
	}
	case Axes.TX: {
	    int vxstart = 0;
	    int vtstart = 0;
	    int pxstart = coords.x;
	    int ptstart = coords.t;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pxstop = coords.x + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = coords.y*xdim + pxstart;
	    final int pz = coords.z;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int x=pxstart, px=pxy, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
		    values[vt][vx] = pixels[pt][pz][px];
	    break;
	}
	case Axes.ZY: {
	    int vystart = 0;
	    int vzstart = 0;
	    int pystart = coords.y;
	    int pzstart = coords.z;
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    int pystop = coords.y + values[0].length;
	    int pzstop = coords.z + values.length;
	    if (pystop > ydim) pystop = ydim;
	    if (pzstop > zdim) pzstop = zdim;
	    final int pxy = pystart*xdim + coords.x;
	    for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		    values[vz][vy] = pixels[coords.t][pz][py];
	    break;
	}
	case Axes.TY: {
	    int vystart = 0;
	    int vtstart = 0;
	    int pystart = coords.y;
	    int ptstart = coords.t;
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pystop = coords.y + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pystop > ydim) pystop = ydim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = pystart*xdim + coords.x;
	    final int pz = coords.z;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		    values[vt][vy] = pixels[pt][pz][py];
	    break;
	}
	case Axes.TZ: {
	    int vzstart = 0;
	    int vtstart = 0;
	    int pzstart = coords.z;
	    int ptstart = coords.t;
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pzstop = coords.z + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pzstop > zdim) pzstop = zdim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = coords.y*xdim + coords.x;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		    values[vt][vz] = pixels[pt][pz][pxy];
	    break;
	}
	default: // wrong number of active axes
	    throw new IllegalStateException("Wrong number of active axes");
	}
    }

    /** Reads pixel values in two dimensions into the given array. Note
        that the pixel values are taken in the unsigned sense and are
        therefore always in the range [0,255]. The dimensions from which
        pixel values are read are determined by the currently active
        Cartesian coordinate axes. Exactly two axes must be active in
        order for this method to work. The two indices of the given
        pixel array correspond to these axes, where the order of the
        indices from left to right is from the highest to the lowest
        dimension. The given coordinates indicate the starting position
        of the given pixel array, relative to the image. That is to say,
        they determine the starting (x,y,z,t) position in the image for
        reading pixel values into the array. Note that the coordinates
        corresponding to the active axes are allowed to be
        negative. However, pixel positions outside the image in the
        active dimensions are not actually read but simply skipped. This
        is useful for creating borders around the pixel data read into
        the given array.

	@exception IllegalStateException if not exactly two axes are active.

	@exception ArrayIndexOutOfBoundsException if the coordinates
        corresponding to the non-active axes are outside the range
        determined by the size of the image.

	@exception NullPointerException if any of the parameters is
	<code>null</code>.
    */
    public void get(final Coordinates coords, final float[][] values) {

	switch (axes) {

	case Axes.YX: {
	    int vxstart = 0;
	    int vystart = 0;
	    int pxstart = coords.x;
	    int pystart = coords.y;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    int pxstop = coords.x + values[0].length;
	    int pystop = coords.y + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pystop > ydim) pystop = ydim;
	    for (int y=pystart, py=pystart*xdim + pxstart, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		for (int x=pxstart, px=py, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
		    values[vy][vx] = 0xFF & pixels[coords.t][coords.z][px];
	    break;
	}
	case Axes.ZX: {
	    int vxstart = 0;
	    int vzstart = 0;
	    int pxstart = coords.x;
	    int pzstart = coords.z;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    int pxstop = coords.x + values[0].length;
	    int pzstop = coords.z + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pzstop > zdim) pzstop = zdim;
	    final int pxy = coords.y*xdim + pxstart;
	    for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		for (int x=pxstart, px=pxy, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
		    values[vz][vx] = 0xFF & pixels[coords.t][pz][px];
	    break;
	}
	case Axes.TX: {
	    int vxstart = 0;
	    int vtstart = 0;
	    int pxstart = coords.x;
	    int ptstart = coords.t;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pxstop = coords.x + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = coords.y*xdim + pxstart;
	    final int pz = coords.z;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int x=pxstart, px=pxy, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
		    values[vt][vx] = 0xFF & pixels[pt][pz][px];
	    break;
	}
	case Axes.ZY: {
	    int vystart = 0;
	    int vzstart = 0;
	    int pystart = coords.y;
	    int pzstart = coords.z;
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    int pystop = coords.y + values[0].length;
	    int pzstop = coords.z + values.length;
	    if (pystop > ydim) pystop = ydim;
	    if (pzstop > zdim) pzstop = zdim;
	    final int pxy = pystart*xdim + coords.x;
	    for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		    values[vz][vy] = 0xFF & pixels[coords.t][pz][py];
	    break;
	}
	case Axes.TY: {
	    int vystart = 0;
	    int vtstart = 0;
	    int pystart = coords.y;
	    int ptstart = coords.t;
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pystop = coords.y + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pystop > ydim) pystop = ydim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = pystart*xdim + coords.x;
	    final int pz = coords.z;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		    values[vt][vy] = 0xFF & pixels[pt][pz][py];
	    break;
	}
	case Axes.TZ: {
	    int vzstart = 0;
	    int vtstart = 0;
	    int pzstart = coords.z;
	    int ptstart = coords.t;
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pzstop = coords.z + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pzstop > zdim) pzstop = zdim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = coords.y*xdim + coords.x;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		    values[vt][vz] = 0xFF & pixels[pt][pz][pxy];
	    break;
	}
	default: // wrong number of active axes
	    throw new IllegalStateException("Wrong number of active axes");
	}
    }

    /** Reads pixel values in three dimensions into the given array. The
        dimensions from which pixel values are read are determined by
        the currently active Cartesian coordinate axes. Exactly three
        axes must be active in order for this method to work. The three
        indices of the given pixel array correspond to these axes, where
        the order of the indices from left to right is from the highest
        to the lowest dimension. The given coordinates indicate the
        starting position of the given pixel array, relative to the
        image. That is to say, they determine the starting (x,y,z,t)
        position in the image for reading pixel values into the
        array. Note that the coordinates corresponding to the active
        axes are allowed to be negative. However, pixel positions
        outside the image in the active dimensions are not actually read
        but simply skipped. This is useful for creating borders around
        the pixel data read into the given array.

	@exception IllegalStateException if not exactly three axes are active.

	@exception ArrayIndexOutOfBoundsException if the coordinate
        corresponding to the non-active axis is outside the range
        determined by the size of the image.

	@exception NullPointerException if any of the parameters is
	<code>null</code>.
    */
    public void get(final Coordinates coords, final byte[][][] values) {

	switch (axes) {

	case Axes.ZYX: {
	    int vxstart = 0;
	    int vystart = 0;
	    int vzstart = 0;
	    int pxstart = coords.x;
	    int pystart = coords.y;
	    int pzstart = coords.z;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    int pxstop = coords.x + values[0][0].length;
	    int pystop = coords.y + values[0].length;
	    int pzstop = coords.z + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pystop > ydim) pystop = ydim;
	    if (pzstop > zdim) pzstop = zdim;
	    final int pxy = pystart*xdim + pxstart;
	    for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		    for (int x=pxstart, px=py, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
			values[vz][vy][vx] = pixels[coords.t][pz][px];
	    break;
	}
	case Axes.TZX: {
	    int vxstart = 0;
	    int vzstart = 0;
	    int vtstart = 0;
	    int pxstart = coords.x;
	    int pzstart = coords.z;
	    int ptstart = coords.t;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pxstop = coords.x + values[0][0].length;
	    int pzstop = coords.z + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pzstop > zdim) pzstop = zdim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = coords.y*xdim + pxstart;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		    for (int x=pxstart, px=pxy, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
			values[vt][vz][vx] = pixels[pt][pz][px];
	    break;
	}
	case Axes.TYX: {
	    int vxstart = 0;
	    int vystart = 0;
	    int vtstart = 0;
	    int pxstart = coords.x;
	    int pystart = coords.y;
	    int ptstart = coords.t;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pxstop = coords.x + values[0][0].length;
	    int pystop = coords.y + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pystop > ydim) pystop = ydim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = pystart*xdim + pxstart;
	    final int pz = coords.z;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		    for (int x=pxstart, px=py, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
			values[vt][vy][vx] = pixels[pt][pz][px];
	    break;
	}
	case Axes.TZY: {
	    int vystart = 0;
	    int vzstart = 0;
	    int vtstart = 0;
	    int pystart = coords.y;
	    int pzstart = coords.z;
	    int ptstart = coords.t;
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pystop = coords.y + values[0][0].length;
	    int pzstop = coords.z + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pystop > ydim) pystop = ydim;
	    if (pzstop > zdim) pzstop = zdim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = pystart*xdim + coords.x;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		    for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
			values[vt][vz][vy] = pixels[pt][pz][py];
	    break;
	}
	default: // wrong number of active axes
	    throw new IllegalStateException("Wrong number of active axes");
	}
    }

    /** Reads pixel values in three dimensions into the given
        array. Note that the pixel values are taken in the unsigned
        sense and are therefore always in the range [0,255]. The
        dimensions from which pixel values are read are determined by
        the currently active Cartesian coordinate axes. Exactly three
        axes must be active in order for this method to work. The three
        indices of the given pixel array correspond to these axes, where
        the order of the indices from left to right is from the highest
        to the lowest dimension. The given coordinates indicate the
        starting position of the given pixel array, relative to the
        image. That is to say, they determine the starting (x,y,z,t)
        position in the image for reading pixel values into the
        array. Note that the coordinates corresponding to the active
        axes are allowed to be negative. However, pixel positions
        outside the image in the active dimensions are not actually read
        but simply skipped. This is useful for creating borders around
        the pixel data read into the given array.

	@exception IllegalStateException if not exactly three axes are active.

	@exception ArrayIndexOutOfBoundsException if the coordinate
        corresponding to the non-active axis is outside the range
        determined by the size of the image.

	@exception NullPointerException if any of the parameters is
	<code>null</code>.
    */
    public void get(final Coordinates coords, final float[][][] values) {

	switch (axes) {

	case Axes.ZYX: {
	    int vxstart = 0;
	    int vystart = 0;
	    int vzstart = 0;
	    int pxstart = coords.x;
	    int pystart = coords.y;
	    int pzstart = coords.z;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    int pxstop = coords.x + values[0][0].length;
	    int pystop = coords.y + values[0].length;
	    int pzstop = coords.z + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pystop > ydim) pystop = ydim;
	    if (pzstop > zdim) pzstop = zdim;
	    final int pxy = pystart*xdim + pxstart;
	    for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		    for (int x=pxstart, px=py, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
			values[vz][vy][vx] = 0xFF & pixels[coords.t][pz][px];
	    break;
	}
	case Axes.TZX: {
	    int vxstart = 0;
	    int vzstart = 0;
	    int vtstart = 0;
	    int pxstart = coords.x;
	    int pzstart = coords.z;
	    int ptstart = coords.t;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pxstop = coords.x + values[0][0].length;
	    int pzstop = coords.z + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pzstop > zdim) pzstop = zdim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = coords.y*xdim + pxstart;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		    for (int x=pxstart, px=pxy, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
			values[vt][vz][vx] = 0xFF & pixels[pt][pz][px];
	    break;
	}
	case Axes.TYX: {
	    int vxstart = 0;
	    int vystart = 0;
	    int vtstart = 0;
	    int pxstart = coords.x;
	    int pystart = coords.y;
	    int ptstart = coords.t;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pxstop = coords.x + values[0][0].length;
	    int pystop = coords.y + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pystop > ydim) pystop = ydim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = pystart*xdim + pxstart;
	    final int pz = coords.z;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		    for (int x=pxstart, px=py, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
			values[vt][vy][vx] = 0xFF & pixels[pt][pz][px];
	    break;
	}
	case Axes.TZY: {
	    int vystart = 0;
	    int vzstart = 0;
	    int vtstart = 0;
	    int pystart = coords.y;
	    int pzstart = coords.z;
	    int ptstart = coords.t;
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pystop = coords.y + values[0][0].length;
	    int pzstop = coords.z + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pystop > ydim) pystop = ydim;
	    if (pzstop > zdim) pzstop = zdim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = pystart*xdim + coords.x;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		    for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
			values[vt][vz][vy] = 0xFF & pixels[pt][pz][py];
	    break;
	}
	default: // wrong number of active axes
	    throw new IllegalStateException("Wrong number of active axes");
	}
    }

    /** Reads pixel values in all four dimensions into the given
        array. All four axes must be active in order for this method to
        work. The four indices of the given pixel array correspond to
        these axes, where the order of the indices from left to right is
        t, z, y, x. The given coordinates indicate the starting position
        of the given pixel array, relative to the image. That is to say,
        they determine the starting (x,y,z,t) position in the image for
        reading pixel values into the array. Note that the coordinates
        are allowed to be negative. However, pixel positions outside the
        image are not actually read but simply skipped. This is useful
        for creating borders around the pixel data read into the given
        array.

	@exception IllegalStateException if not all four axes are active.

	@exception NullPointerException if any of the parameters is
	<code>null</code>.
    */
    public void get(final Coordinates coords, final byte[][][][] values) {

	switch (axes) {

	case Axes.TZYX: {
	    int vxstart = 0;
	    int vystart = 0;
	    int vzstart = 0;
	    int vtstart = 0;
	    int pxstart = coords.x;
	    int pystart = coords.y;
	    int pzstart = coords.z;
	    int ptstart = coords.t;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pxstop = coords.x + values[0][0][0].length;
	    int pystop = coords.y + values[0][0].length;
	    int pzstop = coords.z + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pystop > ydim) pystop = ydim;
	    if (pzstop > zdim) pzstop = zdim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = pystart*xdim + pxstart;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		    for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
			for (int x=pxstart, px=py, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
			    values[vt][vz][vy][vx] = pixels[pt][pz][px];
	    break;
	}
	default: // wrong number of active axes
	    throw new IllegalStateException("Wrong number of active axes");
	}
    }

    /** Reads pixel values in all four dimensions into the given
        array. Note that the pixel values are taken in the unsigned
        sense and are therefore always in the range [0,255]. All four
        axes must be active in order for this method to work. The four
        indices of the given pixel array correspond to these axes, where
        the order of the indices from left to right is t, z, y, x. The
        given coordinates indicate the starting position of the given
        pixel array, relative to the image. That is to say, they
        determine the starting (x,y,z,t) position in the image for
        reading pixel values into the array. Note that the coordinates
        are allowed to be negative. However, pixel positions outside the
        image are not actually read but simply skipped. This is useful
        for creating borders around the pixel data read into the given
        array.

	@exception IllegalStateException if not all four axes are active.

	@exception NullPointerException if any of the parameters is
	<code>null</code>.
    */
    public void get(final Coordinates coords, final float[][][][] values) {

	switch (axes) {

	case Axes.TZYX: {
	    int vxstart = 0;
	    int vystart = 0;
	    int vzstart = 0;
	    int vtstart = 0;
	    int pxstart = coords.x;
	    int pystart = coords.y;
	    int pzstart = coords.z;
	    int ptstart = coords.t;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pxstop = coords.x + values[0][0][0].length;
	    int pystop = coords.y + values[0][0].length;
	    int pzstop = coords.z + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pystop > ydim) pystop = ydim;
	    if (pzstop > zdim) pzstop = zdim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = pystart*xdim + pxstart;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		    for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
			for (int x=pxstart, px=py, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
			    values[vt][vz][vy][vx] = 0xFF & pixels[pt][pz][px];
	    break;
	}
	default: // wrong number of active axes
	    throw new IllegalStateException("Wrong number of active axes");
	}
    }

    /** Sets all pixels to the given value. */
    public void set(final byte value) {

	final int xydim = xdim*ydim;
	for (int t=0; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy)
		    pixels[t][z][xy] = value;
    }

    /** Sets all pixels to the given value. The given value is clipped
        to the range [0,255] and rounded to the nearest <code>byte</code>
        value. */
    public void set(final float value) {

	final int xydim = xdim*ydim;
	final byte val = (byte)FMath.round((value > 255.0f) ? 255.0f : ((value < 0.0f) ? 0.0f : value));
	for (int t=0; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy)
		    pixels[t][z][xy] = val;
    }

    /** Sets the pixel at the given coordinates to the given value.

	@exception ArrayIndexOutOfBoundsException if any of the
	coordinates is out of range.

	@exception NullPointerException if the first parameter is
	<code>null</code>.
    */
    public void set(final Coordinates coords, final byte value) {

	pixels[coords.t][coords.z][xdim*coords.y + coords.x] = value;
    }
    
    /** Sets the pixel at the given coordinates to the given
        value. The given value is clipped to the range [0,255] and
        rounded to the nearest <code>byte</code> value.

	@exception ArrayIndexOutOfBoundsException if any of the
	coordinates is out of range.

	@exception NullPointerException if the first parameter is
	<code>null</code>.
    */
    public void set(final Coordinates coords, final float value) {

	pixels[coords.t][coords.z][xdim*coords.y + coords.x] =
	    (byte)FMath.round((value > 255.0f) ? 255.0f : ((value < 0.0f) ? 0.0f : value));
    }
    
    /** Sets all pixels in the given borders to the given value. The
        size of the borders at the beginning and end in any dimension of
        the image is determined by the corresponding field of
        <code>borders</code>.
	
	@exception IllegalArgumentException if any of the given border
	sizes is larger than half the image size in the corresponding
	dimension.
	
	@exception NullPointerException if the first parameter is
	<code>null</code>.
    */
    public void set(final Borders borders, final byte value) {

	if (borders.x > xdim/2 || borders.y > ydim/2 ||
	    borders.z > zdim/2 || borders.t > tdim/2)
	    throw new IllegalArgumentException("Border(s) too large");

	// Initialize:
	final int tdimb = tdim - borders.t;
	final int zdimb = zdim - borders.z;
	final int ydimb = ydim - borders.y;
	final int xdimb = xdim - borders.x;
	final int xydim = xdim*ydim;

	// Fill borders in t-dimension:
	for (int t=0; t<borders.t; ++t)
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy)
		    pixels[t][z][xy] = value;
	for (int t=tdimb; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy)
		    pixels[t][z][xy] = value;

	// Fill borders in z-dimension:
	for (int t=borders.t; t<tdimb; ++t) {
	    for (int z=0; z<borders.z; ++z)
		for (int xy=0; xy<xydim; ++xy)
		    pixels[t][z][xy] = value;
	    for (int z=zdimb; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy)
		    pixels[t][z][xy] = value;
	}

	// Fill borders in y-dimension:
	for (int t=borders.t; t<tdimb; ++t)
	    for (int z=borders.z; z<zdimb; ++z) {
		for (int y=0, xy=0; y<borders.y; ++y)
		    for (int x=0; x<xdim; ++x, ++xy)
			pixels[t][z][xy] = value;
		for (int y=ydimb, xy=ydimb*xdim; y<ydim; ++y)
		    for (int x=0; x<xdim; ++x, ++xy)
			pixels[t][z][xy] = value;
	    }

	// Fill borders in x-dimension:
	for (int t=borders.t; t<tdimb; ++t)
	    for (int z=borders.z; z<zdimb; ++z)
		for (int y=borders.y; y<ydimb; ++y) {
		    for (int x=0, xy=y*xdim; x<borders.x; ++x, ++xy)
			pixels[t][z][xy] = value;
		    for (int x=xdimb, xy=y*xdim+xdimb; x<xdim; ++x, ++xy)
			pixels[t][z][xy] = value;
		}
    }
    
    /** Sets all pixels in the given borders to the given value. The
        given value is clipped to the range [0,255] and rounded to the
        nearest <code>byte</code> value. The size of the borders at the
        beginning and end in any dimension of the image is determined by
        the corresponding field of <code>borders</code>.
	
	@exception IllegalArgumentException if any of the given border
	sizes is larger than half the image size in the corresponding
	dimension.
	
	@exception NullPointerException if the first parameter is
	<code>null</code>.
    */
    public void set(final Borders borders, final float value) {

	if (borders.x > xdim/2 || borders.y > ydim/2 ||
	    borders.z > zdim/2 || borders.t > tdim/2)
	    throw new IllegalArgumentException("Border(s) too large");

	// Initialize:
	final int tdimb = tdim - borders.t;
	final int zdimb = zdim - borders.z;
	final int ydimb = ydim - borders.y;
	final int xdimb = xdim - borders.x;
	final int xydim = xdim*ydim;
	final byte bvalue = (byte)FMath.round((value > 255.0f) ? 255.0f : ((value < 0.0f) ? 0.0f : value));

	// Fill borders in t-dimension:
	for (int t=0; t<borders.t; ++t)
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy)
		    pixels[t][z][xy] = bvalue;
	for (int t=tdimb; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy)
		    pixels[t][z][xy] = bvalue;

	// Fill borders in z-dimension:
	for (int t=borders.t; t<tdimb; ++t) {
	    for (int z=0; z<borders.z; ++z)
		for (int xy=0; xy<xydim; ++xy)
		    pixels[t][z][xy] = bvalue;
	    for (int z=zdimb; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy)
		    pixels[t][z][xy] = bvalue;
	}

	// Fill borders in y-dimension:
	for (int t=borders.t; t<tdimb; ++t)
	    for (int z=borders.z; z<zdimb; ++z) {
		for (int y=0, xy=0; y<borders.y; ++y)
		    for (int x=0; x<xdim; ++x, ++xy)
			pixels[t][z][xy] = bvalue;
		for (int y=ydimb, xy=ydimb*xdim; y<ydim; ++y)
		    for (int x=0; x<xdim; ++x, ++xy)
			pixels[t][z][xy] = bvalue;
	    }

	// Fill borders in x-dimension:
	for (int t=borders.t; t<tdimb; ++t)
	    for (int z=borders.z; z<zdimb; ++z)
		for (int y=borders.y; y<ydimb; ++y) {
		    for (int x=0, xy=y*xdim; x<borders.x; ++x, ++xy)
			pixels[t][z][xy] = bvalue;
		    for (int x=xdimb, xy=y*xdim+xdimb; x<xdim; ++x, ++xy)
			pixels[t][z][xy] = bvalue;
		}
    }
    
    /** Writes given pixel values in one dimension into the image. The
        dimension into which pixel values are written is determined by
        the currently active Cartesian coordinate axis. Exactly one axis
        must be active in order for this method to work. The given
        coordinates indicate the starting position of the given pixel
        array, relative to the image. That is to say, they determine the
        starting (x,y,z,t) position in the image for writing pixel
        values. Note that the coordinate corresponding to the active
        axis is allowed to be negative. However, pixel positions outside
        the image in the active dimension are not actually written but
        simply skipped. This is useful for ignoring possible borders
        around the pixel data in the given array.

	@exception IllegalStateException if more than one coordinate
	axis is active.

	@exception ArrayIndexOutOfBoundsException if the coordinates
        corresponding to the non-active axes are outside the range
        determined by the size of the image.

	@exception NullPointerException if any of the parameters is
	<code>null</code>.
    */
    public void set(final Coordinates coords, final byte[] values) {

	switch (axes) {

	case Axes.X: {
	    int vxstart = 0;
	    int pxstart = coords.x;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    int pxstop = coords.x + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    for (int x=pxstart, px=coords.y*xdim + pxstart, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
		pixels[coords.t][coords.z][px] = values[vx];
	    break;
	}
	case Axes.Y: {
	    int vystart = 0;
	    int pystart = coords.y;
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    int pystop = coords.y + values.length;
	    if (pystop > ydim) pystop = ydim;
	    for (int y=pystart, py=pystart*xdim + coords.x, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		pixels[coords.t][coords.z][py] = values[vy];
	    break;
	}
	case Axes.Z: {
	    int vzstart = 0;
	    int pzstart = coords.z;
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    int pzstop = coords.z + values.length;
	    if (pzstop > zdim) pzstop = zdim;
	    final int pxy = coords.y*xdim + coords.x;
	    for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		pixels[coords.t][pz][pxy] = values[vz];
	    break;
	}
	case Axes.T: {
	    int vtstart = 0;
	    int ptstart = coords.t;
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int ptstop = coords.t + values.length;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pz = coords.z;
	    final int pxy = coords.y*xdim + coords.x;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		pixels[pt][pz][pxy] = values[vt];
	    break;
	}
	default: // wrong number of active axes
	    throw new IllegalStateException("Wrong number of active axes");
	}
    }

    /** Writes given pixel values in one dimension into the image. The
        values are clipped to the range [0,255] and rounded to the
        nearest <code>byte</code> value. The dimension into which pixel
        values are written is determined by the currently active
        Cartesian coordinate axis. Exactly one axis must be active in
        order for this method to work. The given coordinates indicate
        the starting position of the given pixel array, relative to the
        image. That is to say, they determine the starting (x,y,z,t)
        position in the image for writing pixel values. Note that the
        coordinate corresponding to the active axis is allowed to be
        negative. However, pixel positions outside the image in the
        active dimension are not actually written but simply
        skipped. This is useful for ignoring possible borders around the
        pixel data in the given array.

	@exception IllegalStateException if more than one coordinate
	axis is active.

	@exception ArrayIndexOutOfBoundsException if the coordinates
        corresponding to the non-active axes are outside the range
        determined by the size of the image.

	@exception NullPointerException if any of the parameters is
	<code>null</code>.
    */
    public void set(final Coordinates coords, final float[] values) {

	switch (axes) {

	case Axes.X: {
	    int vxstart = 0;
	    int pxstart = coords.x;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    int pxstop = coords.x + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    for (int x=pxstart, px=coords.y*xdim + pxstart, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
		pixels[coords.t][coords.z][px] = (byte)FMath.round((values[vx] > 255.0f) ? 255.0f : ((values[vx] < 0.0f) ? 0.0f : values[vx]));
	    break;
	}
	case Axes.Y: {
	    int vystart = 0;
	    int pystart = coords.y;
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    int pystop = coords.y + values.length;
	    if (pystop > ydim) pystop = ydim;
	    for (int y=pystart, py=pystart*xdim + coords.x, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		pixels[coords.t][coords.z][py] = (byte)FMath.round((values[vy] > 255.0f) ? 255.0f : ((values[vy] < 0.0f) ? 0.0f : values[vy]));
	    break;
	}
	case Axes.Z: {
	    int vzstart = 0;
	    int pzstart = coords.z;
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    int pzstop = coords.z + values.length;
	    if (pzstop > zdim) pzstop = zdim;
	    final int pxy = coords.y*xdim + coords.x;
	    for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		pixels[coords.t][pz][pxy] = (byte)FMath.round((values[vz] > 255.0f) ? 255.0f : ((values[vz] < 0.0f) ? 0.0f : values[vz]));
	    break;
	}
	case Axes.T: {
	    int vtstart = 0;
	    int ptstart = coords.t;
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int ptstop = coords.t + values.length;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pz = coords.z;
	    final int pxy = coords.y*xdim + coords.x;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		pixels[pt][pz][pxy] = (byte)FMath.round((values[vt] > 255.0f) ? 255.0f : ((values[vt] < 0.0f) ? 0.0f : values[vt]));
	    break;
	}
	default: // wrong number of active axes
	    throw new IllegalStateException("Wrong number of active axes");
	}
    }

    /** Writes given pixel values in two dimensions into the image. The
        dimensions into which pixel values are written are determined by
        the currently active Cartesian coordinate axes. Exactly two axes
        must be active in order for this method to work. The two indices
        of the given pixel array correspond to these axes, where the
        order of the indices from left to right is from the highest to
        the lowest dimension. The given coordinates indicate the
        starting position of the given pixel array, relative to the
        image. That is to say, they determine the starting (x,y,z,t)
        position in the image for writing pixel values. Note that the
        coordinates corresponding to the active axes are allowed to be
        negative. However, pixel positions outside the image in the
        active dimensions are not actually written but simply
        skipped. This is useful for ignoring possible borders around the
        pixel data in the given array.

	@exception IllegalStateException if not exactly two axes are active.

	@exception ArrayIndexOutOfBoundsException if the coordinates
        corresponding to the non-active axes are outside the range
        determined by the size of the image.

	@exception NullPointerException if any of the parameters is
	<code>null</code>.
    */
    public void set(final Coordinates coords, final byte[][] values) {
	
	switch (axes) {

	case Axes.YX: {
	    int vxstart = 0;
	    int vystart = 0;
	    int pxstart = coords.x;
	    int pystart = coords.y;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    int pxstop = coords.x + values[0].length;
	    int pystop = coords.y + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pystop > ydim) pystop = ydim;
	    for (int y=pystart, py=pystart*xdim + pxstart, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		for (int x=pxstart, px=py, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
		    pixels[coords.t][coords.z][px] = values[vy][vx];
	    break;
	}
	case Axes.ZX: {
	    int vxstart = 0;
	    int vzstart = 0;
	    int pxstart = coords.x;
	    int pzstart = coords.z;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    int pxstop = coords.x + values[0].length;
	    int pzstop = coords.z + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pzstop > zdim) pzstop = zdim;
	    final int pxy = coords.y*xdim + pxstart;
	    for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		for (int x=pxstart, px=pxy, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
		    pixels[coords.t][pz][px] = values[vz][vx];
	    break;
	}
	case Axes.TX: {
	    int vxstart = 0;
	    int vtstart = 0;
	    int pxstart = coords.x;
	    int ptstart = coords.t;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pxstop = coords.x + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = coords.y*xdim + pxstart;
	    final int pz = coords.z;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int x=pxstart, px=pxy, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
		    pixels[pt][pz][px] = values[vt][vx];
	    break;
	}
	case Axes.ZY: {
	    int vystart = 0;
	    int vzstart = 0;
	    int pystart = coords.y;
	    int pzstart = coords.z;
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    int pystop = coords.y + values[0].length;
	    int pzstop = coords.z + values.length;
	    if (pystop > ydim) pystop = ydim;
	    if (pzstop > zdim) pzstop = zdim;
	    final int pxy = pystart*xdim + coords.x;
	    for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		    pixels[coords.t][pz][py] = values[vz][vy];
	    break;
	}
	case Axes.TY: {
	    int vystart = 0;
	    int vtstart = 0;
	    int pystart = coords.y;
	    int ptstart = coords.t;
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pystop = coords.y + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pystop > ydim) pystop = ydim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = pystart*xdim + coords.x;
	    final int pz = coords.z;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		    pixels[pt][pz][py] = values[vt][vy];
	    break;
	}
	case Axes.TZ: {
	    int vzstart = 0;
	    int vtstart = 0;
	    int pzstart = coords.z;
	    int ptstart = coords.t;
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pzstop = coords.z + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pzstop > zdim) pzstop = zdim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = coords.y*xdim + coords.x;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		    pixels[pt][pz][pxy] = values[vt][vz];
	    break;
	}
	default: // wrong number of active axes
	    throw new IllegalStateException("Wrong number of active axes");
	}
    }

    /** Writes given pixel values in two dimensions into the image. The
        values are clipped to the range [0,255] and rounded to the
        nearest <code>byte</code> value. The dimensions into which pixel
        values are written are determined by the currently active
        Cartesian coordinate axes. Exactly two axes must be active in
        order for this method to work. The two indices of the given
        pixel array correspond to these axes, where the order of the
        indices from left to right is from the highest to the lowest
        dimension. The given coordinates indicate the starting position
        of the given pixel array, relative to the image. That is to say,
        they determine the starting (x,y,z,t) position in the image for
        writing pixel values. Note that the coordinates corresponding to
        the active axes are allowed to be negative. However, pixel
        positions outside the image in the active dimensions are not
        actually written but simply skipped. This is useful for ignoring
        possible borders around the pixel data in the given array.

	@exception IllegalStateException if not exactly two axes are active.

	@exception ArrayIndexOutOfBoundsException if the coordinates
        corresponding to the non-active axes are outside the range
        determined by the size of the image.

	@exception NullPointerException if any of the parameters is
	<code>null</code>.
    */
    public void set(final Coordinates coords, final float[][] values) {
	
	switch (axes) {

	case Axes.YX: {
	    int vxstart = 0;
	    int vystart = 0;
	    int pxstart = coords.x;
	    int pystart = coords.y;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    int pxstop = coords.x + values[0].length;
	    int pystop = coords.y + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pystop > ydim) pystop = ydim;
	    for (int y=pystart, py=pystart*xdim + pxstart, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		for (int x=pxstart, px=py, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
		    pixels[coords.t][coords.z][px] = (byte)FMath.round((values[vy][vx] > 255.0f) ? 255.0f : ((values[vy][vx] < 0.0f) ? 0.0f : values[vy][vx]));
	    break;
	}
	case Axes.ZX: {
	    int vxstart = 0;
	    int vzstart = 0;
	    int pxstart = coords.x;
	    int pzstart = coords.z;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    int pxstop = coords.x + values[0].length;
	    int pzstop = coords.z + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pzstop > zdim) pzstop = zdim;
	    final int pxy = coords.y*xdim + pxstart;
	    for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		for (int x=pxstart, px=pxy, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
		    pixels[coords.t][pz][px] = (byte)FMath.round((values[vz][vx] > 255.0f) ? 255.0f : ((values[vz][vx] < 0.0f) ? 0.0f : values[vz][vx]));
	    break;
	}
	case Axes.TX: {
	    int vxstart = 0;
	    int vtstart = 0;
	    int pxstart = coords.x;
	    int ptstart = coords.t;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pxstop = coords.x + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = coords.y*xdim + pxstart;
	    final int pz = coords.z;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int x=pxstart, px=pxy, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
		    pixels[pt][pz][px] = (byte)FMath.round((values[vt][vx] > 255.0f) ? 255.0f : ((values[vt][vx] < 0.0f) ? 0.0f : values[vt][vx]));
	    break;
	}
	case Axes.ZY: {
	    int vystart = 0;
	    int vzstart = 0;
	    int pystart = coords.y;
	    int pzstart = coords.z;
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    int pystop = coords.y + values[0].length;
	    int pzstop = coords.z + values.length;
	    if (pystop > ydim) pystop = ydim;
	    if (pzstop > zdim) pzstop = zdim;
	    final int pxy = pystart*xdim + coords.x;
	    for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		    pixels[coords.t][pz][py] = (byte)FMath.round((values[vz][vy] > 255.0f) ? 255.0f : ((values[vz][vy] < 0.0f) ? 0.0f : values[vz][vy]));
	    break;
	}
	case Axes.TY: {
	    int vystart = 0;
	    int vtstart = 0;
	    int pystart = coords.y;
	    int ptstart = coords.t;
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pystop = coords.y + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pystop > ydim) pystop = ydim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = pystart*xdim + coords.x;
	    final int pz = coords.z;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		    pixels[pt][pz][py] = (byte)FMath.round((values[vt][vy] > 255.0f) ? 255.0f : ((values[vt][vy] < 0.0f) ? 0.0f : values[vt][vy]));
	    break;
	}
	case Axes.TZ: {
	    int vzstart = 0;
	    int vtstart = 0;
	    int pzstart = coords.z;
	    int ptstart = coords.t;
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pzstop = coords.z + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pzstop > zdim) pzstop = zdim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = coords.y*xdim + coords.x;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		    pixels[pt][pz][pxy] = (byte)FMath.round((values[vt][vz] > 255.0f) ? 255.0f : ((values[vt][vz] < 0.0f) ? 0.0f : values[vt][vz]));
	    break;
	}
	default: // wrong number of active axes
	    throw new IllegalStateException("Wrong number of active axes");
	}
    }

    /** Writes given pixel values in three dimensions into the
        image. The dimensions into which pixel values are written are
        determined by the currently active Cartesian coordinate
        axes. Exactly three axes must be active in order for this method
        to work. The three indices of the given pixel array correspond
        to these axes, where the order of the indices from left to right
        is from the highest to the lowest dimension. The given
        coordinates indicate the starting position of the given pixel
        array, relative to the image. That is to say, they determine the
        starting (x,y,z,t) position in the image for writing pixel
        values. Note that the coordinates corresponding to the active
        axes are allowed to be negative. However, pixel positions
        outside the image in the active dimensions are not actually
        written but simply skipped. This is useful for ignoring possible
        borders around the pixel data in the given array.

	@exception IllegalStateException if not exactly three axes are
	active.

	@exception ArrayIndexOutOfBoundsException if the coordinate
        corresponding to the non-active axis is outside the range
        determined by the size of the image.

	@exception NullPointerException if any of the parameters is
	<code>null</code>.
    */
    public void set(final Coordinates coords, final byte[][][] values) {

	switch (axes) {

	case Axes.ZYX: {
	    int vxstart = 0;
	    int vystart = 0;
	    int vzstart = 0;
	    int pxstart = coords.x;
	    int pystart = coords.y;
	    int pzstart = coords.z;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    int pxstop = coords.x + values[0][0].length;
	    int pystop = coords.y + values[0].length;
	    int pzstop = coords.z + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pystop > ydim) pystop = ydim;
	    if (pzstop > zdim) pzstop = zdim;
	    final int pxy = pystart*xdim + pxstart;
	    for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		    for (int x=pxstart, px=py, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
			pixels[coords.t][pz][px] = values[vz][vy][vx];
	    break;
	}
	case Axes.TZX: {
	    int vxstart = 0;
	    int vzstart = 0;
	    int vtstart = 0;
	    int pxstart = coords.x;
	    int pzstart = coords.z;
	    int ptstart = coords.t;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pxstop = coords.x + values[0][0].length;
	    int pzstop = coords.z + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pzstop > zdim) pzstop = zdim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = coords.y*xdim + pxstart;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		    for (int x=pxstart, px=pxy, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
			pixels[pt][pz][px] = values[vt][vz][vx];
	    break;
	}
	case Axes.TYX: {
	    int vxstart = 0;
	    int vystart = 0;
	    int vtstart = 0;
	    int pxstart = coords.x;
	    int pystart = coords.y;
	    int ptstart = coords.t;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pxstop = coords.x + values[0][0].length;
	    int pystop = coords.y + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pystop > ydim) pystop = ydim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = pystart*xdim + pxstart;
	    final int pz = coords.z;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		    for (int x=pxstart, px=py, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
			pixels[pt][pz][px] = values[vt][vy][vx];
	    break;
	}
	case Axes.TZY: {
	    int vystart = 0;
	    int vzstart = 0;
	    int vtstart = 0;
	    int pystart = coords.y;
	    int pzstart = coords.z;
	    int ptstart = coords.t;
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pystop = coords.y + values[0][0].length;
	    int pzstop = coords.z + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pystop > ydim) pystop = ydim;
	    if (pzstop > zdim) pzstop = zdim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = pystart*xdim + coords.x;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		    for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
			pixels[pt][pz][py] = values[vt][vz][vy];
	    break;
	}
	default: // wrong number of active axes
	    throw new IllegalStateException("Wrong number of active axes");
	}
    }

    /** Writes given pixel values in three dimensions into the
        image. The values are clipped to the range [0,255] and rounded
        to the nearest <code>byte</code> value. The dimensions into which
        pixel values are written are determined by the currently active
        Cartesian coordinate axes. Exactly three axes must be active in
        order for this method to work. The three indices of the given
        pixel array correspond to these axes, where the order of the
        indices from left to right is from the highest to the lowest
        dimension. The given coordinates indicate the starting position
        of the given pixel array, relative to the image. That is to say,
        they determine the starting (x,y,z,t) position in the image for
        writing pixel values. Note that the coordinates corresponding to
        the active axes are allowed to be negative. However, pixel
        positions outside the image in the active dimensions are not
        actually written but simply skipped. This is useful for ignoring
        possible borders around the pixel data in the given array.

	@exception IllegalStateException if not exactly three axes are
	active.

	@exception ArrayIndexOutOfBoundsException if the coordinate
        corresponding to the non-active axis is outside the range
        determined by the size of the image.

	@exception NullPointerException if any of the parameters is
	<code>null</code>.
    */
    public void set(final Coordinates coords, final float[][][] values) {

	switch (axes) {

	case Axes.ZYX: {
	    int vxstart = 0;
	    int vystart = 0;
	    int vzstart = 0;
	    int pxstart = coords.x;
	    int pystart = coords.y;
	    int pzstart = coords.z;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    int pxstop = coords.x + values[0][0].length;
	    int pystop = coords.y + values[0].length;
	    int pzstop = coords.z + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pystop > ydim) pystop = ydim;
	    if (pzstop > zdim) pzstop = zdim;
	    final int pxy = pystart*xdim + pxstart;
	    for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		    for (int x=pxstart, px=py, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
			pixels[coords.t][pz][px] = (byte)FMath.round((values[vz][vy][vx] > 255.0f) ? 255.0f : ((values[vz][vy][vx] < 0.0f) ? 0.0f : values[vz][vy][vx]));
	    break;
	}
	case Axes.TZX: {
	    int vxstart = 0;
	    int vzstart = 0;
	    int vtstart = 0;
	    int pxstart = coords.x;
	    int pzstart = coords.z;
	    int ptstart = coords.t;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pxstop = coords.x + values[0][0].length;
	    int pzstop = coords.z + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pzstop > zdim) pzstop = zdim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = coords.y*xdim + pxstart;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		    for (int x=pxstart, px=pxy, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
			pixels[pt][pz][px] = (byte)FMath.round((values[vt][vz][vx] > 255.0f) ? 255.0f : ((values[vt][vz][vx] < 0.0f) ? 0.0f : values[vt][vz][vx]));
	    break;
	}
	case Axes.TYX: {
	    int vxstart = 0;
	    int vystart = 0;
	    int vtstart = 0;
	    int pxstart = coords.x;
	    int pystart = coords.y;
	    int ptstart = coords.t;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pxstop = coords.x + values[0][0].length;
	    int pystop = coords.y + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pystop > ydim) pystop = ydim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = pystart*xdim + pxstart;
	    final int pz = coords.z;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		    for (int x=pxstart, px=py, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
			pixels[pt][pz][px] = (byte)FMath.round((values[vt][vy][vx] > 255.0f) ? 255.0f : ((values[vt][vy][vx] < 0.0f) ? 0.0f : values[vt][vy][vx]));
	    break;
	}
	case Axes.TZY: {
	    int vystart = 0;
	    int vzstart = 0;
	    int vtstart = 0;
	    int pystart = coords.y;
	    int pzstart = coords.z;
	    int ptstart = coords.t;
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pystop = coords.y + values[0][0].length;
	    int pzstop = coords.z + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pystop > ydim) pystop = ydim;
	    if (pzstop > zdim) pzstop = zdim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = pystart*xdim + coords.x;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		    for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
			pixels[pt][pz][py] = (byte)FMath.round((values[vt][vz][vy] > 255.0f) ? 255.0f : ((values[vt][vz][vy] < 0.0f) ? 0.0f : values[vt][vz][vy]));
	    break;
	}
	default: // wrong number of active axes
	    throw new IllegalStateException("Wrong number of active axes");
	}
    }

    /** Writes given pixel values in all four dimensions into the
        image. All four axes must be active in order for this method to
        work. The four indices of the given pixel array correspond to
        these axes, where the order of the indices from left to right is
        t, z, y, x. The given coordinates indicate the starting position
        of the given pixel array, relative to the image. That is to say,
        they determine the starting (x,y,z,t) position in the image for
        writing pixel values. Note that the coordinates are allowed to
        be negative. However, pixel positions outside the image are not
        actually written but simply skipped. This is useful for ignoring
        possible borders around the pixel data in the given array.

	@exception IllegalStateException if not all four axes are active.

	@exception NullPointerException if any of the parameters is
	<code>null</code>.
    */
    public void set(final Coordinates coords, final byte[][][][] values) {

	switch (axes) {

	case Axes.TZYX: {
	    int vxstart = 0;
	    int vystart = 0;
	    int vzstart = 0;
	    int vtstart = 0;
	    int pxstart = coords.x;
	    int pystart = coords.y;
	    int pzstart = coords.z;
	    int ptstart = coords.t;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pxstop = coords.x + values[0][0][0].length;
	    int pystop = coords.y + values[0][0].length;
	    int pzstop = coords.z + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pystop > ydim) pystop = ydim;
	    if (pzstop > zdim) pzstop = zdim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = pystart*xdim + pxstart;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		    for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
			for (int x=pxstart, px=py, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
			    pixels[pt][pz][px] = values[vt][vz][vy][vx];
	    break;
	}
	default: // wrong number of active axes
	    throw new IllegalStateException("Wrong number of active axes");
	}
    }

    /** Writes given pixel values in all four dimensions into the
        image. The values are clipped to the range [0,255] and rounded
        to the nearest <code>byte</code> value. All four axes must be active
        in order for this method to work. The four indices of the given
        pixel array correspond to these axes, where the order of the
        indices from left to right is t, z, y, x. The given coordinates
        indicate the starting position of the given pixel array,
        relative to the image. That is to say, they determine the
        starting (x,y,z,t) position in the image for writing pixel
        values. Note that the coordinates are allowed to be
        negative. However, pixel positions outside the image are not
        actually written but simply skipped. This is useful for ignoring
        possible borders around the pixel data in the given array.

	@exception IllegalStateException if not all four axes are active.

	@exception NullPointerException if any of the parameters is
	<code>null</code>.
    */
    public void set(final Coordinates coords, final float[][][][] values) {

	switch (axes) {

	case Axes.TZYX: {
	    int vxstart = 0;
	    int vystart = 0;
	    int vzstart = 0;
	    int vtstart = 0;
	    int pxstart = coords.x;
	    int pystart = coords.y;
	    int pzstart = coords.z;
	    int ptstart = coords.t;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pxstop = coords.x + values[0][0][0].length;
	    int pystop = coords.y + values[0][0].length;
	    int pzstop = coords.z + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pystop > ydim) pystop = ydim;
	    if (pzstop > zdim) pzstop = zdim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = pystart*xdim + pxstart;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		    for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
			for (int x=pxstart, px=py, vx=vxstart; x < pxstop; ++x, ++px, ++vx) {
			    final float value = values[vt][vz][vy][vx];
			    pixels[pt][pz][px] = (byte)FMath.round((value > 255.0f) ? 255.0f : ((value < 0.0f) ? 0.0f : value));
			}
	    break;
	}
	default: // wrong number of active axes
	    throw new IllegalStateException("Wrong number of active axes");
	}
    }

    /** Sets the pixels in the given borders by mirroring the remainder
	of the image. The size of the borders at the beginning and end
	in any dimension of the image is determined by the corresponding
	field in <code>borders</code>. The part of the image that results from
	ignoring the borders is held to contain the relevant data, which
	is mirrored into the borders.

	@exception IllegalArgumentException if any of the given border
	sizes is larger than half the image size minus one in the
	corresponding dimension.

	@exception NullPointerException if the parameter is
	<code>null</code>.
    */
    public void mirror(final Borders borders) {

	if (borders.x > (xdim-1)/2 || borders.y > (ydim-1)/2 ||
	    borders.z > (zdim-1)/2 || borders.t > (tdim-1)/2)
	    throw new IllegalArgumentException("Border(s) too large");

	// Initialize:
	final int xdimb = xdim - borders.x;
	final int ydimb = ydim - borders.y;
	final int zdimb = zdim - borders.z;
	final int tdimb = tdim - borders.t;

	final int remxdim = xdim - 2*borders.x;
	final int remydim = ydim - 2*borders.y;
	final int remzdim = zdim - 2*borders.z;
	final int remtdim = tdim - 2*borders.t;

	final int remxdim1 = (remxdim == 1) ? remxdim : (remxdim - 1);
	final int remydim1 = (remydim == 1) ? remydim : (remydim - 1);
	final int remzdim1 = (remzdim == 1) ? remzdim : (remzdim - 1);
	final int remtdim1 = (remtdim == 1) ? remtdim : (remtdim - 1);

	final int ifx = (remxdim == 1) ? 1 : 2;
	final int ify = (remydim == 1) ? 1 : 2;
	final int ifz = (remzdim == 1) ? 1 : 2;
	final int ift = (remtdim == 1) ? 1 : 2;

	final int xydim = xdim*ydim;

	// Mirror borders in x-dimension:
	final int byxdim = borders.y*xdim;
	for (int t=borders.t; t<tdimb; ++t)
	    for (int z=borders.z; z<zdimb; ++z) {
		for (int x=0; x<borders.x; ++x) {
		    final int xdiff = x - borders.x;
		    int x0 = xdiff/remxdim1; x0 += x0%ifx;
		    final int xmap = borders.x + Math.abs(xdiff - x0*remxdim1);
		    for (int y=borders.y, xy=byxdim+x, xymap=byxdim+xmap; y<ydimb; ++y, xy+=xdim, xymap+=xdim)
			pixels[t][z][xy] = pixels[t][z][xymap];
		}
		for (int x=xdimb; x<xdim; ++x) {
		    final int xdiff = x - borders.x;
		    int x0 = xdiff/remxdim1; x0 += x0%ifx;
		    final int xmap = borders.x + Math.abs(xdiff - x0*remxdim1);
		    for (int y=borders.y, xy=byxdim+x, xymap=byxdim+xmap; y<ydimb; ++y, xy+=xdim, xymap+=xdim)
			pixels[t][z][xy] = pixels[t][z][xymap];
		}
	    }

	// Mirror borders in y-dimension:
	for (int t=borders.t; t<tdimb; ++t)
	    for (int z=borders.z; z<zdimb; ++z) {
		for (int y=0; y<borders.y; ++y) {
		    final int ydiff = y - borders.y;
		    int y0 = ydiff/remydim1; y0 += y0%ify;
		    final int ymap = borders.y + Math.abs(ydiff - y0*remydim1);
		    for (int x=0, xy=y*xdim, xymap=ymap*xdim; x<xdim; ++x, ++xy, ++xymap)
			pixels[t][z][xy] = pixels[t][z][xymap];
		}
		for (int y=ydimb; y<ydim; ++y) {
		    final int ydiff = y - borders.y;
		    int y0 = ydiff/remydim1; y0 += y0%ify;
		    final int ymap = borders.y + Math.abs(ydiff - y0*remydim1);
		    for (int x=0, xy=y*xdim, xymap=ymap*xdim; x<xdim; ++x, ++xy, ++xymap)
			pixels[t][z][xy] = pixels[t][z][xymap];
		}
	    }

	// Mirror borders in z-dimension:
	for (int t=borders.t; t<tdimb; ++t) {
	    for (int z=0; z<borders.z; ++z) {
		final int zdiff = z - borders.z;
		int z0 = zdiff/remzdim1; z0 += z0%ifz;
		final int zmap = borders.z + Math.abs(zdiff - z0*remzdim1);
		for (int xy=0; xy<xydim; ++xy)
		    pixels[t][z][xy] = pixels[t][zmap][xy];
	    }
	    for (int z=zdimb; z<zdim; ++z) {
		final int zdiff = z - borders.z;
		int z0 = zdiff/remzdim1; z0 += z0%ifz;
		final int zmap = borders.z + Math.abs(zdiff - z0*remzdim1);
		for (int xy=0; xy<xydim; ++xy)
		    pixels[t][z][xy] = pixels[t][zmap][xy];
	    }
	}

	// Mirror borders in t-dimension:
	for (int t=0; t<borders.t; ++t) {
	    final int tdiff = t - borders.t;
	    int t0 = tdiff/remtdim1; t0 += t0%ift;
	    final int tmap = borders.t + Math.abs(tdiff - t0*remtdim1);
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy)
		    pixels[t][z][xy] = pixels[tmap][z][xy];
	}
	for (int t=tdimb; t<tdim; ++t) {
	    final int tdiff = t - borders.t;
	    int t0 = tdiff/remtdim1; t0 += t0%ift;
	    final int tmap = borders.t + Math.abs(tdiff - t0*remtdim1);
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy)
		    pixels[t][z][xy] = pixels[tmap][z][xy];
	}
    }
    
} 
