// **********************************************************************
// COPYRIGHT NOTICE AND DISCLAIMER
//
// Copyright (c) Erik Meijering. All rights reserved.
//
// Permission to copy and use this software and accompanying
// documentation for educational, research, and not-for-profit
// purposes, without fee and without a signed licensing agreement,
// is hereby granted, provided that the above copyright notice, this
// paragraph and the following two paragraphs appear in all copies.
// The copyright holder is free to make upgraded or improved versions
// of the software available for a fee or commercially only.
//
// IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE TO ANY PARTY FOR
// DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
// OF ANY KIND WHATSOEVER, ARISING OUT OF THE USE OF THIS SOFTWARE
// AND ITS DOCUMENTATION, EVEN IF HE HAS BEEN ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//
// THE COPYRIGHT HOLDER SPECIFICALLY DISCLAIMS ANY WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
// SOFTWARE AND ACCOMPANYING DOCUMENTATION IS PROVIDED "AS IS".
// THE COPYRIGHT HOLDER HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
// SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
//
// **********************************************************************

package imagescience.images;

import ij.ImageStack;
import imagescience.utilities.FMath;

/** Represents images of up to four dimensions containing data of type
    <code>int</code>. The data are always taken as aRGB-color values,
    with the blue component in bits 0-7 (the first, least significant
    byte), the green component in bits 8-15 (the second byte), the red
    component in bits 16-23 (the third byte), and the alpha value in
    bits 24-31 (the fourth, most significant byte). Each of the
    components is always in the range [0,255]. */
public final class ColorImage extends Image {

    // Components:
    public final static byte ALPHA = 0;
    public final static byte RED = 1;
    public final static byte GREEN = 2;
    public final static byte BLUE = 3;

    // Pixel array:
    final int[][][] pixels;

    // Component parameters:
    byte component = RED;
    int getmask = 0x00FF0000;
    int setmask = 0xFF00FFFF;
    int bitshift = 16;

    /** Constructor for creating a one-dimensional image of <code>int</code>s.

	@exception IllegalArgumentException if the parameter is zero or negative.
    */
    public ColorImage(final int xdim) {
	
	super(xdim,1,1,1);

	pixels = new int[1][1][xdim];
    }

    /** Constructor for creating a two-dimensional image of <code>int</code>s.

	@exception IllegalArgumentException if any of the parameters is zero or negative.
    */
    public ColorImage(final int xdim, final int ydim) {
	
	super(xdim,ydim,1,1);

	pixels = new int[1][1][xdim*ydim];
    }

    /** Constructor for creating a three-dimensional image of <code>int</code>s.

	@exception IllegalArgumentException if any of the parameters is zero or negative.
    */
    public ColorImage(final int xdim, final int ydim, final int zdim) {
	
	super(xdim,ydim,zdim,1);

	pixels = new int[1][zdim][xdim*ydim];
    }

    /** Constructor for creating a four-dimensional image of <code>int</code>s.

	@exception IllegalArgumentException if any of the parameters is zero or negative.
    */
    public ColorImage(final int xdim, final int ydim, final int zdim, final int tdim) {
	
	super(xdim,ydim,zdim,tdim);

	pixels = new int[tdim][zdim][xdim*ydim];
    }

    /** Constructor for creating a four-dimensional image of
	<code>int</code>s. The size of the image in the four dimensions
	is determined by the field values of the parameter.

	@exception NullPointerException if the parameter is <code>null</code>.
    */
    public ColorImage(final Dimensions dims) {
	
	super(dims.x,dims.y,dims.z,dims.t);

	pixels = new int[tdim][zdim][xdim*ydim];
    }

    /** Copy constructor. All pixel data and corresponding information
        contained in the given image is copied, and nothing in the
        resulting image is shared with the given image.

	@exception NullPointerException if the parameter is <code>null</code>.
    */
    public ColorImage(final ColorImage image) {
	
	super(image.xdim,image.ydim,image.zdim,image.tdim);

	pixels = new int[tdim][zdim][xdim*ydim];

	final int xydim = xdim*ydim;
	for (int t=0; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy)
		    pixels[t][z][xy] = image.pixels[t][z][xy];

	axes = image.axes;

	xpixdim = image.xpixdim;
	ypixdim = image.ypixdim;
	zpixdim = image.zpixdim;
	tpixdim = image.tpixdim;

	component = image.component;
	getmask = image.getmask;
	setmask = image.setmask;
	bitshift = image.bitshift;
    }

    /** Constructor for creating a three-dimensional image wrapper
        object. The size information is copied from the given
        <code>ImageStack</code> object. The actual pixel data, however,
        is not copied to, but shared with the newly created image.

	@exception IllegalArgumentException if the data stored in
	<code>stack</code> is not of type <code>int</code>.

	@exception NullPointerException if the parameter is
	<code>null</code>.
    */
    public ColorImage(final ImageStack stack) {
	
	super(stack.getWidth(),stack.getHeight(),stack.getSize(),1);

	final Object[] spixels = stack.getImageArray();
	if (!(spixels[0] instanceof int[]))
	    throw new IllegalArgumentException("Wrong pixel type");

	pixels = new int[tdim][zdim][];
	for (int z=0; z<zdim; ++z)
	    pixels[0][z] = (int[])spixels[z];
    }

    /** Constructor for creating a four-dimensional image wrapper
        object. The size information is copied from the given
        <code>ImageStack</code> object. The actual pixel data, however,
        is not copied to, but shared with the newly created image.

	@exception IllegalArgumentException if the data stored in
	<code>stack</code> is not of type <code>int</code>, or if the
	number of slices in <code>stack</code> is not an integral
	multiple of the value of the second parameter.

	@exception NullPointerException if the first parameter is
	<code>null</code>.
    */
    public ColorImage(final ImageStack stack, final int slicespervolume) {

	super(stack.getWidth(),stack.getHeight(),slicespervolume,stack.getSize()/slicespervolume);

	if (zdim*tdim != stack.getSize())
	    throw new IllegalArgumentException("Incorrect number of slices per volume");

	final Object[] spixels = stack.getImageArray();
	if (!(spixels[0] instanceof int[]))
	    throw new IllegalArgumentException("Wrong pixel type");

	pixels = new int[tdim][zdim][];
	for (int t=0, slice=0; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z, ++slice)
		pixels[t][z] = (int[])spixels[slice];
    }

    /** Returns a handle to the internal pixel array of the current
        image. In the current implementation, this is a
        three-dimensional array, where the first two indices from the
        left correspond to the t- and z-dimension, respectively, while
        the right-most index addresses the x-y-dimension, containing the
        sequence of pixel rows. One should be cautious in using this
        method, as the organization of the internal array containing the
        actual pixel data may change in future versions of the class. */
    public int[][][] array() { return pixels; }

    public ImageStack stack() {

	ImageStack stack = new ImageStack(xdim,ydim);

	for (int t=0; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z)
		stack.addSlice("",pixels[t][z]);

	return stack;
    }

    /** Returns an exact copy of the image. That is to say, all pixel
        data and corresponding information contained in the object is
        duplicated, and nothing in the returned object is shared with
        the object. */
    public ColorImage duplicate() {

	ColorImage dupi = new ColorImage(xdim,ydim,zdim,tdim);

	final int xydim = xdim*ydim;
	for (int t=0; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy)
		    dupi.pixels[t][z][xy] = pixels[t][z][xy];

	dupi.axes = axes;

	dupi.xpixdim = xpixdim;
	dupi.ypixdim = ypixdim;
	dupi.zpixdim = zpixdim;
	dupi.tpixdim = tpixdim;

	dupi.component = component;
	dupi.getmask = getmask;
	dupi.setmask = setmask;
	dupi.bitshift = bitshift;

	return dupi;
    }

    public int memory() { return (4*xdim*ydim*zdim*tdim)/1024; }

    /** Returns the active component. That is,
	<code>ColorImage.RED</code>, <code>ColorImage.GREEN</code>,
	<code>ColorImage.BLUE</code>, or
	<code>ColorImage.ALPHA</code>. */
    public byte component() { return component; }

    /** Sets the active component. The argument must be one of
	<code>ColorImage.RED</code> (the active component at
	construction), <code>ColorImage.GREEN</code>,
	<code>ColorImage.BLUE</code>, or
	<code>ColorImage.ALPHA</code>. Any method with a
	<code>float</code> (array) argument or return value will
	consider or change only the currently active component.

	@exception IllegalArgumentException if <code>component</code> is
	not one of the mentioned values.
    */
    public void component(final byte component) {
	
	this.component = component;

	switch (component) {

	case ALPHA:
	    getmask = 0xFF000000;
	    setmask = 0x00FFFFFF;
	    bitshift = 24;
	    break;

	case RED:
	    getmask = 0x00FF0000;
	    setmask = 0xFF00FFFF;
	    bitshift = 16;
	    break;

	case GREEN:
	    getmask = 0x0000FF00;
	    setmask = 0xFFFF00FF;
	    bitshift = 8;
	    break;

	case BLUE:
	    getmask = 0x000000FF;
	    setmask = 0xFFFFFF00;
	    bitshift = 0;
	    break;

	default:
	    throw new IllegalArgumentException("Non-supported component");
	}
    }

    /** Returns a floating-point representation of the minimum value in
        the image. Note that this method considers only the values of
        the currently active component, which are always in the range
        [0,255]. */
    public float minimum() {

	final int xydim = xdim*ydim;
	float min = Float.MAX_VALUE;
	for (int t=0; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy) {
		    final float value = (getmask & pixels[t][z][xy]) >> bitshift;
		    if (value < min) min = value;
		}

	return min;
    }

    /** Returns a floating-point representation of the maximum value in
        the image. Note that this method considers only the values of
        the currently active component, which are always in the range
        [0,255]. */
    public float maximum() {

	final int xydim = xdim*ydim;
	float max = -Float.MAX_VALUE;
	for (int t=0; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy) {
		    final float value = (getmask & pixels[t][z][xy]) >> bitshift;
		    if (value > max) max = value;
		}

	return max;
    }

    /** Returns the currently active component as a
        <code>ByteImage</code> object. */
    public ByteImage get() {

	final ByteImage c = new ByteImage(xdim,ydim,zdim,tdim);

	final int xydim = xdim*ydim;
	for (int t=0; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy)
		    c.pixels[t][z][xy] = (byte)((getmask & pixels[t][z][xy]) >> bitshift);

	return c;
    }

    /** Returns the pixel value at the given coordinates. Note
        that this method considers only the value of the currently
        active component, which is always in the range [0,255].

	@exception ArrayIndexOutOfBoundsException if any of the
	coordinates is out of range.

	@exception NullPointerException if the parameter is <code>null</code>.
    */
    public float get(final Coordinates coords) {

	return (getmask & pixels[coords.t][coords.z][xdim*coords.y + coords.x]) >> bitshift;
    }

    /** Reads pixel values in one dimension into the given array. The
        dimension from which pixel values are read is determined by the
        currently active Cartesian coordinate axis. Exactly one axis
        must be active in order for this method to work. The given
        coordinates indicate the starting position of the given pixel
        array, relative to the image. That is to say, they determine the
        starting (x,y,z,t) position in the image for reading pixel
        values into the array. Note that the coordinate corresponding to
        the active axis is allowed to be negative. However, pixel
        positions outside the image in the active dimension are not
        actually read but simply skipped. This is useful for creating
        borders around the pixel data read into the given array.

	@exception IllegalStateException if more than one coordinate
	axis is active.

	@exception ArrayIndexOutOfBoundsException if the coordinates
        corresponding to the non-active axes are outside the range
        determined by the size of the image.

	@exception NullPointerException if any of the parameters is
	<code>null</code>.
    */
    public void get(final Coordinates coords, final int[] values) {

	switch (axes) {

	case Axes.X: {
	    int vxstart = 0;
	    int pxstart = coords.x;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    int pxstop = coords.x + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    for (int x=pxstart, px=coords.y*xdim + pxstart, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
		values[vx] = pixels[coords.t][coords.z][px];
	    break;
	}
	case Axes.Y: {
	    int vystart = 0;
	    int pystart = coords.y;
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    int pystop = coords.y + values.length;
	    if (pystop > ydim) pystop = ydim;
	    for (int y=pystart, py=pystart*xdim + coords.x, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		values[vy] = pixels[coords.t][coords.z][py];
	    break;
	}
	case Axes.Z: {
	    int vzstart = 0;
	    int pzstart = coords.z;
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    int pzstop = coords.z + values.length;
	    if (pzstop > zdim) pzstop = zdim;
	    final int pxy = coords.y*xdim + coords.x;
	    for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		values[vz] = pixels[coords.t][pz][pxy];
	    break;
	}
	case Axes.T: {
	    int vtstart = 0;
	    int ptstart = coords.t;
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int ptstop = coords.t + values.length;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pz = coords.z;
	    final int pxy = coords.y*xdim + coords.x;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		values[vt] = pixels[pt][pz][pxy];
	    break;
	}
	default: // wrong number of active axes
	    throw new IllegalStateException("Wrong number of active axes");
	}
    }

    /** Reads pixel values in one dimension into the given array. Note
        that this method considers only the values of the currently
        active component, which are always in the range [0,255]. The
        dimension from which pixel values are read is determined by the
        currently active Cartesian coordinate axis. Exactly one axis
        must be active in order for this method to work. The given
        coordinates indicate the starting position of the given pixel
        array, relative to the image. That is to say, they determine the
        starting (x,y,z,t) position in the image for reading pixel
        values into the array. Note that the coordinate corresponding to
        the active axis is allowed to be negative. However, pixel
        positions outside the image in the active dimension are not
        actually read but simply skipped. This is useful for creating
        borders around the pixel data read into the given array.

	@exception IllegalStateException if more than one coordinate
	axis is active.

	@exception ArrayIndexOutOfBoundsException if the coordinates
        corresponding to the non-active axes are outside the range
        determined by the size of the image.

	@exception NullPointerException if any of the parameters is
	<code>null</code>.
    */
    public void get(final Coordinates coords, final float[] values) {

	switch (axes) {

	case Axes.X: {
	    int vxstart = 0;
	    int pxstart = coords.x;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    int pxstop = coords.x + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    for (int x=pxstart, px=coords.y*xdim + pxstart, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
		values[vx] = (getmask & pixels[coords.t][coords.z][px]) >> bitshift;
	    break;
	}
	case Axes.Y: {
	    int vystart = 0;
	    int pystart = coords.y;
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    int pystop = coords.y + values.length;
	    if (pystop > ydim) pystop = ydim;
	    for (int y=pystart, py=pystart*xdim + coords.x, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		values[vy] = (getmask & pixels[coords.t][coords.z][py]) >> bitshift;
	    break;
	}
	case Axes.Z: {
	    int vzstart = 0;
	    int pzstart = coords.z;
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    int pzstop = coords.z + values.length;
	    if (pzstop > zdim) pzstop = zdim;
	    final int pxy = coords.y*xdim + coords.x;
	    for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		values[vz] = (getmask & pixels[coords.t][pz][pxy]) >> bitshift;
	    break;
	}
	case Axes.T: {
	    int vtstart = 0;
	    int ptstart = coords.t;
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int ptstop = coords.t + values.length;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pz = coords.z;
	    final int pxy = coords.y*xdim + coords.x;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		values[vt] = (getmask & pixels[pt][pz][pxy]) >> bitshift;
	    break;
	}
	default: // wrong number of active axes
	    throw new IllegalStateException("Wrong number of active axes");
	}
    }

    /** Reads pixel values in two dimensions into the given array. The
        dimensions from which pixel values are read are determined by
        the currently active Cartesian coordinate axes. Exactly two axes
        must be active in order for this method to work. The two indices
        of the given pixel array correspond to these axes, where the
        order of the indices from left to right is from the highest to
        the lowest dimension. The given coordinates indicate the
        starting position of the given pixel array, relative to the
        image. That is to say, they determine the starting (x,y,z,t)
        position in the image for reading pixel values into the
        array. Note that the coordinates corresponding to the active
        axes are allowed to be negative. However, pixel positions
        outside the image in the active dimensions are not actually read
        but simply skipped. This is useful for creating borders around
        the pixel data read into the given array.

	@exception IllegalStateException if not exactly two axes are active.

	@exception ArrayIndexOutOfBoundsException if the coordinates
        corresponding to the non-active axes are outside the range
        determined by the size of the image.

	@exception NullPointerException if any of the parameters is
	<code>null</code>.
    */
    public void get(final Coordinates coords, final int[][] values) {

	switch (axes) {

	case Axes.YX: {
	    int vxstart = 0;
	    int vystart = 0;
	    int pxstart = coords.x;
	    int pystart = coords.y;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    int pxstop = coords.x + values[0].length;
	    int pystop = coords.y + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pystop > ydim) pystop = ydim;
	    for (int y=pystart, py=pystart*xdim + pxstart, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		for (int x=pxstart, px=py, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
		    values[vy][vx] = pixels[coords.t][coords.z][px];
	    break;
	}
	case Axes.ZX: {
	    int vxstart = 0;
	    int vzstart = 0;
	    int pxstart = coords.x;
	    int pzstart = coords.z;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    int pxstop = coords.x + values[0].length;
	    int pzstop = coords.z + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pzstop > zdim) pzstop = zdim;
	    final int pxy = coords.y*xdim + pxstart;
	    for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		for (int x=pxstart, px=pxy, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
		    values[vz][vx] = pixels[coords.t][pz][px];
	    break;
	}
	case Axes.TX: {
	    int vxstart = 0;
	    int vtstart = 0;
	    int pxstart = coords.x;
	    int ptstart = coords.t;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pxstop = coords.x + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = coords.y*xdim + pxstart;
	    final int pz = coords.z;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int x=pxstart, px=pxy, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
		    values[vt][vx] = pixels[pt][pz][px];
	    break;
	}
	case Axes.ZY: {
	    int vystart = 0;
	    int vzstart = 0;
	    int pystart = coords.y;
	    int pzstart = coords.z;
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    int pystop = coords.y + values[0].length;
	    int pzstop = coords.z + values.length;
	    if (pystop > ydim) pystop = ydim;
	    if (pzstop > zdim) pzstop = zdim;
	    final int pxy = pystart*xdim + coords.x;
	    for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		    values[vz][vy] = pixels[coords.t][pz][py];
	    break;
	}
	case Axes.TY: {
	    int vystart = 0;
	    int vtstart = 0;
	    int pystart = coords.y;
	    int ptstart = coords.t;
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pystop = coords.y + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pystop > ydim) pystop = ydim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = pystart*xdim + coords.x;
	    final int pz = coords.z;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		    values[vt][vy] = pixels[pt][pz][py];
	    break;
	}
	case Axes.TZ: {
	    int vzstart = 0;
	    int vtstart = 0;
	    int pzstart = coords.z;
	    int ptstart = coords.t;
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pzstop = coords.z + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pzstop > zdim) pzstop = zdim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = coords.y*xdim + coords.x;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		    values[vt][vz] = pixels[pt][pz][pxy];
	    break;
	}
	default: // wrong number of active axes
	    throw new IllegalStateException("Wrong number of active axes");
	}
    }

    /** Reads pixel values in two dimensions into the given array. Note
        that this method considers only the values of the currently
        active component, which are always in the range [0,255]. The
        dimensions from which pixel values are read are determined by
        the currently active Cartesian coordinate axes. Exactly two axes
        must be active in order for this method to work. The two indices
        of the given pixel array correspond to these axes, where the
        order of the indices from left to right is from the highest to
        the lowest dimension. The given coordinates indicate the
        starting position of the given pixel array, relative to the
        image. That is to say, they determine the starting (x,y,z,t)
        position in the image for reading pixel values into the
        array. Note that the coordinates corresponding to the active
        axes are allowed to be negative. However, pixel positions
        outside the image in the active dimensions are not actually read
        but simply skipped. This is useful for creating borders around
        the pixel data read into the given array.

	@exception IllegalStateException if not exactly two axes are active.

	@exception ArrayIndexOutOfBoundsException if the coordinates
        corresponding to the non-active axes are outside the range
        determined by the size of the image.

	@exception NullPointerException if any of the parameters is
	<code>null</code>.
    */
    public void get(final Coordinates coords, final float[][] values) {

	switch (axes) {

	case Axes.YX: {
	    int vxstart = 0;
	    int vystart = 0;
	    int pxstart = coords.x;
	    int pystart = coords.y;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    int pxstop = coords.x + values[0].length;
	    int pystop = coords.y + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pystop > ydim) pystop = ydim;
	    for (int y=pystart, py=pystart*xdim + pxstart, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		for (int x=pxstart, px=py, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
		    values[vy][vx] = (getmask & pixels[coords.t][coords.z][px]) >> bitshift;
	    break;
	}
	case Axes.ZX: {
	    int vxstart = 0;
	    int vzstart = 0;
	    int pxstart = coords.x;
	    int pzstart = coords.z;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    int pxstop = coords.x + values[0].length;
	    int pzstop = coords.z + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pzstop > zdim) pzstop = zdim;
	    final int pxy = coords.y*xdim + pxstart;
	    for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		for (int x=pxstart, px=pxy, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
		    values[vz][vx] = (getmask & pixels[coords.t][pz][px]) >> bitshift;
	    break;
	}
	case Axes.TX: {
	    int vxstart = 0;
	    int vtstart = 0;
	    int pxstart = coords.x;
	    int ptstart = coords.t;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pxstop = coords.x + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = coords.y*xdim + pxstart;
	    final int pz = coords.z;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int x=pxstart, px=pxy, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
		    values[vt][vx] = (getmask & pixels[pt][pz][px]) >> bitshift;
	    break;
	}
	case Axes.ZY: {
	    int vystart = 0;
	    int vzstart = 0;
	    int pystart = coords.y;
	    int pzstart = coords.z;
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    int pystop = coords.y + values[0].length;
	    int pzstop = coords.z + values.length;
	    if (pystop > ydim) pystop = ydim;
	    if (pzstop > zdim) pzstop = zdim;
	    final int pxy = pystart*xdim + coords.x;
	    for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		    values[vz][vy] = (getmask & pixels[coords.t][pz][py]) >> bitshift;
	    break;
	}
	case Axes.TY: {
	    int vystart = 0;
	    int vtstart = 0;
	    int pystart = coords.y;
	    int ptstart = coords.t;
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pystop = coords.y + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pystop > ydim) pystop = ydim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = pystart*xdim + coords.x;
	    final int pz = coords.z;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		    values[vt][vy] = (getmask & pixels[pt][pz][py]) >> bitshift;
	    break;
	}
	case Axes.TZ: {
	    int vzstart = 0;
	    int vtstart = 0;
	    int pzstart = coords.z;
	    int ptstart = coords.t;
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pzstop = coords.z + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pzstop > zdim) pzstop = zdim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = coords.y*xdim + coords.x;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		    values[vt][vz] = (getmask & pixels[pt][pz][pxy]) >> bitshift;
	    break;
	}
	default: // wrong number of active axes
	    throw new IllegalStateException("Wrong number of active axes");
	}
    }

    /** Reads pixel values in three dimensions into the given array. The
        dimensions from which pixel values are read are determined by
        the currently active Cartesian coordinate axes. Exactly three
        axes must be active in order for this method to work. The three
        indices of the given pixel array correspond to these axes, where
        the order of the indices from left to right is from the highest
        to the lowest dimension. The given coordinates indicate the
        starting position of the given pixel array, relative to the
        image. That is to say, they determine the starting (x,y,z,t)
        position in the image for reading pixel values into the
        array. Note that the coordinates corresponding to the active
        axes are allowed to be negative. However, pixel positions
        outside the image in the active dimensions are not actually read
        but simply skipped. This is useful for creating borders around
        the pixel data read into the given array.

	@exception IllegalStateException if not exactly three axes are active.

	@exception ArrayIndexOutOfBoundsException if the coordinate
        corresponding to the non-active axis is outside the range
        determined by the size of the image.

	@exception NullPointerException if any of the parameters is
	<code>null</code>.
    */
    public void get(final Coordinates coords, final int[][][] values) {

	switch (axes) {

	case Axes.ZYX: {
	    int vxstart = 0;
	    int vystart = 0;
	    int vzstart = 0;
	    int pxstart = coords.x;
	    int pystart = coords.y;
	    int pzstart = coords.z;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    int pxstop = coords.x + values[0][0].length;
	    int pystop = coords.y + values[0].length;
	    int pzstop = coords.z + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pystop > ydim) pystop = ydim;
	    if (pzstop > zdim) pzstop = zdim;
	    final int pxy = pystart*xdim + pxstart;
	    for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		    for (int x=pxstart, px=py, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
			values[vz][vy][vx] = pixels[coords.t][pz][px];
	    break;
	}
	case Axes.TZX: {
	    int vxstart = 0;
	    int vzstart = 0;
	    int vtstart = 0;
	    int pxstart = coords.x;
	    int pzstart = coords.z;
	    int ptstart = coords.t;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pxstop = coords.x + values[0][0].length;
	    int pzstop = coords.z + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pzstop > zdim) pzstop = zdim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = coords.y*xdim + pxstart;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		    for (int x=pxstart, px=pxy, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
			values[vt][vz][vx] = pixels[pt][pz][px];
	    break;
	}
	case Axes.TYX: {
	    int vxstart = 0;
	    int vystart = 0;
	    int vtstart = 0;
	    int pxstart = coords.x;
	    int pystart = coords.y;
	    int ptstart = coords.t;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pxstop = coords.x + values[0][0].length;
	    int pystop = coords.y + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pystop > ydim) pystop = ydim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = pystart*xdim + pxstart;
	    final int pz = coords.z;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		    for (int x=pxstart, px=py, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
			values[vt][vy][vx] = pixels[pt][pz][px];
	    break;
	}
	case Axes.TZY: {
	    int vystart = 0;
	    int vzstart = 0;
	    int vtstart = 0;
	    int pystart = coords.y;
	    int pzstart = coords.z;
	    int ptstart = coords.t;
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pystop = coords.y + values[0][0].length;
	    int pzstop = coords.z + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pystop > ydim) pystop = ydim;
	    if (pzstop > zdim) pzstop = zdim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = pystart*xdim + coords.x;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		    for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
			values[vt][vz][vy] = pixels[pt][pz][py];
	    break;
	}
	default: // wrong number of active axes
	    throw new IllegalStateException("Wrong number of active axes");
	}
    }

    /** Reads pixel values in three dimensions into the given
        array. Note that this method considers only the values of the
        currently active component, which are always in the range
        [0,255]. The dimensions from which pixel values are read are
        determined by the currently active Cartesian coordinate
        axes. Exactly three axes must be active in order for this method
        to work. The three indices of the given pixel array correspond
        to these axes, where the order of the indices from left to right
        is from the highest to the lowest dimension. The given
        coordinates indicate the starting position of the given pixel
        array, relative to the image. That is to say, they determine the
        starting (x,y,z,t) position in the image for reading pixel
        values into the array. Note that the coordinates corresponding
        to the active axes are allowed to be negative. However, pixel
        positions outside the image in the active dimensions are not
        actually read but simply skipped. This is useful for creating
        borders around the pixel data read into the given array.

	@exception IllegalStateException if not exactly three axes are active.

	@exception ArrayIndexOutOfBoundsException if the coordinate
        corresponding to the non-active axis is outside the range
        determined by the size of the image.

	@exception NullPointerException if any of the parameters is
	<code>null</code>.
    */
    public void get(final Coordinates coords, final float[][][] values) {

	switch (axes) {

	case Axes.ZYX: {
	    int vxstart = 0;
	    int vystart = 0;
	    int vzstart = 0;
	    int pxstart = coords.x;
	    int pystart = coords.y;
	    int pzstart = coords.z;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    int pxstop = coords.x + values[0][0].length;
	    int pystop = coords.y + values[0].length;
	    int pzstop = coords.z + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pystop > ydim) pystop = ydim;
	    if (pzstop > zdim) pzstop = zdim;
	    final int pxy = pystart*xdim + pxstart;
	    for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		    for (int x=pxstart, px=py, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
			values[vz][vy][vx] = (getmask & pixels[coords.t][pz][px]) >> bitshift;
	    break;
	}
	case Axes.TZX: {
	    int vxstart = 0;
	    int vzstart = 0;
	    int vtstart = 0;
	    int pxstart = coords.x;
	    int pzstart = coords.z;
	    int ptstart = coords.t;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pxstop = coords.x + values[0][0].length;
	    int pzstop = coords.z + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pzstop > zdim) pzstop = zdim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = coords.y*xdim + pxstart;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		    for (int x=pxstart, px=pxy, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
			values[vt][vz][vx] = (getmask & pixels[pt][pz][px]) >> bitshift;
	    break;
	}
	case Axes.TYX: {
	    int vxstart = 0;
	    int vystart = 0;
	    int vtstart = 0;
	    int pxstart = coords.x;
	    int pystart = coords.y;
	    int ptstart = coords.t;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pxstop = coords.x + values[0][0].length;
	    int pystop = coords.y + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pystop > ydim) pystop = ydim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = pystart*xdim + pxstart;
	    final int pz = coords.z;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		    for (int x=pxstart, px=py, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
			values[vt][vy][vx] = (getmask & pixels[pt][pz][px]) >> bitshift;
	    break;
	}
	case Axes.TZY: {
	    int vystart = 0;
	    int vzstart = 0;
	    int vtstart = 0;
	    int pystart = coords.y;
	    int pzstart = coords.z;
	    int ptstart = coords.t;
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pystop = coords.y + values[0][0].length;
	    int pzstop = coords.z + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pystop > ydim) pystop = ydim;
	    if (pzstop > zdim) pzstop = zdim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = pystart*xdim + coords.x;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		    for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
			values[vt][vz][vy] = (getmask & pixels[pt][pz][py]) >> bitshift;
	    break;
	}
	default: // wrong number of active axes
	    throw new IllegalStateException("Wrong number of active axes");
	}
    }

    /** Reads pixel values in all four dimensions into the given
        array. All four axes must be active in order for this method to
        work. The four indices of the given pixel array correspond to
        these axes, where the order of the indices from left to right is
        t, z, y, x. The given coordinates indicate the starting position
        of the given pixel array, relative to the image. That is to say,
        they determine the starting (x,y,z,t) position in the image for
        reading pixel values into the array. Note that the coordinates
        are allowed to be negative. However, pixel positions outside the
        image are not actually read but simply skipped. This is useful
        for creating borders around the pixel data read into the given
        array.

	@exception IllegalStateException if not all four axes are active.

	@exception NullPointerException if any of the parameters is
	<code>null</code>.
    */
    public void get(final Coordinates coords, final int[][][][] values) {

	switch (axes) {

	case Axes.TZYX: {
	    int vxstart = 0;
	    int vystart = 0;
	    int vzstart = 0;
	    int vtstart = 0;
	    int pxstart = coords.x;
	    int pystart = coords.y;
	    int pzstart = coords.z;
	    int ptstart = coords.t;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pxstop = coords.x + values[0][0][0].length;
	    int pystop = coords.y + values[0][0].length;
	    int pzstop = coords.z + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pystop > ydim) pystop = ydim;
	    if (pzstop > zdim) pzstop = zdim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = pystart*xdim + pxstart;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		    for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
			for (int x=pxstart, px=py, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
			    values[vt][vz][vy][vx] = pixels[pt][pz][px];
	    break;
	}
	default: // wrong number of active axes
	    throw new IllegalStateException("Wrong number of active axes");
	}
    }

    /** Reads pixel values in all four dimensions into the given
        array. Note that this method considers only the values of the
        currently active component, which are always in the range
        [0,255]. All four axes must be active in order for this method
        to work. The four indices of the given pixel array correspond to
        these axes, where the order of the indices from left to right is
        t, z, y, x. The given coordinates indicate the starting position
        of the given pixel array, relative to the image. That is to say,
        they determine the starting (x,y,z,t) position in the image for
        reading pixel values into the array. Note that the coordinates
        are allowed to be negative. However, pixel positions outside the
        image are not actually read but simply skipped. This is useful
        for creating borders around the pixel data read into the given
        array.

	@exception IllegalStateException if not all four axes are active.

	@exception NullPointerException if any of the parameters is
	<code>null</code>.
    */
    public void get(final Coordinates coords, final float[][][][] values) {

	switch (axes) {

	case Axes.TZYX: {
	    int vxstart = 0;
	    int vystart = 0;
	    int vzstart = 0;
	    int vtstart = 0;
	    int pxstart = coords.x;
	    int pystart = coords.y;
	    int pzstart = coords.z;
	    int ptstart = coords.t;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pxstop = coords.x + values[0][0][0].length;
	    int pystop = coords.y + values[0][0].length;
	    int pzstop = coords.z + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pystop > ydim) pystop = ydim;
	    if (pzstop > zdim) pzstop = zdim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = pystart*xdim + pxstart;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		    for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
			for (int x=pxstart, px=py, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
			    values[vt][vz][vy][vx] = (getmask & pixels[pt][pz][px]) >> bitshift;
	    break;
	}
	default: // wrong number of active axes
	    throw new IllegalStateException("Wrong number of active axes");
	}
    }

    /** Writes the pixel values in the given image to the currently
        active component.

	@exception IllegalArgumentException if the size of the given
	image is not equal to that of the current image in every
	dimension.

	@exception NullPointerException if <code>image</code> is
	<code>null</code>.
    */
    public void set(final ByteImage image) {

	if (xdim != image.xdim || ydim != image.ydim ||
	    zdim != image.zdim || tdim != image.tdim)
	    throw new IllegalArgumentException("Unequal image sizes");
	
	final int xydim = xdim*ydim;
	for (int t=0; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy)
		    pixels[t][z][xy] = (setmask & pixels[t][z][xy]) | ((0xFF & image.pixels[t][z][xy]) << bitshift);
    }

    /** Sets all pixels to the given value. */
    public void set(final int value) {

	final int xydim = xdim*ydim;
	for (int t=0; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy)
		    pixels[t][z][xy] = value;
    }

    /** Sets all pixels to the given value. Note that this method
        affects only the currently active component. The given value is
        clipped to the range [0,255], rounded to the nearest
        <code>int</code> value, and then replaces the active component. */
    public void set(final float value) {

	final int xydim = xdim*ydim;
	final int val = (FMath.round((value > 255.0f) ? 255.0f : ((value < 0.0f) ? 0.0f : value))) << bitshift;
	for (int t=0; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy)
		    pixels[t][z][xy] = (setmask & pixels[t][z][xy]) | val;
    }

    /** Sets the pixel at the given coordinates to the given value.

	@exception ArrayIndexOutOfBoundsException if any of the
	coordinates is out of range.

	@exception NullPointerException if the first parameter is
	<code>null</code>.
    */
    public void set(final Coordinates coords, final int value) {

	pixels[coords.t][coords.z][xdim*coords.y + coords.x] = value;
    }
    
    /** Sets the pixel at the given coordinates to the given value. Note
	that this method affects only the currently active component.
	The given value is clipped to the range [0,255], rounded to the
	nearest <code>int</code> value, and then replaces the active
	component.

	@exception ArrayIndexOutOfBoundsException if any of the
	coordinates is out of range.

	@exception NullPointerException if the first parameter is
	<code>null</code>.  */
    public void set(final Coordinates coords, final float value) {

	pixels[coords.t][coords.z][xdim*coords.y + coords.x] =
	    (setmask & pixels[coords.t][coords.z][xdim*coords.y + coords.x]) |
	    ((FMath.round((value > 255.0f) ? 255.0f : ((value < 0.0f) ? 0.0f : value))) << bitshift);
    }
    
    /** Writes given pixel values in one dimension into the image. The
        dimension into which pixel values are written is determined by
        the currently active Cartesian coordinate axis. Exactly one axis
        must be active in order for this method to work. The given
        coordinates indicate the starting position of the given pixel
        array, relative to the image. That is to say, they determine the
        starting (x,y,z,t) position in the image for writing pixel
        values. Note that the coordinate corresponding to the active
        axis is allowed to be negative. However, pixel positions outside
        the image in the active dimension are not actually written but
        simply skipped. This is useful for ignoring possible borders
        around the pixel data in the given array.

	@exception IllegalStateException if more than one coordinate
	axis is active.

	@exception ArrayIndexOutOfBoundsException if the coordinates
        corresponding to the non-active axes are outside the range
        determined by the size of the image.

	@exception NullPointerException if any of the parameters is
	<code>null</code>.
    */
    public void set(final Coordinates coords, final int[] values) {

	switch (axes) {

	case Axes.X: {
	    int vxstart = 0;
	    int pxstart = coords.x;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    int pxstop = coords.x + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    for (int x=pxstart, px=coords.y*xdim + pxstart, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
		pixels[coords.t][coords.z][px] = values[vx];
	    break;
	}
	case Axes.Y: {
	    int vystart = 0;
	    int pystart = coords.y;
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    int pystop = coords.y + values.length;
	    if (pystop > ydim) pystop = ydim;
	    for (int y=pystart, py=pystart*xdim + coords.x, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		pixels[coords.t][coords.z][py] = values[vy];
	    break;
	}
	case Axes.Z: {
	    int vzstart = 0;
	    int pzstart = coords.z;
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    int pzstop = coords.z + values.length;
	    if (pzstop > zdim) pzstop = zdim;
	    final int pxy = coords.y*xdim + coords.x;
	    for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		pixels[coords.t][pz][pxy] = values[vz];
	    break;
	}
	case Axes.T: {
	    int vtstart = 0;
	    int ptstart = coords.t;
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int ptstop = coords.t + values.length;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pz = coords.z;
	    final int pxy = coords.y*xdim + coords.x;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		pixels[pt][pz][pxy] = values[vt];
	    break;
	}
	default: // wrong number of active axes
	    throw new IllegalStateException("Wrong number of active axes");
	}
    }

    /** Writes given pixel values in one dimension into the image. Note
        that this method affects only the currently active
        component. The given pixel values are clipped to the range
        [0,255], rounded to the nearest <code>int</code> value, and then
        replace the active component. The dimension into which pixel
        values are written is determined by the currently active
        Cartesian coordinate axis. Exactly one axis must be active in
        order for this method to work. The given coordinates indicate
        the starting position of the given pixel array, relative to the
        image. That is to say, they determine the starting (x,y,z,t)
        position in the image for writing pixel values. Note that the
        coordinate corresponding to the active axis is allowed to be
        negative. However, pixel positions outside the image in the
        active dimension are not actually written but simply
        skipped. This is useful for ignoring possible borders around the
        pixel data in the given array.

	@exception IllegalStateException if more than one coordinate
	axis is active.

	@exception ArrayIndexOutOfBoundsException if the coordinates
        corresponding to the non-active axes are outside the range
        determined by the size of the image.

	@exception NullPointerException if any of the parameters is
	<code>null</code>.
    */
    public void set(final Coordinates coords, final float[] values) {

	switch (axes) {

	case Axes.X: {
	    int vxstart = 0;
	    int pxstart = coords.x;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    int pxstop = coords.x + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    for (int x=pxstart, px=coords.y*xdim + pxstart, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
		pixels[coords.t][coords.z][px] = (setmask & pixels[coords.t][coords.z][px]) |
		    (FMath.round((values[vx] > 255.0f) ? 255.0f : ((values[vx] < 0.0f) ? 0.0f : values[vx])) << bitshift);
	    break;
	}
	case Axes.Y: {
	    int vystart = 0;
	    int pystart = coords.y;
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    int pystop = coords.y + values.length;
	    if (pystop > ydim) pystop = ydim;
	    for (int y=pystart, py=pystart*xdim + coords.x, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		pixels[coords.t][coords.z][py] = (setmask & pixels[coords.t][coords.z][py]) |
		    (FMath.round((values[vy] > 255.0f) ? 255.0f : ((values[vy] < 0.0f) ? 0.0f : values[vy])) << bitshift);
	    break;
	}
	case Axes.Z: {
	    int vzstart = 0;
	    int pzstart = coords.z;
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    int pzstop = coords.z + values.length;
	    if (pzstop > zdim) pzstop = zdim;
	    final int pxy = coords.y*xdim + coords.x;
	    for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		pixels[coords.t][pz][pxy] = (setmask & pixels[coords.t][pz][pxy]) |
		    (FMath.round((values[vz] > 255.0f) ? 255.0f : ((values[vz] < 0.0f) ? 0.0f : values[vz])) << bitshift);
	    break;
	}
	case Axes.T: {
	    int vtstart = 0;
	    int ptstart = coords.t;
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int ptstop = coords.t + values.length;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pz = coords.z;
	    final int pxy = coords.y*xdim + coords.x;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		pixels[pt][pz][pxy] = (setmask & pixels[pt][pz][pxy]) |
		    (FMath.round((values[vt] > 255.0f) ? 255.0f : ((values[vt] < 0.0f) ? 0.0f : values[vt])) << bitshift);
	    break;
	}
	default: // wrong number of active axes
	    throw new IllegalStateException("Wrong number of active axes");
	}
    }

    /** Writes given pixel values in two dimensions into the image. The
        dimensions into which pixel values are written are determined by
        the currently active Cartesian coordinate axes. Exactly two axes
        must be active in order for this method to work. The two indices
        of the given pixel array correspond to these axes, where the
        order of the indices from left to right is from the highest to
        the lowest dimension. The given coordinates indicate the
        starting position of the given pixel array, relative to the
        image. That is to say, they determine the starting (x,y,z,t)
        position in the image for writing pixel values. Note that the
        coordinates corresponding to the active axes are allowed to be
        negative. However, pixel positions outside the image in the
        active dimensions are not actually written but simply
        skipped. This is useful for ignoring possible borders around the
        pixel data in the given array.

	@exception IllegalStateException if not exactly two axes are active.

	@exception ArrayIndexOutOfBoundsException if the coordinates
        corresponding to the non-active axes are outside the range
        determined by the size of the image.

	@exception NullPointerException if any of the parameters is
	<code>null</code>.
    */
    public void set(final Coordinates coords, final int[][] values) {
	
	switch (axes) {

	case Axes.YX: {
	    int vxstart = 0;
	    int vystart = 0;
	    int pxstart = coords.x;
	    int pystart = coords.y;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    int pxstop = coords.x + values[0].length;
	    int pystop = coords.y + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pystop > ydim) pystop = ydim;
	    for (int y=pystart, py=pystart*xdim + pxstart, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		for (int x=pxstart, px=py, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
		    pixels[coords.t][coords.z][px] = values[vy][vx];
	    break;
	}
	case Axes.ZX: {
	    int vxstart = 0;
	    int vzstart = 0;
	    int pxstart = coords.x;
	    int pzstart = coords.z;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    int pxstop = coords.x + values[0].length;
	    int pzstop = coords.z + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pzstop > zdim) pzstop = zdim;
	    final int pxy = coords.y*xdim + pxstart;
	    for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		for (int x=pxstart, px=pxy, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
		    pixels[coords.t][pz][px] = values[vz][vx];
	    break;
	}
	case Axes.TX: {
	    int vxstart = 0;
	    int vtstart = 0;
	    int pxstart = coords.x;
	    int ptstart = coords.t;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pxstop = coords.x + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = coords.y*xdim + pxstart;
	    final int pz = coords.z;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int x=pxstart, px=pxy, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
		    pixels[pt][pz][px] = values[vt][vx];
	    break;
	}
	case Axes.ZY: {
	    int vystart = 0;
	    int vzstart = 0;
	    int pystart = coords.y;
	    int pzstart = coords.z;
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    int pystop = coords.y + values[0].length;
	    int pzstop = coords.z + values.length;
	    if (pystop > ydim) pystop = ydim;
	    if (pzstop > zdim) pzstop = zdim;
	    final int pxy = pystart*xdim + coords.x;
	    for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		    pixels[coords.t][pz][py] = values[vz][vy];
	    break;
	}
	case Axes.TY: {
	    int vystart = 0;
	    int vtstart = 0;
	    int pystart = coords.y;
	    int ptstart = coords.t;
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pystop = coords.y + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pystop > ydim) pystop = ydim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = pystart*xdim + coords.x;
	    final int pz = coords.z;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		    pixels[pt][pz][py] = values[vt][vy];
	    break;
	}
	case Axes.TZ: {
	    int vzstart = 0;
	    int vtstart = 0;
	    int pzstart = coords.z;
	    int ptstart = coords.t;
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pzstop = coords.z + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pzstop > zdim) pzstop = zdim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = coords.y*xdim + coords.x;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		    pixels[pt][pz][pxy] = values[vt][vz];
	    break;
	}
	default: // wrong number of active axes
	    throw new IllegalStateException("Wrong number of active axes");
	}
    }

    /** Writes given pixel values in two dimensions into the image. Note
        that this method affects only the currently active
        component. The given pixel values are clipped to the range
        [0,255], rounded to the nearest <code>int</code> value, and then
        replace the active component. The dimensions into which pixel
        values are written are determined by the currently active
        Cartesian coordinate axes. Exactly two axes must be active in
        order for this method to work. The two indices of the given
        pixel array correspond to these axes, where the order of the
        indices from left to right is from the highest to the lowest
        dimension. The given coordinates indicate the starting position
        of the given pixel array, relative to the image. That is to say,
        they determine the starting (x,y,z,t) position in the image for
        writing pixel values. Note that the coordinates corresponding to
        the active axes are allowed to be negative. However, pixel
        positions outside the image in the active dimensions are not
        actually written but simply skipped. This is useful for ignoring
        possible borders around the pixel data in the given array.

	@exception IllegalStateException if not exactly two axes are active.

	@exception ArrayIndexOutOfBoundsException if the coordinates
        corresponding to the non-active axes are outside the range
        determined by the size of the image.

	@exception NullPointerException if any of the parameters is
	<code>null</code>.
    */
    public void set(final Coordinates coords, final float[][] values) {
	
	switch (axes) {

	case Axes.YX: {
	    int vxstart = 0;
	    int vystart = 0;
	    int pxstart = coords.x;
	    int pystart = coords.y;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    int pxstop = coords.x + values[0].length;
	    int pystop = coords.y + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pystop > ydim) pystop = ydim;
	    for (int y=pystart, py=pystart*xdim + pxstart, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		for (int x=pxstart, px=py, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
		    pixels[coords.t][coords.z][px] = (setmask & pixels[coords.t][coords.z][px]) |
			(FMath.round((values[vy][vx] > 255.0f) ? 255.0f : ((values[vy][vx] < 0.0f) ? 0.0f : values[vy][vx])) << bitshift);
	    break;
	}
	case Axes.ZX: {
	    int vxstart = 0;
	    int vzstart = 0;
	    int pxstart = coords.x;
	    int pzstart = coords.z;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    int pxstop = coords.x + values[0].length;
	    int pzstop = coords.z + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pzstop > zdim) pzstop = zdim;
	    final int pxy = coords.y*xdim + pxstart;
	    for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		for (int x=pxstart, px=pxy, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
		    pixels[coords.t][pz][px] = (setmask & pixels[coords.t][pz][px]) |
			(FMath.round((values[vz][vx] > 255.0f) ? 255.0f : ((values[vz][vx] < 0.0f) ? 0.0f : values[vz][vx])) << bitshift);
	    break;
	}
	case Axes.TX: {
	    int vxstart = 0;
	    int vtstart = 0;
	    int pxstart = coords.x;
	    int ptstart = coords.t;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pxstop = coords.x + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = coords.y*xdim + pxstart;
	    final int pz = coords.z;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int x=pxstart, px=pxy, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
		    pixels[pt][pz][px] = (setmask & pixels[pt][pz][px]) |
			(FMath.round((values[vt][vx] > 255.0f) ? 255.0f : ((values[vt][vx] < 0.0f) ? 0.0f : values[vt][vx])) << bitshift);
	    break;
	}
	case Axes.ZY: {
	    int vystart = 0;
	    int vzstart = 0;
	    int pystart = coords.y;
	    int pzstart = coords.z;
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    int pystop = coords.y + values[0].length;
	    int pzstop = coords.z + values.length;
	    if (pystop > ydim) pystop = ydim;
	    if (pzstop > zdim) pzstop = zdim;
	    final int pxy = pystart*xdim + coords.x;
	    for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		    pixels[coords.t][pz][py] = (setmask & pixels[coords.t][pz][py]) |
			(FMath.round((values[vz][vy] > 255.0f) ? 255.0f : ((values[vz][vy] < 0.0f) ? 0.0f : values[vz][vy])) << bitshift);
	    break;
	}
	case Axes.TY: {
	    int vystart = 0;
	    int vtstart = 0;
	    int pystart = coords.y;
	    int ptstart = coords.t;
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pystop = coords.y + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pystop > ydim) pystop = ydim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = pystart*xdim + coords.x;
	    final int pz = coords.z;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		    pixels[pt][pz][py] = (setmask & pixels[pt][pz][py]) |
			(FMath.round((values[vt][vy] > 255.0f) ? 255.0f : ((values[vt][vy] < 0.0f) ? 0.0f : values[vt][vy])) << bitshift);
	    break;
	}
	case Axes.TZ: {
	    int vzstart = 0;
	    int vtstart = 0;
	    int pzstart = coords.z;
	    int ptstart = coords.t;
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pzstop = coords.z + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pzstop > zdim) pzstop = zdim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = coords.y*xdim + coords.x;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		    pixels[pt][pz][pxy] = (setmask & pixels[pt][pz][pxy]) |
			(FMath.round((values[vt][vz] > 255.0f) ? 255.0f : ((values[vt][vz] < 0.0f) ? 0.0f : values[vt][vz])) << bitshift);
	    break;
	}
	default: // wrong number of active axes
	    throw new IllegalStateException("Wrong number of active axes");
	}
    }

    /** Writes given pixel values in three dimensions into the
        image. The dimensions into which pixel values are written are
        determined by the currently active Cartesian coordinate
        axes. Exactly three axes must be active in order for this method
        to work. The three indices of the given pixel array correspond
        to these axes, where the order of the indices from left to right
        is from the highest to the lowest dimension. The given
        coordinates indicate the starting position of the given pixel
        array, relative to the image. That is to say, they determine the
        starting (x,y,z,t) position in the image for writing pixel
        values. Note that the coordinates corresponding to the active
        axes are allowed to be negative. However, pixel positions
        outside the image in the active dimensions are not actually
        written but simply skipped. This is useful for ignoring possible
        borders around the pixel data in the given array.

	@exception IllegalStateException if not exactly three axes are
	active.

	@exception ArrayIndexOutOfBoundsException if the coordinate
        corresponding to the non-active axis is outside the range
        determined by the size of the image.

	@exception NullPointerException if any of the parameters is
	<code>null</code>.
    */
    public void set(final Coordinates coords, final int[][][] values) {

	switch (axes) {

	case Axes.ZYX: {
	    int vxstart = 0;
	    int vystart = 0;
	    int vzstart = 0;
	    int pxstart = coords.x;
	    int pystart = coords.y;
	    int pzstart = coords.z;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    int pxstop = coords.x + values[0][0].length;
	    int pystop = coords.y + values[0].length;
	    int pzstop = coords.z + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pystop > ydim) pystop = ydim;
	    if (pzstop > zdim) pzstop = zdim;
	    final int pxy = pystart*xdim + pxstart;
	    for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		    for (int x=pxstart, px=py, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
			pixels[coords.t][pz][px] = values[vz][vy][vx];
	    break;
	}
	case Axes.TZX: {
	    int vxstart = 0;
	    int vzstart = 0;
	    int vtstart = 0;
	    int pxstart = coords.x;
	    int pzstart = coords.z;
	    int ptstart = coords.t;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pxstop = coords.x + values[0][0].length;
	    int pzstop = coords.z + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pzstop > zdim) pzstop = zdim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = coords.y*xdim + pxstart;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		    for (int x=pxstart, px=pxy, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
			pixels[pt][pz][px] = values[vt][vz][vx];
	    break;
	}
	case Axes.TYX: {
	    int vxstart = 0;
	    int vystart = 0;
	    int vtstart = 0;
	    int pxstart = coords.x;
	    int pystart = coords.y;
	    int ptstart = coords.t;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pxstop = coords.x + values[0][0].length;
	    int pystop = coords.y + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pystop > ydim) pystop = ydim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = pystart*xdim + pxstart;
	    final int pz = coords.z;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		    for (int x=pxstart, px=py, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
			pixels[pt][pz][px] = values[vt][vy][vx];
	    break;
	}
	case Axes.TZY: {
	    int vystart = 0;
	    int vzstart = 0;
	    int vtstart = 0;
	    int pystart = coords.y;
	    int pzstart = coords.z;
	    int ptstart = coords.t;
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pystop = coords.y + values[0][0].length;
	    int pzstop = coords.z + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pystop > ydim) pystop = ydim;
	    if (pzstop > zdim) pzstop = zdim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = pystart*xdim + coords.x;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		    for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
			pixels[pt][pz][py] = values[vt][vz][vy];
	    break;
	}
	default: // wrong number of active axes
	    throw new IllegalStateException("Wrong number of active axes");
	}
    }

    /** Writes given pixel values in three dimensions into the
        image. Note that this method affects only the currently active
        component. The given pixel values are clipped to the range
        [0,255], rounded to the nearest <code>int</code> value, and then
        replace the active component. The dimensions into which pixel
        values are written are determined by the currently active
        Cartesian coordinate axes. Exactly three axes must be active in
        order for this method to work. The three indices of the given
        pixel array correspond to these axes, where the order of the
        indices from left to right is from the highest to the lowest
        dimension. The given coordinates indicate the starting position
        of the given pixel array, relative to the image. That is to say,
        they determine the starting (x,y,z,t) position in the image for
        writing pixel values. Note that the coordinates corresponding to
        the active axes are allowed to be negative. However, pixel
        positions outside the image in the active dimensions are not
        actually written but simply skipped. This is useful for ignoring
        possible borders around the pixel data in the given array.

	@exception IllegalStateException if not exactly three axes are
	active.

	@exception ArrayIndexOutOfBoundsException if the coordinate
        corresponding to the non-active axis is outside the range
        determined by the size of the image.

	@exception NullPointerException if any of the parameters is
	<code>null</code>.
    */
    public void set(final Coordinates coords, final float[][][] values) {

	switch (axes) {

	case Axes.ZYX: {
	    int vxstart = 0;
	    int vystart = 0;
	    int vzstart = 0;
	    int pxstart = coords.x;
	    int pystart = coords.y;
	    int pzstart = coords.z;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    int pxstop = coords.x + values[0][0].length;
	    int pystop = coords.y + values[0].length;
	    int pzstop = coords.z + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pystop > ydim) pystop = ydim;
	    if (pzstop > zdim) pzstop = zdim;
	    final int pxy = pystart*xdim + pxstart;
	    for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		    for (int x=pxstart, px=py, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
			pixels[coords.t][pz][px] = (setmask & pixels[coords.t][pz][px]) |
			    (FMath.round((values[vz][vy][vx] > 255.0f) ? 255.0f : ((values[vz][vy][vx] < 0.0f) ? 0.0f : values[vz][vy][vx])) << bitshift);
	    break;
	}
	case Axes.TZX: {
	    int vxstart = 0;
	    int vzstart = 0;
	    int vtstart = 0;
	    int pxstart = coords.x;
	    int pzstart = coords.z;
	    int ptstart = coords.t;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pxstop = coords.x + values[0][0].length;
	    int pzstop = coords.z + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pzstop > zdim) pzstop = zdim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = coords.y*xdim + pxstart;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		    for (int x=pxstart, px=pxy, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
			pixels[pt][pz][px] = (setmask & pixels[pt][pz][px]) |
			    (FMath.round((values[vt][vz][vx] > 255.0f) ? 255.0f : ((values[vt][vz][vx] < 0.0f) ? 0.0f : values[vt][vz][vx])) << bitshift);
	    break;
	}
	case Axes.TYX: {
	    int vxstart = 0;
	    int vystart = 0;
	    int vtstart = 0;
	    int pxstart = coords.x;
	    int pystart = coords.y;
	    int ptstart = coords.t;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pxstop = coords.x + values[0][0].length;
	    int pystop = coords.y + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pystop > ydim) pystop = ydim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = pystart*xdim + pxstart;
	    final int pz = coords.z;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		    for (int x=pxstart, px=py, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
			pixels[pt][pz][px] = (setmask & pixels[pt][pz][px]) |
			    (FMath.round((values[vt][vy][vx] > 255.0f) ? 255.0f : ((values[vt][vy][vx] < 0.0f) ? 0.0f : values[vt][vy][vx])) << bitshift);
	    break;
	}
	case Axes.TZY: {
	    int vystart = 0;
	    int vzstart = 0;
	    int vtstart = 0;
	    int pystart = coords.y;
	    int pzstart = coords.z;
	    int ptstart = coords.t;
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pystop = coords.y + values[0][0].length;
	    int pzstop = coords.z + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pystop > ydim) pystop = ydim;
	    if (pzstop > zdim) pzstop = zdim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = pystart*xdim + coords.x;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		    for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
			pixels[pt][pz][py] = (setmask & pixels[pt][pz][py]) |
			    (FMath.round((values[vt][vz][vy] > 255.0f) ? 255.0f : ((values[vt][vz][vy] < 0.0f) ? 0.0f : values[vt][vz][vy])) << bitshift);
	    break;
	}
	default: // wrong number of active axes
	    throw new IllegalStateException("Wrong number of active axes");
	}
    }

    /** Writes given pixel values in all four dimensions into the
        image. All four axes must be active in order for this method to
        work. The four indices of the given pixel array correspond to
        these axes, where the order of the indices from left to right is
        t, z, y, x. The given coordinates indicate the starting position
        of the given pixel array, relative to the image. That is to say,
        they determine the starting (x,y,z,t) position in the image for
        writing pixel values. Note that the coordinates are allowed to
        be negative. However, pixel positions outside the image are not
        actually written but simply skipped. This is useful for ignoring
        possible borders around the pixel data in the given array.

	@exception IllegalStateException if not all four axes are active.

	@exception NullPointerException if any of the parameters is
	<code>null</code>.
    */
    public void set(final Coordinates coords, final int[][][][] values) {

	switch (axes) {

	case Axes.TZYX: {
	    int vxstart = 0;
	    int vystart = 0;
	    int vzstart = 0;
	    int vtstart = 0;
	    int pxstart = coords.x;
	    int pystart = coords.y;
	    int pzstart = coords.z;
	    int ptstart = coords.t;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pxstop = coords.x + values[0][0][0].length;
	    int pystop = coords.y + values[0][0].length;
	    int pzstop = coords.z + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pystop > ydim) pystop = ydim;
	    if (pzstop > zdim) pzstop = zdim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = pystart*xdim + pxstart;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		    for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
			for (int x=pxstart, px=py, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
			    pixels[pt][pz][px] = values[vt][vz][vy][vx];
	    break;
	}
	default: // wrong number of active axes
	    throw new IllegalStateException("Wrong number of active axes");
	}
    }

    /** Writes given pixel values in all four dimensions into the
        image. Note that this method affects only the currently active
        component. The given pixel values are clipped to the range
        [0,255], rounded to the nearest <code>int</code> value, and then
        replace the active component. All four axes must be active in
        order for this method to work. The four indices of the given
        pixel array correspond to these axes, where the order of the
        indices from left to right is t, z, y, x. The given coordinates
        indicate the starting position of the given pixel array,
        relative to the image. That is to say, they determine the
        starting (x,y,z,t) position in the image for writing pixel
        values. Note that the coordinates are allowed to be
        negative. However, pixel positions outside the image are not
        actually written but simply skipped. This is useful for ignoring
        possible borders around the pixel data in the given array.

	@exception IllegalStateException if not all four axes are active.

	@exception NullPointerException if any of the parameters is
	<code>null</code>.
    */
    public void set(final Coordinates coords, final float[][][][] values) {

	switch (axes) {

	case Axes.TZYX: {
	    int vxstart = 0;
	    int vystart = 0;
	    int vzstart = 0;
	    int vtstart = 0;
	    int pxstart = coords.x;
	    int pystart = coords.y;
	    int pzstart = coords.z;
	    int ptstart = coords.t;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pxstop = coords.x + values[0][0][0].length;
	    int pystop = coords.y + values[0][0].length;
	    int pzstop = coords.z + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pystop > ydim) pystop = ydim;
	    if (pzstop > zdim) pzstop = zdim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = pystart*xdim + pxstart;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		    for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
			for (int x=pxstart, px=py, vx=vxstart; x < pxstop; ++x, ++px, ++vx) {
			    final float value = values[vt][vz][vy][vx];
			    pixels[pt][pz][px] = (setmask & pixels[pt][pz][px]) |
				(FMath.round((value > 255.0f) ? 255.0f : ((value < 0.0f) ? 0.0f : value)) << bitshift);
			}
	    break;
	}
	default: // wrong number of active axes
	    throw new IllegalStateException("Wrong number of active axes");
	}
    }

} 
