// **********************************************************************
// COPYRIGHT NOTICE AND DISCLAIMER
//
// Copyright (c) Erik Meijering. All rights reserved.
//
// Permission to copy and use this software and accompanying
// documentation for educational, research, and not-for-profit
// purposes, without fee and without a signed licensing agreement,
// is hereby granted, provided that the above copyright notice, this
// paragraph and the following two paragraphs appear in all copies.
// The copyright holder is free to make upgraded or improved versions
// of the software available for a fee or commercially only.
//
// IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE TO ANY PARTY FOR
// DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
// OF ANY KIND WHATSOEVER, ARISING OUT OF THE USE OF THIS SOFTWARE
// AND ITS DOCUMENTATION, EVEN IF HE HAS BEEN ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//
// THE COPYRIGHT HOLDER SPECIFICALLY DISCLAIMS ANY WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
// SOFTWARE AND ACCOMPANYING DOCUMENTATION IS PROVIDED "AS IS".
// THE COPYRIGHT HOLDER HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
// SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
//
// **********************************************************************

package imagescience.images;

/** Contains size information about (at most) 4D data objects. */
public final class Dimensions {

    /** The size of the x-dimension. */
    public final int x;

    /** The size of the y-dimension. */
    public final int y;

    /** The size of the z-dimension. */
    public final int z;

    /** The size of the t-dimension. */
    public final int t;

    /** Default constructor. Each of the dimensions is set to 1. */
    public Dimensions() {
	this.x = 1;
	this.y = 1;
	this.z = 1;
	this.t = 1;
    }

    /** One-dimensional constructor. The x-dimension is set to the
        given value. The y-, z-, and t-dimensions are set to 1.

	@exception IllegalArgumentException if the parameter is negative
	or zero.
    */
    public Dimensions(final int x) {
	if (x <= 0) throw new IllegalArgumentException("Argument negative or zero");
	this.x = x;
	this.y = 1;
	this.z = 1;
	this.t = 1;
    }

    /** Two-dimensional constructor. The x- and y-dimensions are set to
        the given values. The z- and t-dimensions are set to 1.

	@exception IllegalArgumentException if any of the parameters is
	negative or zero.
    */
    public Dimensions(final int x, final int y) {
	if (x <= 0 || y <= 0) throw new IllegalArgumentException("Argument(s) negative or zero");
	this.x = x;
	this.y = y;
	this.z = 1;
	this.t = 1;
    }

    /** Three-dimensional constructor. The x-, y-, and z-dimensions are
        set to the given values. The t-dimension is set to 1.

	@exception IllegalArgumentException if any of the parameters is
	negative or zero.
    */
    public Dimensions(final int x, final int y, final int z) {
	if (x <= 0 || y <= 0 || z <= 0) throw new IllegalArgumentException("Argument(s) negative or zero");
	this.x = x;
	this.y = y;
	this.z = z;
	this.t = 1;
    }

    /** Four-dimensional constructor. The x-, y-, z-, and t-dimensions
        are set to the given values.

	@exception IllegalArgumentException if any of the parameters is
	negative or zero.
    */
    public Dimensions(final int x, final int y, final int z, final int t) {
	if (x <= 0 || y <= 0 || z <= 0 || t <= 0) throw new IllegalArgumentException("Argument(s) negative or zero");
	this.x = x;
	this.y = y;
	this.z = z;
	this.t = t;
    }

    /** Returns a copy of the current <tt>Dimensions</tt> object. */
    public Dimensions duplicate() { return new Dimensions(x,y,z,t); }

    /** Indicates whether or not the argument dimensions are equal to
        the current dimensions.

	@return <tt>true</tt> if each of the current dimensions is equal
	to the corresponding parameter dimension; <tt>false</tt> if this
	is not the case.

	@exception NullPointerException if the parameter is <tt>null</tt>.
    */
    public boolean equals(Dimensions dims) {
	if (x==dims.x && y==dims.y && z==dims.z && t==dims.t) return true;
	return false;
    }

}
