// **********************************************************************
// COPYRIGHT NOTICE AND DISCLAIMER
//
// Copyright (c) Erik Meijering. All rights reserved.
//
// Permission to copy and use this software and accompanying
// documentation for educational, research, and not-for-profit
// purposes, without fee and without a signed licensing agreement,
// is hereby granted, provided that the above copyright notice, this
// paragraph and the following two paragraphs appear in all copies.
// The copyright holder is free to make upgraded or improved versions
// of the software available for a fee or commercially only.
//
// IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE TO ANY PARTY FOR
// DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
// OF ANY KIND WHATSOEVER, ARISING OUT OF THE USE OF THIS SOFTWARE
// AND ITS DOCUMENTATION, EVEN IF HE HAS BEEN ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//
// THE COPYRIGHT HOLDER SPECIFICALLY DISCLAIMS ANY WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
// SOFTWARE AND ACCOMPANYING DOCUMENTATION IS PROVIDED "AS IS".
// THE COPYRIGHT HOLDER HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
// SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
//
// **********************************************************************

package imagescience.images;

import ij.ImageStack;

/** Represents images of up to four dimensions containing data of type
    <code>float</code>. */
public final class FloatImage extends Image {
    
    // Pixel array:
    final float[][][] pixels;

    /** Constructor for creating a one-dimensional image of <code>float</code>s.

	@exception IllegalArgumentException if the parameter is zero or negative.
    */
    public FloatImage(final int xdim) {
	
	super(xdim,1,1,1);

	pixels = new float[1][1][xdim];
    }

    /** Constructor for creating a two-dimensional image of <code>float</code>s.

	@exception IllegalArgumentException if any of the parameters is zero or negative.
    */
    public FloatImage(final int xdim, final int ydim) {
	
	super(xdim,ydim,1,1);

	pixels = new float[1][1][xdim*ydim];
    }

    /** Constructor for creating a three-dimensional image of <code>float</code>s.

	@exception IllegalArgumentException if any of the parameters is zero or negative.
    */
    public FloatImage(final int xdim, final int ydim, final int zdim) {
	
	super(xdim,ydim,zdim,1);

	pixels = new float[1][zdim][xdim*ydim];
    }

    /** Constructor for creating a four-dimensional image of <code>float</code>s.

	@exception IllegalArgumentException if any of the parameters is zero or negative.
    */
    public FloatImage(final int xdim, final int ydim, final int zdim, final int tdim) {
	
	super(xdim,ydim,zdim,tdim);

	pixels = new float[tdim][zdim][xdim*ydim];
    }

    /** Constructor for creating a four-dimensional image of
	<code>float</code>s. The size of the image in the four dimensions
	is determined by the field values of the parameter.

	@exception NullPointerException if the parameter is <code>null</code>.
    */
    public FloatImage(final Dimensions dims) {
	
	super(dims.x,dims.y,dims.z,dims.t);

	pixels = new float[tdim][zdim][xdim*ydim];
    }

    /** Copy constructor. All pixel data and corresponding information
        contained in the given image is copied, and nothing in the
        resulting image is shared with the given image. Since pixel
        values are taken in the unsigned sense, the values in the
        resulting image are always in the range [0,255].

	@exception NullPointerException if the parameter is <code>null</code>.
    */
    public FloatImage(final ByteImage image) {

	super(image.xdim,image.ydim,image.zdim,image.tdim);

	pixels = new float[tdim][zdim][xdim*ydim];

	final int xydim = xdim*ydim;
	for (int t=0; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy)
		    pixels[t][z][xy] = 0xFF & image.pixels[t][z][xy];

	axes = image.axes;

	xpixdim = image.xpixdim;
	ypixdim = image.ypixdim;
	zpixdim = image.zpixdim;
	tpixdim = image.tpixdim;
    }

    /** Copy constructor. All pixel data and corresponding information
        contained in the given image is copied, and nothing in the
        resulting image is shared with the given image. Since pixel
        values are taken in the unsigned sense, the values in the
        resulting image are always in the range [0,65535].

	@exception NullPointerException if the parameter is <code>null</code>.
    */
    public FloatImage(final ShortImage image) {
	
	super(image.xdim,image.ydim,image.zdim,image.tdim);

	pixels = new float[tdim][zdim][xdim*ydim];

	final int xydim = xdim*ydim;
	for (int t=0; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy)
		    pixels[t][z][xy] = 0xFFFF & image.pixels[t][z][xy];

	axes = image.axes;

	xpixdim = image.xpixdim;
	ypixdim = image.ypixdim;
	zpixdim = image.zpixdim;
	tpixdim = image.tpixdim;
    }

    /** Copy constructor. All pixel data and corresponding information
        contained in the given image is copied, and nothing in the
        resulting image is shared with the given image.

	@exception NullPointerException if the parameter is <code>null</code>.
    */
    public FloatImage(final FloatImage image) {
	
	super(image.xdim,image.ydim,image.zdim,image.tdim);

	pixels = new float[tdim][zdim][xdim*ydim];

	final int xydim = xdim*ydim;
	for (int t=0; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy)
		    pixels[t][z][xy] = image.pixels[t][z][xy];

	axes = image.axes;

	xpixdim = image.xpixdim;
	ypixdim = image.ypixdim;
	zpixdim = image.zpixdim;
	tpixdim = image.tpixdim;
    }

    /** Copy constructor that allows adding borders. Creates a new image
        whose size in every dimension is equal to that of the given
        image plus twice the value of the given borders size in that
        dimension. All pixel data and corresponding information
        contained in the given image is copied, and nothing in the
        resulting image is shared with the given image. The
        <code>borders</code> parameter determines the starting position
        for copying into the new image: the data is copied in such a way
        that in each dimension, the resulting image has an empty border
        of given size on both sides. Since pixel values are taken in the
        unsigned sense, the values in the resulting image are always in
        the range [0,255].

	@exception NullPointerException if any of the parameters is
	<code>null</code>.
    */
    public FloatImage(final ByteImage image, final Borders borders) {
	
	super(image.xdim+2*borders.x,image.ydim+2*borders.y,image.zdim+2*borders.z,image.tdim+2*borders.t);

	pixels = new float[tdim][zdim][xdim*ydim];

	for (int tin=0, tnew=borders.t; tin<image.tdim; ++tin, ++tnew)
	    for (int zin=0, znew=borders.z; zin<image.zdim; ++zin, ++znew)
		for (int y=0, xyin=0, ynew=borders.y*xdim+borders.x; y<image.ydim; ++y, ynew+=xdim)
		    for (int x=0, xynew=ynew; x<image.xdim; ++x, ++xyin, ++xynew)
			pixels[tnew][znew][xynew] = 0xFF & image.pixels[tin][zin][xyin];

	axes = image.axes;

	xpixdim = image.xpixdim;
	ypixdim = image.ypixdim;
	zpixdim = image.zpixdim;
	tpixdim = image.tpixdim;
    }

    /** Copy constructor that allows adding borders. Creates a new image
        whose size in every dimension is equal to that of the given
        image plus twice the value of the given borders size in that
        dimension. All pixel data and corresponding information
        contained in the given image is copied, and nothing in the
        resulting image is shared with the given image. The
        <code>borders</code> parameter determines the starting position
        for copying into the new image: the data is copied in such a way
        that in each dimension, the resulting image has an empty border
        of given size on both sides. Since pixel values are taken in the
        unsigned sense, the values in the resulting image are always in
        the range [0,65535.

	@exception NullPointerException if any of the parameters is
	<code>null</code>.
    */
    public FloatImage(final ShortImage image, final Borders borders) {
	
	super(image.xdim+2*borders.x,image.ydim+2*borders.y,image.zdim+2*borders.z,image.tdim+2*borders.t);

	pixels = new float[tdim][zdim][xdim*ydim];

	for (int tin=0, tnew=borders.t; tin<image.tdim; ++tin, ++tnew)
	    for (int zin=0, znew=borders.z; zin<image.zdim; ++zin, ++znew)
		for (int y=0, xyin=0, ynew=borders.y*xdim+borders.x; y<image.ydim; ++y, ynew+=xdim)
		    for (int x=0, xynew=ynew; x<image.xdim; ++x, ++xyin, ++xynew)
			pixels[tnew][znew][xynew] = 0xFFFF & image.pixels[tin][zin][xyin];

	axes = image.axes;

	xpixdim = image.xpixdim;
	ypixdim = image.ypixdim;
	zpixdim = image.zpixdim;
	tpixdim = image.tpixdim;
    }

    /** Copy constructor that allows adding borders. Creates a new image
        whose size in every dimension is equal to that of the given
        image plus twice the value of the given borders size in that
        dimension. All pixel data and corresponding information
        contained in the given image is copied, and nothing in the
        resulting image is shared with the given image. The
        <code>borders</code> parameter determines the starting position
        for copying into the new image: the data is copied in such a way
        that in each dimension, the resulting image has an empty border
        of given size on both sides.

	@exception NullPointerException if any of the parameters is
	<code>null</code>.
    */
    public FloatImage(final FloatImage image, final Borders borders) {
	
	super(image.xdim+2*borders.x,image.ydim+2*borders.y,image.zdim+2*borders.z,image.tdim+2*borders.t);

	pixels = new float[tdim][zdim][xdim*ydim];

	for (int tin=0, tnew=borders.t; tin<image.tdim; ++tin, ++tnew)
	    for (int zin=0, znew=borders.z; zin<image.zdim; ++zin, ++znew)
		for (int y=0, xyin=0, ynew=borders.y*xdim+borders.x; y<image.ydim; ++y, ynew+=xdim)
		    for (int x=0, xynew=ynew; x<image.xdim; ++x, ++xyin, ++xynew)
			pixels[tnew][znew][xynew] = image.pixels[tin][zin][xyin];

	axes = image.axes;

	xpixdim = image.xpixdim;
	ypixdim = image.ypixdim;
	zpixdim = image.zpixdim;
	tpixdim = image.tpixdim;
    }

    /** Constructor for creating a three-dimensional image wrapper
        object. The size information is copied from the given
        <code>ImageStack</code> object. The actual pixel data, however,
        is not copied to, but shared with the newly created image.

	@exception IllegalArgumentException if the data stored in
	<code>stack</code> is not of type <code>float</code>.

	@exception NullPointerException if the parameter is <code>null</code>.
    */
    public FloatImage(final ImageStack stack) {
	
	super(stack.getWidth(),stack.getHeight(),stack.getSize(),1);

	final Object[] spixels = stack.getImageArray();
	if (!(spixels[0] instanceof float[]))
	    throw new IllegalArgumentException("Wrong pixel type");

	pixels = new float[tdim][zdim][];
	for (int z=0; z<zdim; ++z)
	    pixels[0][z] = (float[])spixels[z];
    }

    /** Constructor for creating a four-dimensional image wrapper
        object. The size information is copied from the given
        <code>ImageStack</code> object. The actual pixel data, however,
        is not copied to, but shared with the newly created image.

	@exception IllegalArgumentException if the data stored in
	<code>stack</code> is not of type <code>float</code>, or if the
	number of slices in <code>stack</code> is not an integral
	multiple of the value of the second parameter.

	@exception NullPointerException if the first parameter is <code>null</code>.
    */
    public FloatImage(final ImageStack stack, final int slicespervolume) {

	super(stack.getWidth(),stack.getHeight(),slicespervolume,stack.getSize()/slicespervolume);

	if (zdim*tdim != stack.getSize())
	    throw new IllegalArgumentException("Incorrect number of slices per volume");

	final Object[] spixels = stack.getImageArray();
	if (!(spixels[0] instanceof float[]))
	    throw new IllegalArgumentException("Wrong pixel type");

	pixels = new float[tdim][zdim][];
	for (int t=0, slice=0; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z, ++slice)
		pixels[t][z] = (float[])spixels[slice];
    }

    /** Returns a handle to the internal pixel array of the current
        image. In the current implementation, this is a
        three-dimensional array, where the first two indices from the
        left correspond to the t- and z-dimension, respectively, while
        the right-most index addresses the x-y-dimension, containing the
        sequence of pixel rows. One should be cautious in using this
        method, as the organization of the internal array containing the
        actual pixel data may change in future versions of the class. */
    public float[][][] array() { return pixels; }

    public ImageStack stack() {

	ImageStack stack = new ImageStack(xdim,ydim);

	for (int t=0; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z)
		stack.addSlice("",pixels[t][z]);

	return stack;
    }

    /** Returns an exact copy of the image. That is to say, all pixel
        data and corresponding information contained in the object is
        duplicated, and nothing in the returned object is shared with
        the object. */
    public FloatImage duplicate() {

	FloatImage dupi = new FloatImage(xdim,ydim,zdim,tdim);

	final int xydim = xdim*ydim;
	for (int t=0; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy)
		    dupi.pixels[t][z][xy] = pixels[t][z][xy];

	dupi.axes = axes;

	dupi.xpixdim = xpixdim;
	dupi.ypixdim = ypixdim;
	dupi.zpixdim = zpixdim;
	dupi.tpixdim = tpixdim;

	return dupi;
    }

    public int memory() { return (4*xdim*ydim*zdim*tdim)/1024; }

    /** Returns the minimum value in the image. */
    public float minimum() {

	final int xydim = xdim*ydim;
	float min = Float.MAX_VALUE;
	for (int t=0; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy) {
		    final float value = pixels[t][z][xy];
		    if (value < min) min = value;
		}

	return min;
    }

    /** Returns the maximum value in the image. */
    public float maximum() {

	final int xydim = xdim*ydim;
	float max = -Float.MAX_VALUE;
	for (int t=0; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy) {
		    final float value = pixels[t][z][xy];
		    if (value > max) max = value;
		}

	return max;
    }

    public float get(final Coordinates coords) {

	return pixels[coords.t][coords.z][xdim*coords.y + coords.x];
    }

    public void get(final Coordinates coords, final float[] values) {

	switch (axes) {

	case Axes.X: {
	    int vxstart = 0;
	    int pxstart = coords.x;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    int pxstop = coords.x + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    for (int x=pxstart, px=coords.y*xdim + pxstart, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
		values[vx] = pixels[coords.t][coords.z][px];
	    break;
	}
	case Axes.Y: {
	    int vystart = 0;
	    int pystart = coords.y;
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    int pystop = coords.y + values.length;
	    if (pystop > ydim) pystop = ydim;
	    for (int y=pystart, py=pystart*xdim + coords.x, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		values[vy] = pixels[coords.t][coords.z][py];
	    break;
	}
	case Axes.Z: {
	    int vzstart = 0;
	    int pzstart = coords.z;
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    int pzstop = coords.z + values.length;
	    if (pzstop > zdim) pzstop = zdim;
	    final int pxy = coords.y*xdim + coords.x;
	    for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		values[vz] = pixels[coords.t][pz][pxy];
	    break;
	}
	case Axes.T: {
	    int vtstart = 0;
	    int ptstart = coords.t;
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int ptstop = coords.t + values.length;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pz = coords.z;
	    final int pxy = coords.y*xdim + coords.x;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		values[vt] = pixels[pt][pz][pxy];
	    break;
	}
	default: // wrong number of active axes
	    throw new IllegalStateException("Wrong number of active axes");
	}
    }

    public void get(final Coordinates coords, final float[][] values) {

	switch (axes) {

	case Axes.YX: {
	    int vxstart = 0;
	    int vystart = 0;
	    int pxstart = coords.x;
	    int pystart = coords.y;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    int pxstop = coords.x + values[0].length;
	    int pystop = coords.y + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pystop > ydim) pystop = ydim;
	    for (int y=pystart, py=pystart*xdim + pxstart, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		for (int x=pxstart, px=py, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
		    values[vy][vx] = pixels[coords.t][coords.z][px];
	    break;
	}
	case Axes.ZX: {
	    int vxstart = 0;
	    int vzstart = 0;
	    int pxstart = coords.x;
	    int pzstart = coords.z;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    int pxstop = coords.x + values[0].length;
	    int pzstop = coords.z + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pzstop > zdim) pzstop = zdim;
	    final int pxy = coords.y*xdim + pxstart;
	    for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		for (int x=pxstart, px=pxy, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
		    values[vz][vx] = pixels[coords.t][pz][px];
	    break;
	}
	case Axes.TX: {
	    int vxstart = 0;
	    int vtstart = 0;
	    int pxstart = coords.x;
	    int ptstart = coords.t;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pxstop = coords.x + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = coords.y*xdim + pxstart;
	    final int pz = coords.z;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int x=pxstart, px=pxy, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
		    values[vt][vx] = pixels[pt][pz][px];
	    break;
	}
	case Axes.ZY: {
	    int vystart = 0;
	    int vzstart = 0;
	    int pystart = coords.y;
	    int pzstart = coords.z;
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    int pystop = coords.y + values[0].length;
	    int pzstop = coords.z + values.length;
	    if (pystop > ydim) pystop = ydim;
	    if (pzstop > zdim) pzstop = zdim;
	    final int pxy = pystart*xdim + coords.x;
	    for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		    values[vz][vy] = pixels[coords.t][pz][py];
	    break;
	}
	case Axes.TY: {
	    int vystart = 0;
	    int vtstart = 0;
	    int pystart = coords.y;
	    int ptstart = coords.t;
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pystop = coords.y + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pystop > ydim) pystop = ydim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = pystart*xdim + coords.x;
	    final int pz = coords.z;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		    values[vt][vy] = pixels[pt][pz][py];
	    break;
	}
	case Axes.TZ: {
	    int vzstart = 0;
	    int vtstart = 0;
	    int pzstart = coords.z;
	    int ptstart = coords.t;
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pzstop = coords.z + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pzstop > zdim) pzstop = zdim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = coords.y*xdim + coords.x;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		    values[vt][vz] = pixels[pt][pz][pxy];
	    break;
	}
	default: // wrong number of active axes
	    throw new IllegalStateException("Wrong number of active axes");
	}
    }

    public void get(final Coordinates coords, final float[][][] values) {

	switch (axes) {

	case Axes.ZYX: {
	    int vxstart = 0;
	    int vystart = 0;
	    int vzstart = 0;
	    int pxstart = coords.x;
	    int pystart = coords.y;
	    int pzstart = coords.z;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    int pxstop = coords.x + values[0][0].length;
	    int pystop = coords.y + values[0].length;
	    int pzstop = coords.z + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pystop > ydim) pystop = ydim;
	    if (pzstop > zdim) pzstop = zdim;
	    final int pxy = pystart*xdim + pxstart;
	    for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		    for (int x=pxstart, px=py, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
			values[vz][vy][vx] = pixels[coords.t][pz][px];
	    break;
	}
	case Axes.TZX: {
	    int vxstart = 0;
	    int vzstart = 0;
	    int vtstart = 0;
	    int pxstart = coords.x;
	    int pzstart = coords.z;
	    int ptstart = coords.t;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pxstop = coords.x + values[0][0].length;
	    int pzstop = coords.z + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pzstop > zdim) pzstop = zdim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = coords.y*xdim + pxstart;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		    for (int x=pxstart, px=pxy, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
			values[vt][vz][vx] = pixels[pt][pz][px];
	    break;
	}
	case Axes.TYX: {
	    int vxstart = 0;
	    int vystart = 0;
	    int vtstart = 0;
	    int pxstart = coords.x;
	    int pystart = coords.y;
	    int ptstart = coords.t;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pxstop = coords.x + values[0][0].length;
	    int pystop = coords.y + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pystop > ydim) pystop = ydim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = pystart*xdim + pxstart;
	    final int pz = coords.z;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		    for (int x=pxstart, px=py, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
			values[vt][vy][vx] = pixels[pt][pz][px];
	    break;
	}
	case Axes.TZY: {
	    int vystart = 0;
	    int vzstart = 0;
	    int vtstart = 0;
	    int pystart = coords.y;
	    int pzstart = coords.z;
	    int ptstart = coords.t;
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pystop = coords.y + values[0][0].length;
	    int pzstop = coords.z + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pystop > ydim) pystop = ydim;
	    if (pzstop > zdim) pzstop = zdim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = pystart*xdim + coords.x;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		    for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
			values[vt][vz][vy] = pixels[pt][pz][py];
	    break;
	}
	default: // wrong number of active axes
	    throw new IllegalStateException("Wrong number of active axes");
	}
    }

    public void get(final Coordinates coords, final float[][][][] values) {

	switch (axes) {

	case Axes.TZYX: {
	    int vxstart = 0;
	    int vystart = 0;
	    int vzstart = 0;
	    int vtstart = 0;
	    int pxstart = coords.x;
	    int pystart = coords.y;
	    int pzstart = coords.z;
	    int ptstart = coords.t;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pxstop = coords.x + values[0][0][0].length;
	    int pystop = coords.y + values[0][0].length;
	    int pzstop = coords.z + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pystop > ydim) pystop = ydim;
	    if (pzstop > zdim) pzstop = zdim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = pystart*xdim + pxstart;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		    for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
			for (int x=pxstart, px=py, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
			    values[vt][vz][vy][vx] = pixels[pt][pz][px];
	    break;
	}
	default: // wrong number of active axes
	    throw new IllegalStateException("Wrong number of active axes");
	}
    }

    public void set(final float value) {

	final int xydim = xdim*ydim;
	for (int t=0; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy)
		    pixels[t][z][xy] = value;
    }

    public void set(final Coordinates coords, final float value) {

	pixels[coords.t][coords.z][xdim*coords.y + coords.x] = value;
    }
    
    /** Sets all pixels in the given borders to the given value. The
        size of the borders at the beginning and end in any dimension of
        the image is determined by the corresponding field of
        <code>borders</code>.
	
	@exception IllegalArgumentException if any of the given border
	sizes is larger than half the image size in the corresponding
	dimension.
	
	@exception NullPointerException if the first parameter is
	<code>null</code>.
    */
    public void set(final Borders borders, final float value) {

	if (borders.x > xdim/2 || borders.y > ydim/2 ||
	    borders.z > zdim/2 || borders.t > tdim/2)
	    throw new IllegalArgumentException("Border(s) too large");

	// Initialize:
	final int tdimb = tdim - borders.t;
	final int zdimb = zdim - borders.z;
	final int ydimb = ydim - borders.y;
	final int xdimb = xdim - borders.x;
	final int xydim = xdim*ydim;

	// Fill borders in t-dimension:
	for (int t=0; t<borders.t; ++t)
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy)
		    pixels[t][z][xy] = value;
	for (int t=tdimb; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy)
		    pixels[t][z][xy] = value;

	// Fill borders in z-dimension:
	for (int t=borders.t; t<tdimb; ++t) {
	    for (int z=0; z<borders.z; ++z)
		for (int xy=0; xy<xydim; ++xy)
		    pixels[t][z][xy] = value;
	    for (int z=zdimb; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy)
		    pixels[t][z][xy] = value;
	}

	// Fill borders in y-dimension:
	for (int t=borders.t; t<tdimb; ++t)
	    for (int z=borders.z; z<zdimb; ++z) {
		for (int y=0, xy=0; y<borders.y; ++y)
		    for (int x=0; x<xdim; ++x, ++xy)
			pixels[t][z][xy] = value;
		for (int y=ydimb, xy=ydimb*xdim; y<ydim; ++y)
		    for (int x=0; x<xdim; ++x, ++xy)
			pixels[t][z][xy] = value;
	    }

	// Fill borders in x-dimension:
	for (int t=borders.t; t<tdimb; ++t)
	    for (int z=borders.z; z<zdimb; ++z)
		for (int y=borders.y; y<ydimb; ++y) {
		    for (int x=0, xy=y*xdim; x<borders.x; ++x, ++xy)
			pixels[t][z][xy] = value;
		    for (int x=xdimb, xy=y*xdim+xdimb; x<xdim; ++x, ++xy)
			pixels[t][z][xy] = value;
		}
    }
    
    public void set(final Coordinates coords, final float[] values) {

	switch (axes) {

	case Axes.X: {
	    int vxstart = 0;
	    int pxstart = coords.x;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    int pxstop = coords.x + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    for (int x=pxstart, px=coords.y*xdim + pxstart, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
		pixels[coords.t][coords.z][px] = values[vx];
	    break;
	}
	case Axes.Y: {
	    int vystart = 0;
	    int pystart = coords.y;
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    int pystop = coords.y + values.length;
	    if (pystop > ydim) pystop = ydim;
	    for (int y=pystart, py=pystart*xdim + coords.x, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		pixels[coords.t][coords.z][py] = values[vy];
	    break;
	}
	case Axes.Z: {
	    int vzstart = 0;
	    int pzstart = coords.z;
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    int pzstop = coords.z + values.length;
	    if (pzstop > zdim) pzstop = zdim;
	    final int pxy = coords.y*xdim + coords.x;
	    for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		pixels[coords.t][pz][pxy] = values[vz];
	    break;
	}
	case Axes.T: {
	    int vtstart = 0;
	    int ptstart = coords.t;
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int ptstop = coords.t + values.length;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pz = coords.z;
	    final int pxy = coords.y*xdim + coords.x;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		pixels[pt][pz][pxy] = values[vt];
	    break;
	}
	default: // wrong number of active axes
	    throw new IllegalStateException("Wrong number of active axes");
	}
    }

    public void set(final Coordinates coords, final float[][] values) {
	
	switch (axes) {

	case Axes.YX: {
	    int vxstart = 0;
	    int vystart = 0;
	    int pxstart = coords.x;
	    int pystart = coords.y;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    int pxstop = coords.x + values[0].length;
	    int pystop = coords.y + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pystop > ydim) pystop = ydim;
	    for (int y=pystart, py=pystart*xdim + pxstart, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		for (int x=pxstart, px=py, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
		    pixels[coords.t][coords.z][px] = values[vy][vx];
	    break;
	}
	case Axes.ZX: {
	    int vxstart = 0;
	    int vzstart = 0;
	    int pxstart = coords.x;
	    int pzstart = coords.z;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    int pxstop = coords.x + values[0].length;
	    int pzstop = coords.z + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pzstop > zdim) pzstop = zdim;
	    final int pxy = coords.y*xdim + pxstart;
	    for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		for (int x=pxstart, px=pxy, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
		    pixels[coords.t][pz][px] = values[vz][vx];
	    break;
	}
	case Axes.TX: {
	    int vxstart = 0;
	    int vtstart = 0;
	    int pxstart = coords.x;
	    int ptstart = coords.t;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pxstop = coords.x + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = coords.y*xdim + pxstart;
	    final int pz = coords.z;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int x=pxstart, px=pxy, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
		    pixels[pt][pz][px] = values[vt][vx];
	    break;
	}
	case Axes.ZY: {
	    int vystart = 0;
	    int vzstart = 0;
	    int pystart = coords.y;
	    int pzstart = coords.z;
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    int pystop = coords.y + values[0].length;
	    int pzstop = coords.z + values.length;
	    if (pystop > ydim) pystop = ydim;
	    if (pzstop > zdim) pzstop = zdim;
	    final int pxy = pystart*xdim + coords.x;
	    for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		    pixels[coords.t][pz][py] = values[vz][vy];
	    break;
	}
	case Axes.TY: {
	    int vystart = 0;
	    int vtstart = 0;
	    int pystart = coords.y;
	    int ptstart = coords.t;
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pystop = coords.y + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pystop > ydim) pystop = ydim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = pystart*xdim + coords.x;
	    final int pz = coords.z;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		    pixels[pt][pz][py] = values[vt][vy];
	    break;
	}
	case Axes.TZ: {
	    int vzstart = 0;
	    int vtstart = 0;
	    int pzstart = coords.z;
	    int ptstart = coords.t;
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pzstop = coords.z + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pzstop > zdim) pzstop = zdim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = coords.y*xdim + coords.x;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		    pixels[pt][pz][pxy] = values[vt][vz];
	    break;
	}
	default: // wrong number of active axes
	    throw new IllegalStateException("Wrong number of active axes");
	}
    }

    public void set(final Coordinates coords, final float[][][] values) {

	switch (axes) {

	case Axes.ZYX: {
	    int vxstart = 0;
	    int vystart = 0;
	    int vzstart = 0;
	    int pxstart = coords.x;
	    int pystart = coords.y;
	    int pzstart = coords.z;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    int pxstop = coords.x + values[0][0].length;
	    int pystop = coords.y + values[0].length;
	    int pzstop = coords.z + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pystop > ydim) pystop = ydim;
	    if (pzstop > zdim) pzstop = zdim;
	    final int pxy = pystart*xdim + pxstart;
	    for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		    for (int x=pxstart, px=py, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
			pixels[coords.t][pz][px] = values[vz][vy][vx];
	    break;
	}
	case Axes.TZX: {
	    int vxstart = 0;
	    int vzstart = 0;
	    int vtstart = 0;
	    int pxstart = coords.x;
	    int pzstart = coords.z;
	    int ptstart = coords.t;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pxstop = coords.x + values[0][0].length;
	    int pzstop = coords.z + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pzstop > zdim) pzstop = zdim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = coords.y*xdim + pxstart;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		    for (int x=pxstart, px=pxy, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
			pixels[pt][pz][px] = values[vt][vz][vx];
	    break;
	}
	case Axes.TYX: {
	    int vxstart = 0;
	    int vystart = 0;
	    int vtstart = 0;
	    int pxstart = coords.x;
	    int pystart = coords.y;
	    int ptstart = coords.t;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pxstop = coords.x + values[0][0].length;
	    int pystop = coords.y + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pystop > ydim) pystop = ydim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = pystart*xdim + pxstart;
	    final int pz = coords.z;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
		    for (int x=pxstart, px=py, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
			pixels[pt][pz][px] = values[vt][vy][vx];
	    break;
	}
	case Axes.TZY: {
	    int vystart = 0;
	    int vzstart = 0;
	    int vtstart = 0;
	    int pystart = coords.y;
	    int pzstart = coords.z;
	    int ptstart = coords.t;
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pystop = coords.y + values[0][0].length;
	    int pzstop = coords.z + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pystop > ydim) pystop = ydim;
	    if (pzstop > zdim) pzstop = zdim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = pystart*xdim + coords.x;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		    for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
			pixels[pt][pz][py] = values[vt][vz][vy];
	    break;
	}
	default: // wrong number of active axes
	    throw new IllegalStateException("Wrong number of active axes");
	}
    }

    public void set(final Coordinates coords, final float[][][][] values) {

	switch (axes) {

	case Axes.TZYX: {
	    int vxstart = 0;
	    int vystart = 0;
	    int vzstart = 0;
	    int vtstart = 0;
	    int pxstart = coords.x;
	    int pystart = coords.y;
	    int pzstart = coords.z;
	    int ptstart = coords.t;
	    if (pxstart < 0) { pxstart = 0; vxstart = -coords.x; }
	    if (pystart < 0) { pystart = 0; vystart = -coords.y; }
	    if (pzstart < 0) { pzstart = 0; vzstart = -coords.z; }
	    if (ptstart < 0) { ptstart = 0; vtstart = -coords.t; }
	    int pxstop = coords.x + values[0][0][0].length;
	    int pystop = coords.y + values[0][0].length;
	    int pzstop = coords.z + values[0].length;
	    int ptstop = coords.t + values.length;
	    if (pxstop > xdim) pxstop = xdim;
	    if (pystop > ydim) pystop = ydim;
	    if (pzstop > zdim) pzstop = zdim;
	    if (ptstop > tdim) ptstop = tdim;
	    final int pxy = pystart*xdim + pxstart;
	    for (int pt=ptstart, vt=vtstart; pt < ptstop; ++pt, ++vt)
		for (int pz=pzstart, vz=vzstart; pz < pzstop; ++pz, ++vz)
		    for (int y=pystart, py=pxy, vy=vystart; y < pystop; ++y, py+=xdim, ++vy)
			for (int x=pxstart, px=py, vx=vxstart; x < pxstop; ++x, ++px, ++vx)
			    pixels[pt][pz][px] = values[vt][vz][vy][vx];
	    break;
	}
	default: // wrong number of active axes
	    throw new IllegalStateException("Wrong number of active axes");
	}
    }

    /** Sets the pixels in the given borders by mirroring the remainder
	of the image. The size of the borders at the beginning and end
	in any dimension of the image is determined by the corresponding
	field in <code>borders</code>. The part of the image that results from
	ignoring the borders is held to contain the relevant data, which
	is mirrored into the borders.

	@exception IllegalArgumentException if any of the given border
	sizes is larger than half the image size minus one in the
	corresponding dimension.

	@exception NullPointerException if the parameter is
	<code>null</code>.
    */
    public void mirror(final Borders borders) {

	if (borders.x > (xdim-1)/2 || borders.y > (ydim-1)/2 ||
	    borders.z > (zdim-1)/2 || borders.t > (tdim-1)/2)
	    throw new IllegalArgumentException("Border(s) too large");

	// Initialize:
	final int xdimb = xdim - borders.x;
	final int ydimb = ydim - borders.y;
	final int zdimb = zdim - borders.z;
	final int tdimb = tdim - borders.t;

	final int remxdim = xdim - 2*borders.x;
	final int remydim = ydim - 2*borders.y;
	final int remzdim = zdim - 2*borders.z;
	final int remtdim = tdim - 2*borders.t;

	final int remxdim1 = (remxdim == 1) ? remxdim : (remxdim - 1);
	final int remydim1 = (remydim == 1) ? remydim : (remydim - 1);
	final int remzdim1 = (remzdim == 1) ? remzdim : (remzdim - 1);
	final int remtdim1 = (remtdim == 1) ? remtdim : (remtdim - 1);

	final int ifx = (remxdim == 1) ? 1 : 2;
	final int ify = (remydim == 1) ? 1 : 2;
	final int ifz = (remzdim == 1) ? 1 : 2;
	final int ift = (remtdim == 1) ? 1 : 2;

	final int xydim = xdim*ydim;

	// Mirror borders in x-dimension:
	final int byxdim = borders.y*xdim;
	for (int t=borders.t; t<tdimb; ++t)
	    for (int z=borders.z; z<zdimb; ++z) {
		for (int x=0; x<borders.x; ++x) {
		    final int xdiff = x - borders.x;
		    int x0 = xdiff/remxdim1; x0 += x0%ifx;
		    final int xmap = borders.x + Math.abs(xdiff - x0*remxdim1);
		    for (int y=borders.y, xy=byxdim+x, xymap=byxdim+xmap; y<ydimb; ++y, xy+=xdim, xymap+=xdim)
			pixels[t][z][xy] = pixels[t][z][xymap];
		}
		for (int x=xdimb; x<xdim; ++x) {
		    final int xdiff = x - borders.x;
		    int x0 = xdiff/remxdim1; x0 += x0%ifx;
		    final int xmap = borders.x + Math.abs(xdiff - x0*remxdim1);
		    for (int y=borders.y, xy=byxdim+x, xymap=byxdim+xmap; y<ydimb; ++y, xy+=xdim, xymap+=xdim)
			pixels[t][z][xy] = pixels[t][z][xymap];
		}
	    }

	// Mirror borders in y-dimension:
	for (int t=borders.t; t<tdimb; ++t)
	    for (int z=borders.z; z<zdimb; ++z) {
		for (int y=0; y<borders.y; ++y) {
		    final int ydiff = y - borders.y;
		    int y0 = ydiff/remydim1; y0 += y0%ify;
		    final int ymap = borders.y + Math.abs(ydiff - y0*remydim1);
		    for (int x=0, xy=y*xdim, xymap=ymap*xdim; x<xdim; ++x, ++xy, ++xymap)
			pixels[t][z][xy] = pixels[t][z][xymap];
		}
		for (int y=ydimb; y<ydim; ++y) {
		    final int ydiff = y - borders.y;
		    int y0 = ydiff/remydim1; y0 += y0%ify;
		    final int ymap = borders.y + Math.abs(ydiff - y0*remydim1);
		    for (int x=0, xy=y*xdim, xymap=ymap*xdim; x<xdim; ++x, ++xy, ++xymap)
			pixels[t][z][xy] = pixels[t][z][xymap];
		}
	    }

	// Mirror borders in z-dimension:
	for (int t=borders.t; t<tdimb; ++t) {
	    for (int z=0; z<borders.z; ++z) {
		final int zdiff = z - borders.z;
		int z0 = zdiff/remzdim1; z0 += z0%ifz;
		final int zmap = borders.z + Math.abs(zdiff - z0*remzdim1);
		for (int xy=0; xy<xydim; ++xy)
		    pixels[t][z][xy] = pixels[t][zmap][xy];
	    }
	    for (int z=zdimb; z<zdim; ++z) {
		final int zdiff = z - borders.z;
		int z0 = zdiff/remzdim1; z0 += z0%ifz;
		final int zmap = borders.z + Math.abs(zdiff - z0*remzdim1);
		for (int xy=0; xy<xydim; ++xy)
		    pixels[t][z][xy] = pixels[t][zmap][xy];
	    }
	}

	// Mirror borders in t-dimension:
	for (int t=0; t<borders.t; ++t) {
	    final int tdiff = t - borders.t;
	    int t0 = tdiff/remtdim1; t0 += t0%ift;
	    final int tmap = borders.t + Math.abs(tdiff - t0*remtdim1);
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy)
		    pixels[t][z][xy] = pixels[tmap][z][xy];
	}
	for (int t=tdimb; t<tdim; ++t) {
	    final int tdiff = t - borders.t;
	    int t0 = tdiff/remtdim1; t0 += t0%ift;
	    final int tmap = borders.t + Math.abs(tdiff - t0*remtdim1);
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy)
		    pixels[t][z][xy] = pixels[tmap][z][xy];
	}
    }
    
    /** Replaces each pixel value by its absolute value. This is done
        simply by applying method <code>Math.abs()</code> from the Java
        language to every pixel value. */
    public void abs() {

	final int xydim = xdim*ydim;
	for (int t=0; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy)
		    pixels[t][z][xy] = Math.abs(pixels[t][z][xy]);
    }

    /** Adds the given value to every pixel. */
    public void add(final float value) {

	final int xydim = xdim*ydim;
	for (int t=0; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy)
		    pixels[t][z][xy] += value;
    }

    /** Adds the given image pixelwise to the current image.

	@exception IllegalArgumentException if the size of the given
	image is not equal to that of the current image in every
	dimension.

	@exception NullPointerException if the parameter is
	<code>null</code>.
    */
    public void add(final FloatImage image) {

	if (xdim != image.xdim || ydim != image.ydim ||
	    zdim != image.zdim || tdim != image.tdim)
	    throw new IllegalArgumentException("Unequal image sizes");

	final int xydim = xdim*ydim;
	for (int t=0; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy)
		    pixels[t][z][xy] += image.pixels[t][z][xy];
    }

    /** Adds the given image pixelwise to the current image. Note that
        the pixel values in the given image are taken in the unsigned
        sense and thus are always in the range [0,255].

	@exception IllegalArgumentException if the size of the given
	image is not equal to that of the current image in every
	dimension.

	@exception NullPointerException if the parameter is
	<code>null</code>.
    */
    public void add(final ByteImage image) {

	if (xdim != image.xdim || ydim != image.ydim ||
	    zdim != image.zdim || tdim != image.tdim)
	    throw new IllegalArgumentException("Unequal image sizes");

	final int xydim = xdim*ydim;
	for (int t=0; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy)
		    pixels[t][z][xy] += 0xFF & image.pixels[t][z][xy];
    }

    /** Adds the given image pixelwise to the current image. Note that
        the pixel values in the given image are taken in the unsigned
        sense and thus are always in the range [0,65535].

	@exception IllegalArgumentException if the size of the given
	image is not equal to that of the current image in every
	dimension.

	@exception NullPointerException if the parameter is
	<code>null</code>.
    */
    public void add(final ShortImage image) {

	if (xdim != image.xdim || ydim != image.ydim ||
	    zdim != image.zdim || tdim != image.tdim)
	    throw new IllegalArgumentException("Unequal image sizes");

	final int xydim = xdim*ydim;
	for (int t=0; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy)
		    pixels[t][z][xy] += 0xFFFF & image.pixels[t][z][xy];
    }

    /** Subtracts the given value from every pixel. */
    public void subtract(final float value) {

	final int xydim = xdim*ydim;
	for (int t=0; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy)
		    pixels[t][z][xy] -= value;
    }

    /** Subtracts the given image pixelwise from the current image.

	@exception IllegalArgumentException if the size of the given
	image is not equal to that of the current image in every
	dimension.

	@exception NullPointerException if the parameter is
	<code>null</code>.
    */
    public void subtract(final FloatImage image) {

	if (xdim != image.xdim || ydim != image.ydim ||
	    zdim != image.zdim || tdim != image.tdim)
	    throw new IllegalArgumentException("Unequal image sizes");

	final int xydim = xdim*ydim;
	for (int t=0; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy)
		    pixels[t][z][xy] -= image.pixels[t][z][xy];
    }

    /** Subtracts the given image pixelwise from the current image. Note
        that the pixel values in the given image are taken in the
        unsigned sense and thus are always in the range [0,255].

	@exception IllegalArgumentException if the size of the given
	image is not equal to that of the current image in every
	dimension.

	@exception NullPointerException if the parameter is
	<code>null</code>.
    */
    public void subtract(final ByteImage image) {

	if (xdim != image.xdim || ydim != image.ydim ||
	    zdim != image.zdim || tdim != image.tdim)
	    throw new IllegalArgumentException("Unequal image sizes");

	final int xydim = xdim*ydim;
	for (int t=0; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy)
		    pixels[t][z][xy] -= 0xFF & image.pixels[t][z][xy];
    }

    /** Subtracts the given image pixelwise from the current image. Note
        that the pixel values in the given image are taken in the
        unsigned sense and thus are always in the range [0,65535].

	@exception IllegalArgumentException if the size of the given
	image is not equal to that of the current image in every
	dimension.

	@exception NullPointerException if the parameter is
	<code>null</code>.
    */
    public void subtract(final ShortImage image) {

	if (xdim != image.xdim || ydim != image.ydim ||
	    zdim != image.zdim || tdim != image.tdim)
	    throw new IllegalArgumentException("Unequal image sizes");

	final int xydim = xdim*ydim;
	for (int t=0; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy)
		    pixels[t][z][xy] -= 0xFFFF & image.pixels[t][z][xy];
    }

    /** Multiplies each pixel with the given value. */
    public void multiply(final float value) {

	final int xydim = xdim*ydim;
	for (int t=0; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy)
		    pixels[t][z][xy] *= value;
    }

    /** Multiplies the current image pixelwise with the given image.

	@exception IllegalArgumentException if the size of the given
	image is not equal to that of the current image in every
	dimension.

	@exception NullPointerException if the parameter is
	<code>null</code>.
    */
    public void multiply(final FloatImage image) {

	if (xdim != image.xdim || ydim != image.ydim ||
	    zdim != image.zdim || tdim != image.tdim)
	    throw new IllegalArgumentException("Unequal image sizes");

	final int xydim = xdim*ydim;
	for (int t=0; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy)
		    pixels[t][z][xy] *= image.pixels[t][z][xy];
    }

    /** Multiplies the current image pixelwise with the given
        image. Note that the pixel values in the given image are taken
        in the unsigned sense and thus are always in the range [0,255].

	@exception IllegalArgumentException if the size of the given
	image is not equal to that of the current image in every
	dimension.

	@exception NullPointerException if the parameter is
	<code>null</code>.
    */
    public void multiply(final ByteImage image) {

	if (xdim != image.xdim || ydim != image.ydim ||
	    zdim != image.zdim || tdim != image.tdim)
	    throw new IllegalArgumentException("Unequal image sizes");

	final int xydim = xdim*ydim;
	for (int t=0; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy)
		    pixels[t][z][xy] *= 0xFF & image.pixels[t][z][xy];
    }

    /** Multiplies the current image pixelwise with the given
        image. Note that the pixel values in the given image are taken
        in the unsigned sense and thus are always in the range
        [0,65535].

	@exception IllegalArgumentException if the size of the given
	image is not equal to that of the current image in every
	dimension.

	@exception NullPointerException if the parameter is
	<code>null</code>.
    */
    public void multiply(final ShortImage image) {

	if (xdim != image.xdim || ydim != image.ydim ||
	    zdim != image.zdim || tdim != image.tdim)
	    throw new IllegalArgumentException("Unequal image sizes");

	final int xydim = xdim*ydim;
	for (int t=0; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy)
		    pixels[t][z][xy] *= 0xFFFF & image.pixels[t][z][xy];
    }

    /** Divides each pixel by the given value. */
    public void divide(final float value) {

	final int xydim = xdim*ydim;
	for (int t=0; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy)
		    pixels[t][z][xy] /= value;
    }

    /** Divides the current image pixelwise by the given image.

	@exception IllegalArgumentException if the size of the given
	image is not equal to that of the current image in every
	dimension.

	@exception NullPointerException if the parameter is
	<code>null</code>.
    */
    public void divide(final FloatImage image) {

	if (xdim != image.xdim || ydim != image.ydim ||
	    zdim != image.zdim || tdim != image.tdim)
	    throw new IllegalArgumentException("Unequal image sizes");

	final int xydim = xdim*ydim;
	for (int t=0; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy)
		    pixels[t][z][xy] /= image.pixels[t][z][xy];
    }

    /** Divides the current image pixelwise by the given image. Note
        that the pixel values in the given image are taken in the
        unsigned sense and thus are always in the range [0,255].

	@exception IllegalArgumentException if the size of the given
	image is not equal to that of the current image in every
	dimension.

	@exception NullPointerException if the parameter is
	<code>null</code>.
    */
    public void divide(final ByteImage image) {

	if (xdim != image.xdim || ydim != image.ydim ||
	    zdim != image.zdim || tdim != image.tdim)
	    throw new IllegalArgumentException("Unequal image sizes");

	final int xydim = xdim*ydim;
	for (int t=0; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy)
		    pixels[t][z][xy] /= 0xFF & image.pixels[t][z][xy];
    }

    /** Divides the current image pixelwise by the given image. Note
        that the pixel values in the given image are taken in the
        unsigned sense and thus are always in the range [0,65535].

	@exception IllegalArgumentException if the size of the given
	image is not equal to that of the current image in every
	dimension.

	@exception NullPointerException if the parameter is
	<code>null</code>.
    */
    public void divide(final ShortImage image) {

	if (xdim != image.xdim || ydim != image.ydim ||
	    zdim != image.zdim || tdim != image.tdim)
	    throw new IllegalArgumentException("Unequal image sizes");

	final int xydim = xdim*ydim;
	for (int t=0; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy)
		    pixels[t][z][xy] /= 0xFFFF & image.pixels[t][z][xy];
    }

    /** Replaces each pixel by its square value. */
    public void square() {

	final int xydim = xdim*ydim;
	for (int t=0; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy)
		    pixels[t][z][xy] *= pixels[t][z][xy];
    }

    /** Replaces each pixel by its square root. This is done simply by
        applying the method <code>Math.sqrt()</code> from the Java
        language to every pixel value. */
    public void sqrt() {

	final int xydim = xdim*ydim;
	for (int t=0; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy)
		    pixels[t][z][xy] = (float)Math.sqrt(pixels[t][z][xy]);
    }

    /** Replaces each pixel value by its value raised to the given
        power. This is done simply by applying the method
        <code>Math.pow()</code> from the Java language to every
        pixel value. */
    public void power(final float power) {

	final int xydim = xdim*ydim;
	for (int t=0; t<tdim; ++t)
	    for (int z=0; z<zdim; ++z)
		for (int xy=0; xy<xydim; ++xy)
		    pixels[t][z][xy] = (float)Math.pow(pixels[t][z][xy],power);
    }

} 
