// **********************************************************************
// COPYRIGHT NOTICE AND DISCLAIMER
//
// Copyright (c) Erik Meijering. All rights reserved.
//
// Permission to copy and use this software and accompanying
// documentation for educational, research, and not-for-profit
// purposes, without fee and without a signed licensing agreement,
// is hereby granted, provided that the above copyright notice, this
// paragraph and the following two paragraphs appear in all copies.
// The copyright holder is free to make upgraded or improved versions
// of the software available for a fee or commercially only.
//
// IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE TO ANY PARTY FOR
// DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
// OF ANY KIND WHATSOEVER, ARISING OUT OF THE USE OF THIS SOFTWARE
// AND ITS DOCUMENTATION, EVEN IF HE HAS BEEN ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//
// THE COPYRIGHT HOLDER SPECIFICALLY DISCLAIMS ANY WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
// SOFTWARE AND ACCOMPANYING DOCUMENTATION IS PROVIDED "AS IS".
// THE COPYRIGHT HOLDER HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
// SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
//
// **********************************************************************

package imagescience.images;

import ij.ImageStack;

/** Abstract base class for images of up to four dimensions. */
public abstract class Image {
    
    // Properties:
    final int xdim;
    final int ydim;
    final int zdim;
    final int tdim;

    float xpixdim = 1.0f;
    float ypixdim = 1.0f;
    float zpixdim = 1.0f;
    float tpixdim = 1.0f;

    int axes = 0;

    static int lastID;
    final int thisID;

    /** Constructor for creating a four-dimensional image. Serves
	merely to initialize fields common to all subclasses.

	@exception IllegalArgumentException if any of the parameters is
	zero or negative.
    */
    public Image(final int xdim, final int ydim, final int zdim, final int tdim) {
	
	if (xdim <= 0 || ydim <= 0 || zdim <= 0 || tdim <= 0)
	    throw new IllegalArgumentException("Negative or zero dimension(s)");

	this.xdim = xdim;
	this.ydim = ydim;
	this.zdim = zdim;
	this.tdim = tdim;

	thisID = ++lastID;
    }

    /** Returns the image in the form of an <code>ImageStack</code>
        object. Note that the fourth dimension is dissolved into the
        third dimension. Also note that the actual pixel data in the
        image is not copied to, but shared by the returned
        <code>ImageStack</code>. */
    public abstract ImageStack stack();

    /** Returns the number of pixels in the image. */
    public int pixelcount() { return xdim*ydim*zdim*tdim; }

    /** Sets the pixel sizes to the given values. By default, the
        pixel size is 1.0 in all dimensions. */
    public void pixelsizes(final PixelSizes pixelsizes) {

	xpixdim = pixelsizes.x;
	ypixdim = pixelsizes.y;
	zpixdim = pixelsizes.z;
	tpixdim = pixelsizes.t;
    }

    /** Returns the pixel sizes as a new <code>PixelSizes</code> object. */
    public PixelSizes pixelsizes() { return new PixelSizes(xpixdim,ypixdim,zpixdim,tpixdim); }

    /** Returns the image's ID. */
    public int id() { return thisID; }

    /** Returns the amount of memory occupied by the pixel data in the
        image. The amount is expressed in kilobytes. */
    public abstract int memory();

    /** Returns the dimensions of the image as a new
        <code>Dimensions</code> object. */
    public Dimensions dimensions() { return new Dimensions(xdim,ydim,zdim,tdim); }

    /** Returns a floating-point representation of the minimum value in
        the image. */
    public abstract float minimum();

    /** Returns a floating-point representation of the maximum value in
        the image. */
    public abstract float maximum();

    /** Sets the currently active Cartesian coordinate axis (axes). This
        method affects the operation of the <code>get</code> and
        <code>set</code> methods for reading and writing pixel values. The
        argument must be one or a combination (by addition) of the
        static fields of class <code>Axes</code>.

	@exception IllegalArgumentException if the value of
	<code>axes</code> does not correspond to a valid combination of
	axes, that is, if it is less than zero or larger than 15.
    */
    public void axes(final int axes) {

	if (axes < Axes.NONE || axes > Axes.ALL)
	    throw new IllegalArgumentException("Non-supported (combination of) axes");

	this.axes = axes;
    }

    /** Returns a floating-point representation of the pixel value at
	the given coordinates.

	@exception ArrayIndexOutOfBoundsException if any of the
	coordinates is out of range.

	@exception NullPointerException if the parameter is <code>null</code>.
    */
    public abstract float get(final Coordinates coords);

    /** Reads pixel values in one dimension into the given array. The
        dimension from which pixel values are read is determined by the
        currently active Cartesian coordinate axis. Exactly one axis
        must be active in order for this method to work. The given
        coordinates indicate the starting position of the given pixel
        array, relative to the image. That is to say, they determine the
        starting (x,y,z,t) position in the image for reading pixel
        values into the array. Note that the coordinate corresponding to
        the active axis is allowed to be negative. However, pixel
        positions outside the image in the active dimension are not
        actually read but simply skipped. This is useful for creating
        borders around the pixel data read into the given array.

	@exception IllegalStateException if more than one coordinate
	axis is active.

	@exception ArrayIndexOutOfBoundsException if the coordinates
        corresponding to the non-active axes are outside the range
        determined by the size of the image.

	@exception NullPointerException if any of the parameters is
	<code>null</code>.
    */
    public abstract void get(final Coordinates coords, final float[] values);

    /** Reads pixel values in two dimensions into the given array. The
        dimensions from which pixel values are read are determined by
        the currently active Cartesian coordinate axes. Exactly two axes
        must be active in order for this method to work. The two indices
        of the given pixel array correspond to these axes, where the
        order of the indices from left to right is from the highest to
        the lowest dimension. The given coordinates indicate the
        starting position of the given pixel array, relative to the
        image. That is to say, they determine the starting (x,y,z,t)
        position in the image for reading pixel values into the
        array. Note that the coordinates corresponding to the active
        axes are allowed to be negative. However, pixel positions
        outside the image in the active dimensions are not actually read
        but simply skipped. This is useful for creating borders around
        the pixel data read into the given array.

	@exception IllegalStateException if not exactly two axes are active.

	@exception ArrayIndexOutOfBoundsException if the coordinates
        corresponding to the non-active axes are outside the range
        determined by the size of the image.

	@exception NullPointerException if any of the parameters is
	<code>null</code>.
    */
    public abstract void get(final Coordinates coords, final float[][] values);

    /** Reads pixel values in three dimensions into the given array. The
        dimensions from which pixel values are read are determined by
        the currently active Cartesian coordinate axes. Exactly three
        axes must be active in order for this method to work. The three
        indices of the given pixel array correspond to these axes, where
        the order of the indices from left to right is from the highest
        to the lowest dimension. The given coordinates indicate the
        starting position of the given pixel array, relative to the
        image. That is to say, they determine the starting (x,y,z,t)
        position in the image for reading pixel values into the
        array. Note that the coordinates corresponding to the active
        axes are allowed to be negative. However, pixel positions
        outside the image in the active dimensions are not actually read
        but simply skipped. This is useful for creating borders around
        the pixel data read into the given array.

	@exception IllegalStateException if not exactly three axes are active.

	@exception ArrayIndexOutOfBoundsException if the coordinate
        corresponding to the non-active axis is outside the range
        determined by the size of the image.

	@exception NullPointerException if any of the parameters is
	<code>null</code>.
    */
    public abstract void get(final Coordinates coords, final float[][][] values);

    /** Reads pixel values in all four dimensions into the given
        array. All four axes must be active in order for this method to
        work. The four indices of the given pixel array correspond to
        these axes, where the order of the indices from left to right is
        t, z, y, x. The given coordinates indicate the starting position
        of the given pixel array, relative to the image. That is to say,
        they determine the starting (x,y,z,t) position in the image for
        reading pixel values into the array. Note that the coordinates
        are allowed to be negative. However, pixel positions outside the
        image are not actually read but simply skipped. This is useful
        for creating borders around the pixel data read into the given
        array.

	@exception IllegalStateException if not all four axes are active.

	@exception NullPointerException if any of the parameters is
	<code>null</code>.
    */
    public abstract void get(final Coordinates coords, final float[][][][] values);

    /** Sets all pixels to the given value. */
    public abstract void set(final float value);

    /** Sets the pixel at the given coordinates to the given value. 

	@exception ArrayIndexOutOfBoundsException if any of the
	coordinates is out of range.

	@exception NullPointerException if the first parameter is
	<code>null</code>.
    */
    public abstract void set(final Coordinates coords, final float value);
    
    /** Writes given pixel values in one dimension into the image. The
        dimension into which pixel values are written is determined by
        the currently active Cartesian coordinate axis. Exactly one axis
        must be active in order for this method to work. The given
        coordinates indicate the starting position of the given pixel
        array, relative to the image. That is to say, they determine the
        starting (x,y,z,t) position in the image for writing pixel
        values. Note that the coordinate corresponding to the active
        axis is allowed to be negative. However, pixel positions outside
        the image in the active dimension are not actually written but
        simply skipped. This is useful for ignoring possible borders
        around the pixel data in the given array.

	@exception IllegalStateException if more than one coordinate
	axis is active.

	@exception ArrayIndexOutOfBoundsException if the coordinates
        corresponding to the non-active axes are outside the range
        determined by the size of the image.

	@exception NullPointerException if any of the parameters is
	<code>null</code>.
    */
    public abstract void set(final Coordinates coords, final float[] values);

    /** Writes given pixel values in two dimensions into the image. The
        dimensions into which pixel values are written are determined by
        the currently active Cartesian coordinate axes. Exactly two axes
        must be active in order for this method to work. The two indices
        of the given pixel array correspond to these axes, where the
        order of the indices from left to right is from the highest to
        the lowest dimension. The given coordinates indicate the
        starting position of the given pixel array, relative to the
        image. That is to say, they determine the starting (x,y,z,t)
        position in the image for writing pixel values. Note that the
        coordinates corresponding to the active axes are allowed to be
        negative. However, pixel positions outside the image in the
        active dimensions are not actually written but simply
        skipped. This is useful for ignoring possible borders around the
        pixel data in the given array.

	@exception IllegalStateException if not exactly two axes are active.

	@exception ArrayIndexOutOfBoundsException if the coordinates
        corresponding to the non-active axes are outside the range
        determined by the size of the image.

	@exception NullPointerException if any of the parameters is
	<code>null</code>.
    */
    public abstract void set(final Coordinates coords, final float[][] values);

    /** Writes given pixel values in three dimensions into the
        image. The dimensions into which pixel values are written are
        determined by the currently active Cartesian coordinate
        axes. Exactly three axes must be active in order for this method
        to work. The three indices of the given pixel array correspond
        to these axes, where the order of the indices from left to right
        is from the highest to the lowest dimension. The given
        coordinates indicate the starting position of the given pixel
        array, relative to the image. That is to say, they determine the
        starting (x,y,z,t) position in the image for writing pixel
        values. Note that the coordinates corresponding to the active
        axes are allowed to be negative. However, pixel positions
        outside the image in the active dimensions are not actually
        written but simply skipped. This is useful for ignoring possible
        borders around the pixel data in the given array.

	@exception IllegalStateException if not exactly three axes are
	active.

	@exception ArrayIndexOutOfBoundsException if the coordinate
        corresponding to the non-active axis is outside the range
        determined by the size of the image.

	@exception NullPointerException if any of the parameters is
	<code>null</code>.
    */
    public abstract void set(final Coordinates coords, final float[][][] values);

    /** Writes given pixel values in all four dimensions into the
        image. All four axes must be active in order for this method to
        work. The four indices of the given pixel array correspond to
        these axes, where the order of the indices from left to right is
        t, z, y, x. The given coordinates indicate the starting position
        of the given pixel array, relative to the image. That is to say,
        they determine the starting (x,y,z,t) position in the image for
        writing pixel values. Note that the coordinates are allowed to
        be negative. However, pixel positions outside the image are not
        actually written but simply skipped. This is useful for ignoring
        possible borders around the pixel data in the given array.

	@exception IllegalStateException if not all four axes are active.

	@exception NullPointerException if any of the parameters is
	<code>null</code>.
    */
    public abstract void set(final Coordinates coords, final float[][][][] values);

} 
