// **********************************************************************
// COPYRIGHT NOTICE AND DISCLAIMER
//
// Copyright (c) Erik Meijering. All rights reserved.
//
// Permission to copy and use this software and accompanying
// documentation for educational, research, and not-for-profit
// purposes, without fee and without a signed licensing agreement,
// is hereby granted, provided that the above copyright notice, this
// paragraph and the following two paragraphs appear in all copies.
// The copyright holder is free to make upgraded or improved versions
// of the software available for a fee or commercially only.
//
// IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE TO ANY PARTY FOR
// DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
// OF ANY KIND WHATSOEVER, ARISING OUT OF THE USE OF THIS SOFTWARE
// AND ITS DOCUMENTATION, EVEN IF HE HAS BEEN ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//
// THE COPYRIGHT HOLDER SPECIFICALLY DISCLAIMS ANY WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
// SOFTWARE AND ACCOMPANYING DOCUMENTATION IS PROVIDED "AS IS".
// THE COPYRIGHT HOLDER HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
// SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
//
// **********************************************************************

package imagescience.utilities;

import ij.IJ;

/** Wrapper around ImageJ's progress indication bar. The advantage of
    using this class is that it <em>always</em> shows the bar,
    regardless of the time difference between the first and second call
    to the <tt>step()</tt> method. This in contrast with ImageJ's
    <tt>IJ.showProgress()</tt> method, which does not show the bar if
    the difference is less than 50msec. Furthermore, this class allows
    for running multiple progress indication processes in parallel
    (which actually makes the bar jump up and down to indicate the
    respective progresses). Finally, it allows to specify the number of
    indication steps (20, by default), thereby limiting the number of
    "under-water" calls to <tt>IJ.showProgress()</tt>, which could be
    relatively expensive (depending on the cost of the operations within
    the loops from where this method is called every time). */
public final class Progressor {

    private static int iNrOfProgressors = 0;
    private int iNrOfSteps = 1;
    private int iStepNr = 0;
    private int iNextIndicationNr = 0;
    private int iNextStepNrForIndication = 0;
    private double dIndicationFraction = 0.0;
    private int iNrOfIndications = 20;
    private String sStatusInfo = "";

    /** Constructor. The progress bar should be visible directly after
        construction. */
    public Progressor() {
	++iNrOfProgressors;
	IJ.showProgress(0.0);
	IJ.wait(50);
	IJ.showProgress(0.0);
    }

    /** Constructor. The progress bar should be visible directly after
        construction. The parameter should be a few words describing the
        process whose progress will be indicated, and appears in the
        status bar. */
    public Progressor(final String statusInfo) {
	++iNrOfProgressors;
	IJ.showProgress(0.0);
	IJ.wait(50);
	IJ.showProgress(0.0);
	if (statusInfo != null) sStatusInfo = statusInfo;
	IJ.showStatus(sStatusInfo);
    }

    /** Method to specify the number of progress steps as displayed by
        the progress bar. The default number of steps is 20.

	@exception IllegalArgumentException if the parameter is zero or
	negative.
    */
    public void indications(final int nrOfIndications) {
	if (nrOfIndications <= 0)
	    throw new IllegalArgumentException("Number of indications zero or negative");
	iNrOfIndications = nrOfIndications;
    }
	
    /** Method to specify the total number of steps in the
        process. Should be called before starting the process whose
        progress is to be indicated.

	@exception IllegalArgumentException if the parameter is negative.
    */
    public void steps(final int nrOfSteps) {
	if (nrOfSteps < 0)
	    throw new IllegalArgumentException("Negative number of steps");
	iNrOfSteps = nrOfSteps;
	dIndicationFraction = (double)iNrOfSteps/(double)iNrOfIndications;
	iNextIndicationNr = 1;
	iNextStepNrForIndication = (int)dIndicationFraction;
    }
	
    /** Method to increase the internal step counter. The total number
        of calls to this method by the end of a process should be equal
        to the number given as parameter to method <tt>steps</tt> prior to
        starting the process. */
    public void step() {
	++iStepNr;
	if (iStepNr >= iNextStepNrForIndication) {
	    IJ.showProgress(((double)iStepNr)/((double)iNrOfSteps));
	    IJ.showStatus(sStatusInfo);
	    ++iNextIndicationNr;
	    iNextStepNrForIndication = (int)(dIndicationFraction*iNextIndicationNr);
	}
    }
	
    /** Method to increase the internal step counter by the given
        amount. By the end of a process, the sum of the parameter values
        of all calls to this method should be equal to the number given
        as parameter to method <tt>steps</tt> prior to starting the
        process.

	@exception IllegalArgumentException if the parameter is zero or
	negative.
    */
    public void step(final int nrOfSteps) {
	if (nrOfSteps <= 0)
	    throw new IllegalArgumentException("Number of steps zero or negative");
	iStepNr += nrOfSteps;
	if (iStepNr >= iNextStepNrForIndication) {
	    final double dStepFraction = (double)iStepNr/(double)iNrOfSteps;
	    IJ.showProgress(dStepFraction);
	    IJ.showStatus(sStatusInfo);
	    iNextIndicationNr = (int)(iNrOfIndications*dStepFraction) + 1;
	    iNextStepNrForIndication = (int)(dIndicationFraction*iNextIndicationNr);
	}
    }
	
    /** Method to reset the internal step and indication counters. Can
        be used in cases where the same <tt>Progressor</tt> object is to
        be used for several subprocesses in succession, and where the
        subprocesses all have the same number of steps. */
    public void reset() {
	iStepNr = 0;
	iNextIndicationNr = 1;
	iNextStepNrForIndication = (int)dIndicationFraction;
	IJ.showProgress(0.0);
    }
	
    /** Method to change the information appearing in the status bar. */
    public void status(final String statusInfo) {
	if (statusInfo != null) sStatusInfo = statusInfo;
	IJ.showStatus(sStatusInfo);
    }
	
    /** Method to end the progress indication. This method should be
        called at the end of the process whose progress was
        indicated. The progress bar disappears only after all
        <tt>Progressor</tt> objects have been <tt>end</tt>ed. */
    public void end() {
	IJ.showStatus("");
	--iNrOfProgressors;
	if (iNrOfProgressors <= 0)
	    IJ.showProgress(1.0);
    }

}
