//**********************************************************************
//COPYRIGHT NOTICE AND DISCLAIMER
//
//Copyright (c) Erik Meijering. All rights reserved.
//
//Permission to copy and use this software and accompanying
//documentation for educational, research, and not-for-profit
//purposes, without fee and without a signed licensing agreement,
//is hereby granted, provided that the above copyright notice, this
//paragraph and the following two paragraphs appear in all copies.
//The copyright holder is free to make upgraded or improved versions
//of the software available for a fee or commercially only.
//
//IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE TO ANY PARTY FOR
//DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
//OF ANY KIND WHATSOEVER, ARISING OUT OF THE USE OF THIS SOFTWARE
//AND ITS DOCUMENTATION, EVEN IF HE HAS BEEN ADVISED OF THE
//POSSIBILITY OF SUCH DAMAGE.
//
//THE COPYRIGHT HOLDER SPECIFICALLY DISCLAIMS ANY WARRANTIES,
//INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
//MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
//SOFTWARE AND ACCOMPANYING DOCUMENTATION IS PROVIDED "AS IS".
//THE COPYRIGHT HOLDER HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
//SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
//
//**********************************************************************
package neuronJgui;

import ij.IJ;
import ij.ImagePlus;
import ij.gui.ImageCanvas;
import ij.gui.ImageWindow;
import ij.gui.Roi;
import ij.gui.Toolbar;
import ij.process.ByteProcessor;
import imagescience.utilities.FMath;

import java.awt.Color;
import java.awt.Cursor;
import java.awt.Event;
import java.awt.FileDialog;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.image.IndexColorModel;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;

public final class TracingHandler extends Roi implements KeyListener, MouseListener, MouseMotionListener {

	
    private ImagePlus imp;
    private ImageCanvas imc;
    private ImageWindow imw;
    private ByteProcessor ipG8;

    private final Dijkstra dijkstra = new Dijkstra();
    private float[][][] costs;
    private byte[][] dirsimage;

    private final Point clckPoint = new Point();
    private final Point currPoint = new Point();
    private final Point mousPoint = new Point();
    private final Point snapPoint = new Point();
    private final Point scrlPoint = new Point();
	
    private Tracings tracings = new Tracings();
    private Tracing currTracing;
    private Segment currSegment = new Segment();
    private Segment ssmpSegment = new Segment();
	
    private boolean bTracingActive;
    private boolean bManualTracing;
    private boolean bSnapCursor;
    private boolean bSmoothSegment;
    private boolean bComputedCosts;
    private boolean bDijkstra;
	
    private int iXSize, iYSize;

    private long lastClckTime = System.currentTimeMillis();

    TracingHandler() { super(0,0,0,0); }

    void attach(final ImagePlus impNew) {

	// Copy handles:
	imp = impNew;
	imw = imp.getWindow();
	imc = imw.getCanvas();

	// Create a copy that is surely a gray-scale image (the pixels
	// are already of type byte, but may represent color indices,
	// not actual gray-values):
	NJ.write("Creating gray-scale copy of new image...");
	iXSize = imp.getWidth(); iYSize = imp.getHeight();
	final ByteProcessor ipIn = (ByteProcessor)imp.getProcessor();
	final IndexColorModel icm = (IndexColorModel)ipIn.getColorModel();
	final int iMapSize = icm.getMapSize();
	final byte[] r = new byte[iMapSize]; icm.getReds(r);
	final byte[] g = new byte[iMapSize]; icm.getGreens(g);
	final byte[] b = new byte[iMapSize]; icm.getBlues(b);
	try {
	    ipG8 = new ByteProcessor(iXSize,iYSize);
	    final byte[] g8pxs = (byte[])ipG8.getPixels();
	    final byte[] inpxs = (byte[])ipIn.getPixels();
	    final int nrpxs = inpxs.length;
	    for (int i=0; i<nrpxs; ++i) {
		final int index = inpxs[i]&0xFF;
		g8pxs[i] = (byte)FMath.round((r[index]&0xFF)*0.3 + (g[index]&0xFF)*0.6 + (b[index]&0xFF)*0.1);
	    }
	} catch (OutOfMemoryError e) {
	    NJ.outOfMemory();
	    ipG8 = null;
	}

	// Remove and add listeners from and to canvas:
	NJ.write("Detaching ImageJ listeners...");
	imw.removeKeyListener(IJ.getInstance());
	imc.removeKeyListener(IJ.getInstance());
	imc.removeMouseListener(imc);
	imc.removeMouseMotionListener(imc);
	NJ.write("Attaching NeuronJ listeners...");
	imc.addMouseMotionListener(this);
	imc.addMouseListener(this);
	imc.addKeyListener(this);
	imc.requestFocus();
	NJ.write("Done");

	// Reset variables:
	costs = null;
	dirsimage = null;
	tracings.reset();
	Tracing.resetID();
	currSegment.reset();
	ssmpSegment.reset();
	currPoint.setLocation(-100,-100);
	bTracingActive = false;
	bManualTracing = false;
	bSnapCursor = true;
	bSmoothSegment = true;
	bComputedCosts = false;
	bDijkstra = false;

	// Enable displaying tracings:
	imp.setRoi(this);
    }
    
    void restoreListeners() {

	// Remove and add listeners from and to canvas:
	NJ.write("Detaching NeuronJ listeners from current image...");
	imc.removeKeyListener(this);
	imc.removeMouseListener(this);
	imc.removeMouseMotionListener(this);
	NJ.write("Attaching ImageJ listeners to current image...");
	imc.addMouseMotionListener(imc);
	imc.addMouseListener(imc);
	imc.addKeyListener(IJ.getInstance());
	imw.addKeyListener(IJ.getInstance());

	// Stop displaying tracings:
	imp.killRoi();
    }

    void computeCosts() {
	final Costs ci = new Costs();
	final long lStartTime = System.currentTimeMillis();
	try {
	    if (ipG8 != null) costs = ci.run(ipG8,NJ.scale);
	    else throw new OutOfMemoryError();
	    NJ.write("Finished in "+IJ.d2s((System.currentTimeMillis()-lStartTime)/1000f,2)+" seconds");
	    bComputedCosts = true;
	} catch (OutOfMemoryError e) {
	    NJ.outOfMemory();
	    NJ.ntb.setTool(NJ.ntb.MAGNIFY);
	}
	IJ.showStatus(NJ.COPYRIGHT);
    }
    
    boolean computedCosts() { return bComputedCosts; }

    void doDijkstra() {
	if (bTracingActive) bDijkstra = true;
    }

    Tracings tracings() { return tracings; }

    void redraw() { imc.repaint(); }
    
    public void draw(final Graphics g) {
	
	// Draw finished tracings:
	tracings.draw(g,imc);

	// Draw currently active tracing and segment:
	if (bTracingActive) {
	    currTracing.draw(g,imc);
	    currSegment.draw(g,imc,NJ.ACTIVECOLOR);
	}

	// Draw currPoint cursor:
	g.setColor(Color.red);
	final double mag = imc.getMagnification();
	final int csx = imc.screenX(currPoint.x) + (int)(mag/2.0);
	final int csy = imc.screenY(currPoint.y) + (int)(mag/2.0);
	g.drawLine(csx,csy-5,csx,csy+5);
	g.drawLine(csx-5,csy,csx+5,csy);
    }

    public void keyPressed(final KeyEvent e) {

	final int iKeyCode = e.getKeyCode();

	if (iKeyCode == KeyEvent.VK_C && costs != null && NJ.hkeys) {
	    final ByteProcessor ip = new ByteProcessor(iXSize,iYSize);
	    final byte[] pixels = (byte[])ip.getPixels();
	    for (int y=0, i=0; y<iYSize; ++y)
		for (int x=0; x<iXSize; ++x, ++i)
		    pixels[i] = (byte)costs[0][y][x];
	    final ImagePlus tmp = new ImagePlus("NeuronJ: Costs",ip);
	    tmp.show(); tmp.updateAndRepaintWindow();
	}
	else if (iKeyCode == KeyEvent.VK_D && dirsimage != null && NJ.hkeys) {
	    final ByteProcessor ip = new ByteProcessor(iXSize,iYSize);
	    final byte[] pixels = (byte[])ip.getPixels();
	    for (int y=0, i=0; y<iYSize; ++y)
		for (int x=0; x<iXSize; ++x, ++i)
		    pixels[i] = (byte)(31*(0xFF & dirsimage[y][x]));
	    final ImagePlus tmp = new ImagePlus("NeuronJ: Directions",ip);
	    tmp.show(); tmp.updateAndRepaintWindow();
	}
	else if (iKeyCode == KeyEvent.VK_V && costs != null && NJ.hkeys) {
	    final ByteProcessor ip = new ByteProcessor(iXSize,iYSize);
	    final byte[] pixels = (byte[])ip.getPixels();
	    for (int y=0, i=0; y<iYSize; ++y)
		for (int x=0; x<iXSize; ++x, ++i)
		    pixels[i] = (byte)(255.0f - costs[0][y][x]);
	    final ImagePlus tmp = new ImagePlus("NeuronJ: Vectors",ip);
	    tmp.show(); tmp.updateAndRepaintWindow();
	    final VectorField vf = new VectorField(tmp,costs);
	}
	else if (iKeyCode == KeyEvent.VK_CONTROL && bSnapCursor) {
	    NJ.write("Switching off local snapping");
	    currPoint.x = mousPoint.x;
	    currPoint.y = mousPoint.y;
	    bSnapCursor = false;
	    if (bTracingActive) updateCurrSegment();
	    redraw();
	}
	else if (iKeyCode == KeyEvent.VK_SHIFT && !bManualTracing) {
	    NJ.write("Switching to manual tracing mode");
	    bManualTracing = true;
	    if (bTracingActive) updateCurrSegment();
	    redraw();
	}
	else if (iKeyCode == KeyEvent.VK_S && bSmoothSegment && NJ.hkeys) {
	    NJ.write("Disabling segment smoothing");
	    bSmoothSegment = false;
	    if (bTracingActive) updateCurrSegment();
	    redraw();
	}
	else if ((iKeyCode == KeyEvent.VK_TAB || iKeyCode == KeyEvent.VK_SPACE) && bTracingActive) {
	    bTracingActive = false;
	    bDijkstra = false;
	    finishCurrSegment();
	    finishCurrTracing();
	    redraw();
	}
	else if (iKeyCode == KeyEvent.VK_LEFT && NJ.workimages != null) {
	    if (NJ.workimagenr <= 0)
		NJ.notify("The current image is the first image.");
	    else {
		--NJ.workimagenr;
		NJ.write("Request to load image "+NJ.workdir+NJ.workimages[NJ.workimagenr]);
		NJ.ntb.loadImage(NJ.workdir,NJ.workimages[NJ.workimagenr]);
		NJ.ntb.setTool(NJ.ntb.MAGNIFY);
	    }
	}
	else if (iKeyCode == KeyEvent.VK_RIGHT && NJ.workimages != null) {
	    if (NJ.workimagenr >= NJ.workimages.length-1)
		NJ.notify("The current image is the last image.");
	    else {
		++NJ.workimagenr;
		NJ.write("Request to load image "+NJ.workdir+NJ.workimages[NJ.workimagenr]);
		NJ.ntb.loadImage(NJ.workdir,NJ.workimages[NJ.workimagenr]);
		NJ.ntb.setTool(NJ.ntb.MAGNIFY);
	    }
	}
    }

    public void keyReleased(final KeyEvent e) {

	final int iKeyCode = e.getKeyCode();

	if (iKeyCode == KeyEvent.VK_CONTROL) {
	    NJ.write("Switching on local snapping");
	    currPoint.x = snapPoint.x;
	    currPoint.y = snapPoint.y;
	    bSnapCursor = true;
	    if (bTracingActive) updateCurrSegment();
	    redraw();
	}
	else if (iKeyCode == KeyEvent.VK_SHIFT) {
	    NJ.write("Back to automatic tracing mode");
	    bManualTracing = false;
	    if (bTracingActive) updateCurrSegment();
	    redraw();
	}
	else if (iKeyCode == KeyEvent.VK_S) {
	    NJ.write("Enabling segment smoothing");
	    bSmoothSegment = true;
	    if (bTracingActive) updateCurrSegment();
	    redraw();
	}
    }

    public void keyTyped(final KeyEvent e) {}
    
    public void mouseClicked(final MouseEvent e) {}
	
    public void mouseDragged(final MouseEvent e) {

        final int x = e.getX();
        final int y = e.getY();

	switch (NJ.ntb.currentTool()) {
	case TracingToolbar.SCROLL: { scroll(x,y); break; }
	case TracingToolbar.RECTANGLE: {
		imc.mouseDragged(e);
		break;
	}
        }
    }

    public void mouseEntered(final MouseEvent e) { /** imw.toFront(); imc.requestFocus(); */ }

    public void mouseExited(final MouseEvent e) { IJ.showStatus(NJ.COPYRIGHT); }
    
    public void mouseMoved(final MouseEvent e) {
		
	final int x = imc.offScreenX(e.getX());
	final int y = imc.offScreenY(e.getY());

	showValue(x,y);

	switch (NJ.ntb.currentTool()) {
	case TracingToolbar.ADD: {
	    final int prevMouseX = mousPoint.x;
	    final int prevMouseY = mousPoint.y;
	    mousPoint.x = x;
	    mousPoint.y = y;

	    // Move away from the border (the Dijkstra algorithm does
	    // not allow the starting point to be on the border):
	    if (mousPoint.x == 0) ++mousPoint.x;
	    else if (mousPoint.x == iXSize-1) --mousPoint.x;
	    if (mousPoint.y == 0) ++mousPoint.y;
	    else if (mousPoint.y == iYSize-1) --mousPoint.y;

	    // If the mouse point is still on the same pixel, there is
	    // no need to do anything (this prevents superfluous screen
	    // refreshments at zoom levels > 100%):
	    if (prevMouseX != mousPoint.x || prevMouseY != mousPoint.y) {

		snapPoint.x = currPoint.x = mousPoint.x;
		snapPoint.y = currPoint.y = mousPoint.y;
		
		// Update directions map if necessary:
		if (bDijkstra) {
		    bDijkstra = false;
		    NJ.write("Computing shortest paths to clicked point...");
		    IJ.showStatus("Computing optimal paths");
		    final long lStartTime = System.currentTimeMillis();
		    dirsimage = dijkstra.run(costs,clckPoint);
		    NJ.write("Finished in "+IJ.d2s((System.currentTimeMillis()-lStartTime)/1000f,2)+" seconds");
		    IJ.showStatus(NJ.COPYRIGHT);
		}
		
		// Compute locally lowest cost point for snapping:
		int startx = mousPoint.x - NJ.snaprange; if (startx < 1) startx = 1;
		int starty = mousPoint.y - NJ.snaprange; if (starty < 1) starty = 1;
		int stopx = mousPoint.x + NJ.snaprange; if (stopx > iXSize-2) stopx = iXSize-2;
		int stopy = mousPoint.y + NJ.snaprange; if (stopy > iYSize-2) stopy = iYSize-2;
		for (int sy=starty; sy<=stopy; ++sy)
		    for (int sx=startx; sx<=stopx; ++sx)
			if (costs[0][sy][sx] < costs[0][snapPoint.y][snapPoint.x]) {
			    snapPoint.x = sx;
			    snapPoint.y = sy;
			}
		
		// Snap if requested:
		if (bSnapCursor) {
		    currPoint.x = snapPoint.x;
		    currPoint.y = snapPoint.y;
		}
		
		if (bTracingActive) updateCurrSegment();
		
		// Draw:
		redraw();
	    }

	    break;
	}
	case TracingToolbar.DELETE:
	case TracingToolbar.ATTRIBS: {
	    final Point tempPoint = new Point();
	    tempPoint.x = x;
	    tempPoint.y = y;
	    final int nrtracings = tracings.nrtracings();
	    for (int i=0; i<nrtracings; ++i) tracings.element(i).highlight(false);
	    int nindex;
	    for (nindex=0; nindex<nrtracings; ++nindex) {
		final Tracing tracing = tracings.element(nindex);
		if (tracing.nearby(tempPoint)) { tracing.highlight(true); break; }
	    }
	    if (tracings.changed()) {
		if (NJ.adg != null)
		    if (nindex < nrtracings) NJ.adg.select(nindex+1);
		    else NJ.adg.select(0);
		redraw();
	    }
	    break;
	}
	case TracingToolbar.RECTANGLE: {
		imc.mouseMoved(e);
		break;
	}
	}
    }
    
  
    public void mousePressed(final MouseEvent e) {
	
	final int x = e.getX();
	final int y = e.getY();
	
	switch (NJ.ntb.currentTool()) {
	case TracingToolbar.SCROLL: {
	    scrlPoint.x = imc.offScreenX(x);
	    scrlPoint.y = imc.offScreenY(y);
	    break;
	}
	case TracingToolbar.MAGNIFY: {
	    final int flags = e.getModifiers();
	    if ((flags & (Event.ALT_MASK | Event.META_MASK | Event.CTRL_MASK)) != 0) imc.zoomOut(x,y);
	    else imc.zoomIn(x,y);
	    showValue(imc.offScreenX(x),imc.offScreenY(y));
	    break;
	}
	case TracingToolbar.RECTANGLE: {
		imp.killRoi();
		imp.createNewRoi(x,y);
		try {
			Toolbar.getInstance().setTool(Toolbar.RECTANGLE);
		} catch (Exception exp) {
			
		}
	   imc.mousePressed(e);
	   break;
	}
	case TracingToolbar.ADD: {
	    final long currClckTime = System.currentTimeMillis();
	    final int prevClckX = clckPoint.x;
	    final int prevClckY = clckPoint.y;
	    clckPoint.x = currPoint.x;
	    clckPoint.y = currPoint.y;
	    NJ.write("Clicked point ("+clckPoint.x+","+clckPoint.y+")");

	    if (!bTracingActive) {
		currTracing = new Tracing();
		bTracingActive = true;
		bDijkstra = true;
	    } else {
		finishCurrSegment();
		bDijkstra = true;

		if ((currClckTime - lastClckTime < 500) &&
		    (prevClckX == clckPoint.x && prevClckY == clckPoint.y)) {
		    finishCurrTracing();
		    bTracingActive = false;
		    bDijkstra = false;
		}
	    }

	    lastClckTime = currClckTime;
	    break;
	}
	case TracingToolbar.DELETE: {
	    final int nrtracings = tracings.nrtracings();
	    for (int w=0; w<nrtracings; ++w) {
		final Tracing tracing = tracings.element(w);
		if (tracing.highlighted()) {
		    final YesNoDialog ynd =
			new YesNoDialog("Delete","Do you really want to delete this tracing?");
		    if (ynd.yesPressed()) {
			NJ.write("Deleting tracing N"+tracing.id());
			tracings.remove(w);
			IJ.showStatus("Deleted tracing");
			if (NJ.adg != null) NJ.adg.reset();
		    } else {
			tracing.highlight(false);
			IJ.showStatus(NJ.COPYRIGHT);
			if (NJ.adg != null) NJ.adg.select(0);
		    }
		    break;
		}
	    }
	    break;
	}
	case TracingToolbar.ATTRIBS: {
	    final int nrtracings = tracings.nrtracings();
	    for (int i=0; i<nrtracings; ++i) {
		final Tracing tracing = tracings.element(i);
		if (tracing.highlighted()) {
		    if (!tracing.selected()) {
			NJ.write("Selecting tracing N"+tracing.id());
			tracing.select(true);
			IJ.showStatus("Selected tracing");
		    } else {
			NJ.write("Deselecting tracing N"+tracing.id());
			tracing.select(false);
			tracing.highlight(false);
			if (NJ.adg != null) NJ.adg.select(0);
			IJ.showStatus("Deselected tracing");
		    }
		    break;
		}
	    }
	    break;
	}
	}

	redraw();
    }

    private void finishCurrTracing() {

	if (currTracing.nrsegments() == 0) {
	    NJ.write("Dumping tracing of zero units length");
	    IJ.showStatus("Dumped tracing");
	} else {
	    NJ.write("Adding tracing of length "+IJ.d2s(currTracing.length(),3)+" "+NJ.units);
	    tracings.add(currTracing);
	    if (NJ.adg != null) NJ.adg.reset();
	    IJ.showStatus("Added tracing");
	}
    }
	
    private void finishCurrSegment() {

	if (currSegment.nrpoints() < 2) {
	    NJ.write("Dumping segment of zero units length");
	    IJ.showStatus("Dumped segment");
	} else {
	    NJ.write("Adding segment of length "+IJ.d2s(currSegment.length(),3)+" "+NJ.units);
	    currTracing.add(currSegment.duplicate());
	    IJ.showStatus("Added segment");
	}
	
	currSegment.reset();
    }

    private void updateCurrSegment() {

	currSegment.reset();

	if (currPoint.x != clckPoint.x || currPoint.y != clckPoint.y) {
	    
	    // Extract current segment from direction map:
	    currSegment.add(currPoint);
	    final Point pnt = new Point(currPoint);
	    if (bManualTracing) {
		currSegment.add(clckPoint);
		currSegment.reverse();
	    } else {
		while (pnt.x != clckPoint.x || pnt.y != clckPoint.y) {
		    switch (dirsimage[pnt.y][pnt.x]) {
		    case 0: { pnt.x = clckPoint.x; pnt.y = clckPoint.y; break; }
		    case 1: { --pnt.x; --pnt.y; break; }
		    case 2: { --pnt.y; break; }
		    case 3: { ++pnt.x; --pnt.y; break; }
		    case 4: { --pnt.x; break; }
		    case 5: { ++pnt.x; break; }
		    case 6: { --pnt.x; ++pnt.y; break; }
		    case 7: { ++pnt.y; break; }
		    case 8: { ++pnt.x; ++pnt.y; break; }
		    }
		    currSegment.add(pnt);
		}
		currSegment.reverse();
		// Smooth and subsample current segment:
		if (bSmoothSegment) smoothsample();
	    }
	}
    }

    private void smoothsample() {

	// Copy current segment with borders:
	final Point pnt = new Point();
	currSegment.element(0,pnt);
	ssmpSegment.reset();
	for (int i=0; i<NJ.halfsmoothrange; ++i) { ssmpSegment.add(pnt); }
	final int clckPoint = currSegment.nrpoints() - 1;
	for (int i=0; i<=clckPoint; ++i) { currSegment.element(i,pnt); ssmpSegment.add(pnt); }
	for (int i=0; i<NJ.halfsmoothrange; ++i) { ssmpSegment.add(pnt); }
	    
	// Smooth and subsample except first and last point:
	int smppos = NJ.halfsmoothrange;
	ssmpSegment.element(smppos,pnt);
	currSegment.reset();
	currSegment.add(pnt);

	final float kernval = 1.0f/(2*NJ.halfsmoothrange + 1);
	final int lastsmp = clckPoint + NJ.halfsmoothrange;
	smppos += NJ.subsamplefactor;
	while (smppos < lastsmp) {
	    ssmpSegment.element(smppos,pnt);
	    float xpos = kernval*pnt.x;
	    float ypos = kernval*pnt.y;
	    for (int i=1; i<=NJ.halfsmoothrange; ++i) {
		ssmpSegment.element(smppos+i,pnt);
		xpos += kernval*pnt.x;
		ypos += kernval*pnt.y;
		ssmpSegment.element(smppos-i,pnt);
		xpos += kernval*pnt.x;
		ypos += kernval*pnt.y;
	    }
	    pnt.x = FMath.round(xpos);
	    pnt.y = FMath.round(ypos);
	    currSegment.add(pnt);
	    smppos += NJ.subsamplefactor;
	}

	ssmpSegment.element(lastsmp,pnt);
	currSegment.add(pnt);
    }

    public void mouseReleased(final MouseEvent e) {
    	switch (NJ.ntb.currentTool()) {
    	case TracingToolbar.RECTANGLE: {
    		imc.mouseReleased(e);
    		break;
    	}
    	}
    }
	
    private void showValue(final int xp, final int yp) {
	IJ.showStatus("Pointer at ("+xp+","+yp+")");
    }

    private void scroll(final int x, final int y) {

	final Rectangle vofRect = imc.getSrcRect();
	final double mag = imc.getMagnification();
        int newx = scrlPoint.x - (int)(x/mag);
        int newy = scrlPoint.y - (int)(y/mag);
        if (newx < 0) newx = 0;
        if (newy < 0) newy = 0;
        if ((newx + vofRect.width) > iXSize) newx = iXSize - vofRect.width;
        if ((newy + vofRect.height) > iYSize) newy = iYSize - vofRect.height;
        vofRect.x = newx;
        vofRect.y = newy;
        imp.draw();
	Thread.yield();
    }   
    
    void eraseTracings() { tracings.reset(); redraw(); }

    void resetTracings() { tracings.reset(); }

    void setCursor(final Cursor c) { imc.setCursor(c); }

    void loadTracings(final String dir, final String file) {

	final String path = (dir.endsWith(File.separator) ? dir : dir+File.separator) + file;
	NJ.write("Loading tracings from "+path);

	try {
	    final BufferedReader br = new BufferedReader(new FileReader(path));
	    if (!br.readLine().startsWith("// NeuronJ Data File")) throw new IOException();
	    final float version = Float.valueOf(br.readLine()).floatValue();

	    float brscale = NJ.scale;
	    float brgamma = NJ.gamma;
	    int brsnaprange = NJ.snaprange;
	    int brdijkrange = NJ.dijkrange;
	    int brhalfsmoothrange = NJ.halfsmoothrange;
	    int brsubsamplefactor = NJ.subsamplefactor;
	    float brxsize = NJ.xsize;
	    float brysize = NJ.ysize;
	    String brunits = NJ.units;
	    boolean brautosave = NJ.autosave;
	    boolean brdebug = NJ.debug;
	    final String[] brtypes = new String[11];
	    final Color[] brtypecolors = new Color[11];
	    final String[] brclusters = new String[11];
	    final Tracings brtracings = new Tracings();

	    if (version <= 1.01f) {
		NJ.write("   Opened NeuronJ version "+version+" data file");

		br.readLine(); // Parameters
		brscale = Float.valueOf(br.readLine()).floatValue();
		brgamma = Float.valueOf(br.readLine()).floatValue();
		brsnaprange = Integer.valueOf(br.readLine()).intValue();
		brdijkrange = Integer.valueOf(br.readLine()).intValue();
		brhalfsmoothrange = Integer.valueOf(br.readLine()).intValue();
		brsubsamplefactor = Integer.valueOf(br.readLine()).intValue();
		brxsize = Float.valueOf(br.readLine()).floatValue();
		brysize = Float.valueOf(br.readLine()).floatValue();
		brunits = br.readLine();
		brautosave = Boolean.valueOf(br.readLine()).booleanValue();
		brdebug = Boolean.valueOf(br.readLine()).booleanValue();
		NJ.write("   Read parameters");

		br.readLine(); // Type names and colors
		for (int i=0; i<=10; ++i) {
		    brtypes[i] = br.readLine();
		    brtypecolors[i] = NJ.colors[Integer.valueOf(br.readLine()).intValue()]; }
		NJ.write("   Read type names and colors");

		br.readLine(); // Cluster names
		for (int i=0; i<=10; ++i) brclusters[i] = br.readLine();
		NJ.write("   Read cluster names");

		// Tracings
		String line = br.readLine();
		while (line.startsWith("// Tracing")) {
		    final Tracing tracing = new Tracing();
		    tracing.id(Integer.valueOf(br.readLine()).intValue());
		    tracing.type(Integer.valueOf(br.readLine()).intValue());
		    tracing.cluster(Integer.valueOf(br.readLine()).intValue());
		    tracing.label(br.readLine());
		    line = br.readLine();
		    while (line.startsWith("// Segment")) {
			final Segment segment = new Segment();
			final Point pnt = new Point();
			line = br.readLine();
			while (!line.startsWith("//")) {
			    pnt.x = Integer.valueOf(line).intValue();
			    pnt.y = Integer.valueOf(br.readLine()).intValue();
			    segment.add(pnt);
			    line = br.readLine();
			}
			if (segment.length() > 0.0) tracing.add(segment);
		    }
		    if (tracing.length() > 0.0) brtracings.add(tracing);
		}
		NJ.write("   Read tracings");
	    }

	    br.close();

	    boolean bScaleChanged = false;
	    if (NJ.scale != brscale) {
		bScaleChanged = true;
		NJ.scale = brscale;
	    }
	    boolean bGammaChanged = false;
	    if (NJ.gamma != brgamma) {
		bGammaChanged = true;
		NJ.gamma = brgamma;
	    }
	    NJ.snaprange = brsnaprange;
	    NJ.dijkrange = brdijkrange;
	    NJ.halfsmoothrange = brhalfsmoothrange;
	    NJ.subsamplefactor = brsubsamplefactor;
	    NJ.xsize = brxsize;
	    NJ.ysize = brysize;
	    NJ.units = brunits;
	    NJ.autosave = brautosave;
	    NJ.types = brtypes;
	    NJ.typecolors = brtypecolors;
	    NJ.clusters = brclusters;
	    tracings = brtracings;
	    NJ.write("   Effectuated read data");

	    NJ.write("Done");
	    NJ.debug = brdebug;
	    NJ.save = false;

	    IJ.showStatus("Loaded tracings from "+path);

	    if (bComputedCosts) {
		if (bScaleChanged) { computeCosts(); doDijkstra(); }
		else if (bGammaChanged) { doDijkstra(); }
	    }

	} catch (IOException ioe) {
	    NJ.write("Unable to read from file");
	    NJ.error("Unable to read from file.");
	    IJ.showStatus(NJ.COPYRIGHT);
	} catch (NumberFormatException nfe) {
	    NJ.write("Error reading from file");
	    NJ.error("Error reading from file.");
	    IJ.showStatus(NJ.COPYRIGHT);
	}

	if (NJ.mdg != null) NJ.mdg.reset();
	if (NJ.adg != null) NJ.adg.reset();

	redraw();
    }

    void closeTracings() {

	String status = "Dumped image";
	if (NJ.save) {
	    if (NJ.autosave) {
		NJ.write("Automatically saving tracings");
		saveTracings(NJ.workdir,NJ.imagename+".ndf");
		status += " but saved tracings";
	    } else {
		NJ.write("Asking user to save tracings");
		final YesNoDialog ynd =
		    new YesNoDialog("Save","Do you want to save the tracings?");
		if (ynd.yesPressed()) {
		    final FileDialog fdg = new FileDialog(IJ.getInstance(),"NeuronJ: Save",FileDialog.SAVE);
		    fdg.setFilenameFilter(new ImageDataFilter());
		    fdg.setFile(NJ.imagename+".ndf");
		    fdg.show();
		    final String dir = fdg.getDirectory();
		    final String file = fdg.getFile();
		    fdg.dispose();
		    if (dir != null && file != null) {
			saveTracings(dir,file);
			status += " but saved tracings";
		    } else {
			NJ.write("Dumping tracings");
			status += " and tracings";
		    }
		} else {
		    NJ.write("Dumping tracings");
		    status += " and tracings";
		}
	    }
	} else NJ.write("No need to save current tracings");

	costs = null; // To free more memory
	IJ.showStatus(status);
    }
    
    void saveTracings(final String dir, final String file) {

	final String path = (dir.endsWith(File.separator) ? dir : dir+File.separator) + file;
	NJ.write("Saving tracings to "+path);

	try {
	    final FileWriter fw = new FileWriter(path);
	    fw.write("// NeuronJ Data File - DO NOT CHANGE\n");
	    fw.write(NJ.VERSION+"\n");
	    NJ.write("   Opened NeuronJ version "+NJ.VERSION+" data file");

	    fw.write("// Parameters\n");
	    fw.write(NJ.scale+"\n");
	    fw.write(NJ.gamma+"\n");
	    fw.write(NJ.snaprange+"\n");
	    fw.write(NJ.dijkrange+"\n");
	    fw.write(NJ.halfsmoothrange+"\n");
	    fw.write(NJ.subsamplefactor+"\n");
	    fw.write(NJ.xsize+"\n");
	    fw.write(NJ.ysize+"\n");
	    fw.write(NJ.units+"\n");
	    fw.write(NJ.autosave+"\n");
	    fw.write(NJ.debug+"\n");
	    NJ.write("   Wrote parameters");

	    fw.write("// Type names and colors\n");
	    final int nrtypes = NJ.types.length;
	    for (int i=0; i<nrtypes; ++i) fw.write(NJ.types[i]+"\n"+NJ.colorIndex(NJ.typecolors[i])+"\n");
	    NJ.write("   Wrote type names and colors");

	    fw.write("// Cluster names\n");
	    final int nrclusters = NJ.clusters.length;
	    for (int i=0; i<nrclusters; ++i) fw.write(NJ.clusters[i]+"\n");
	    NJ.write("   Wrote cluster names");

	    final int nrtracings = tracings.nrtracings();
	    for (int n=0; n<nrtracings; ++n) {
		fw.write("// Tracing "+n+"\n");
		final Tracing tracing = tracings.element(n);
		fw.write(tracing.id()+"\n");
		fw.write(tracing.type()+"\n");
		fw.write(tracing.cluster()+"\n");
		fw.write(tracing.label()+"\n");
		final int nrsegments = tracing.nrsegments();

		for (int s=0; s<nrsegments; ++s) {
		    fw.write("// Segment "+s+" of Tracing "+n+"\n");
		    final Segment segment = tracing.element(s);
		    final int nrpoints = segment.nrpoints();
		    final Point pnt = new Point();

		    for (int p=0; p<nrpoints; ++p) {
			segment.element(p,pnt);
			fw.write(pnt.x+"\n"+pnt.y+"\n");
		    }
		}
	    }
	    NJ.write("   Wrote tracings");

	    fw.write("// End of NeuronJ Data File\n");
	    fw.close();
	    NJ.write("Done");
	    IJ.showStatus("Saved tracings to "+path);
	    NJ.save = false;

	} catch (IOException ioe) {
	    NJ.write("Unable to write to file");
	    NJ.error("Unable to write to file.");
	    IJ.showStatus(NJ.COPYRIGHT);
	}
    }
    
	/** to fix a bug introtuced in the new imagej version 
	    that has the consequence that the current segment is not drawn */
    
    public synchronized Object clone() {
    	return this;
    }
}
