package transforms;


/*
 * LineHoughTransform.java
 * Created on 20.07.2005
 *
 * LineHoughTransform has been derived from the class LineHoughTransformOp from the generation5 jdk (see below). 
 *  
 * This file is part of MRI Cell Image Analyzer.
 * (c) 2005-2006 Volker Bcker
 * MRI Cell Image Analyzer has been created at Montpellier RIO Imaging
 * www.mri.cnrs.fr
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 * 
 */

/*
 * LineHoughTransformOp.java
 * Created on 03 December 2004, 01:45
 *
 * Copyright 2004, Generation5. All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License as published by the Free Software 
 * Foundation; either version 2 of the License, or (at your option) any later 
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with 
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple 
 * Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

import ij.ImagePlus;
import ij.gui.NewImage;
import java.awt.Color;
import java.util.Vector;
import java.awt.geom.Line2D;

/**
 * 
 */
public class LineHoughTransform {
	
	protected Vector lines = new Vector(); 
	  /**
     * The accumulator.
     */    
    protected long[][] houghAccumulator;
    /**
     * The maximum accumulator value.
     */    
    protected long maxAccValue = -1;
    /**
     * The local peak neighbourhood.
     */    
    protected int localPeakNeighbourhood = 5;
    /**
     * The theta step.
     */    
    protected double thetaStep = 0.0;
    /**
     * The line colour used for the superimposed image (default WHITE).
     */    
    protected Color lineColor = Color.WHITE;
    
    protected ImagePlus inputImage;
    
	/**
	 * @return Returns the inputImage.
	 */
	public ImagePlus getInputImage() {
		return inputImage;
	}
	/**
	 * @param inputImage The inputImage to set.
	 */
	public void setInputImage(ImagePlus inputImage) {
		this.inputImage = inputImage;
	}
    /** Creates a new instance of LineHoughTransformOp */
    public LineHoughTransform() {
    }
    
    /**
     * Set the line colour used for the superimposed image
     * @param lc the new line colour.
     */    
    public void setLineColor(Color lc) {
        lineColor = lc;
    }
    
    /**
     * Get the line colour used.
     * @return the current line colour used.
     */    
    public Color getLineColor() {
        return lineColor;
    }
    
    /**
     * Set the local peak neighbourhood.
     * @param lpn the new local peak neighbourhood.
     */    
    public void setLocalPeakNeighbourhood(int lpn) {
        localPeakNeighbourhood = lpn;
    }
    
    /**
     * Get the local peak neighbourhood.
     * @return the current local peak neighbourhood.
     */    
    public int getLocalPeakNeighbourhood() {
        return localPeakNeighbourhood;
    }
    
    /**
     * Get the maximum value in the accumulator.
     * @return the maximum value.
     */    
    protected long getMaximum() {
        if (houghAccumulator == null) return -1;
        
        long accMax = Long.MIN_VALUE;
        for (int i=0; i<houghAccumulator.length; i++) {
            for (int j=0; j<houghAccumulator[0].length; j++) {
                if (houghAccumulator[i][j] > accMax) 
                    accMax = houghAccumulator[i][j];                
            }
        }
        
        return accMax;
    }
    
    /**
     * Returns the accumulator as an image. Values are scaled linearly according to the
     * maximum value.
     * @return the accumulator image.
     */    
    public ImagePlus getAccumulatorImage() {
        // throw exception?
        if (houghAccumulator == null) 
            return null;
        
        double scaleFactor = 255.0 / getMaximum();
        
        ImagePlus acc = NewImage.createByteImage("hough accumulator", houghAccumulator.length, houghAccumulator[0].length, 
			1, NewImage.GRAY8);
        	        
        for (int i=0; i<houghAccumulator.length; i++)
            for (int j=0; j<houghAccumulator[0].length; j++)
                acc.getProcessor().putPixelValue(i, j, (int)(houghAccumulator[i][j] * scaleFactor));
        
        return acc;
    }
    
    /**
     * Runs the line Hough Transform. Note that the input image must be an edge-only,
     * greyscale image (ie, must have been passed through an edge detector and thresholding
     * filter).
     * @param img input image.
     */    
    public void run() {
    	if (inputImage==null) return;
        if (inputImage.getType() != ImagePlus.GRAY8)
            throw new IllegalArgumentException("input image must be greyscale.");

        int width = inputImage.getWidth();
        int height = inputImage.getHeight();
        
        int tmp = Math.max(width, height);
        int h_h = (int) (Math.sqrt(2)*tmp);
        int h_w = 180;

        houghAccumulator = new long[h_w][2*h_h];
        thetaStep = Math.PI / h_w;

        int src_rgb;
        int centre_x = width / 2;
        int centre_y = height / 2;                        

        for (int i=0; i<h_w; i++) {
            for (int j=0; j<2*h_h; j++){
                houghAccumulator[i][j] = 0;
            }
        }
               
        //Now find edge points and update hough array
        for(int i = 0; i < width; i++){
            for(int j = 0; j < height; j++){
                src_rgb = inputImage.getProcessor().getPixel(i, j);
                if (src_rgb != 0) {
                    // Edge pixel found
                    for(int k = 0; k < h_w; k++){
                        //Work out the r values for each theta step
                        tmp = (int) (((i-centre_x)*Math.cos(k*thetaStep)) +
                                     ((j-centre_y)*Math.sin(k*thetaStep)));
                        
                        //Move all values into positive range for display purposes
                        //if (tmp== -1 && k ==150 ) System.out.println("LOW  ij: "+i+" "+j);
                        //if (tmp==  0 && k ==150 ) System.out.println("ON   ij: "+i+" "+j);
                        //if (tmp==  1 && k ==150 ) System.out.println("HIGH ij: "+i+" "+j);
                        tmp = tmp + h_h;
                        if (tmp < 0 || tmp >= 2*h_h) continue;
                        
                        //Increment hough array
                        houghAccumulator[k][tmp]++;
                    }
                }
            }
        }
        
    }
    
    public void computeLines(double threshold) {
         lines = new Vector();
    	 if (inputImage==null) return;
    	 maxAccValue = getMaximum();
         
         long thresh = (int) (threshold * maxAccValue);
         
         // Search for local peaks above threshold to draw
         boolean draw = false;
         int k;
         int l;
         int dt;     // test theta
         int dr;     // test offset
         int width = inputImage.getWidth();
         int height = inputImage.getHeight();
         int centre_x = width / 2;
         int centre_y = height / 2;

         int tmp = Math.max(width, height);
         int hh = (int) (Math.sqrt(2)*tmp);
         int hw = 180;
         
         for (int i=0; i<hw; i++){
             for (int j=0; j<2*hh; j++){
                 
                 if(houghAccumulator[i][j] >= thresh){
                     // This is HORRIBLY inefficient, if another local peak is
                     // found,then that is going to be the line that needs to
                     // be drawn...why check it hundreds more times?
                     if (localPeak(i, j, hw, hh, localPeakNeighbourhood) == false) continue;

                     int x1, x2, y1, y2;
                     double tsin = Math.sin(i*thetaStep);
                     double tcos = Math.cos(i*thetaStep);
                     if (i <= hw/4 || i >= (3*hw)/4) {
                         y1 = 0;
                         x1 = (int) (((j-hh) - ((y1-centre_y)*tsin)) / tcos) + centre_x;
                         y2 = height - 1;
                         x2 = (int) (((j-hh) - ((y2-centre_y)*tsin)) / tcos) + centre_x;
                     } else {
                         x1 = 0;
                         y1 = (int) (((j-hh) - ((x1-centre_x)*tcos)) / tsin) + centre_y;
                         x2 = width - 1;
                         y2 = (int) (((j-hh) - ((x2-centre_x)*tcos)) / tsin) + centre_y;
                     }
                     lines.add(new Line2D.Float(x1, y1, x2, y2));
                 }
             }
         }
    }
    
    /**
     * Returns an image showing the lines found.
     */    
    public ImagePlus getLinesImage() {
    	
		ImagePlus out = NewImage.createByteImage(inputImage.getTitle() + " lines", inputImage.getWidth(), inputImage.getHeight(), 
				inputImage.getStackSize(), NewImage.GRAY8); 
		for (int i=0; i<lines.size();i++) {
			Line2D aLine = (Line2D) lines.elementAt(i);
			out.getProcessor().drawLine((int)Math.round(aLine.getX1()), 
										   (int)Math.round(aLine.getY1()), 
										   (int)Math.round(aLine.getX2()), 
										   (int)Math.round(aLine.getY2()));
		} 
        return out;
    }
    
    private boolean localPeak(int i, int j, int hw, int hh, int neighbourhood) {
        int dt, dr;
        long peak = houghAccumulator[i][j];
        
        for (int k=-neighbourhood; k<=neighbourhood; k++) {
            for (int l = -neighbourhood; l<=neighbourhood; l++) {
                if (k == 0 && l == 0) continue;
                dt = i+k;
                dr = j+l;
                if (dr < 0 || dr >= 2*hh) continue;
                if (dt < 0) dt = dt + hw;
                if (dt >= hw) dt = dt - hw;
                if (houghAccumulator[dt][dr] > peak) {
                    return false;
                }
            }
        }
        
        return true;
    }

	/**
	 * @return Returns the lines.
	 */
	public Vector getLines() {
		return lines;
	}
}
